/* rdminor - process the cris minor frames
 *
 *	Raw telemetry data processed by cris_test_extr is the input 
 *	to this program. See end of file for a description of the 
 *	CrisData format.
 *
 * Marty Olevitch, Jan 1997
 */

#include <stdio.h>
#include <sys/types.h>
#include <string.h>	/* memcpy */
#include <errno.h>
/*#include <softcris.h>*/

/* number of bytes in the cris telemetry minor frame. This is equivalent to
 * the NUM_CRIS_TELEMETRY_BYTES seen in some non-Lexas programs. */
#define SC_CRIS_TELEMBYTES	58

/* sc_crisSubset - this is the cris portion of the spacecraft telemetry
 * data. From Robert Radocynski. */
typedef struct {
    u_int PacketSecondaryHeader;
    u_int TimeOfDay;
    u_short DayOfYear;
    u_short year;
    u_short ErrorFlags;
    u_short PhaseAngle;
    u_char facility;
    u_char subfacility;
    u_char assembly;
    u_char DataNature;
    u_char VirtualChannelId;
    u_char MajorFrameCount;
    u_char MinorFrameCount;
    u_char SpacecraftId;
    u_char FormatId;
    u_char MainBusVoltage;
    u_char MainBusCurrent;
    u_char CrisInternalTemperature1;
    u_char CrisInternalTemperature2;
    u_char CrisInterfaceTemperature;
    u_char CrisCurrent;
    u_char CrisData[SC_CRIS_TELEMBYTES];
} sc_crisSubset;

#define SC_TLM_SYNC	0x80
#define SC_TLM_CMD	0x40
#define SC_TLM_DIAG	0x20

struct sc_minorFrame {
    sc_crisSubset s;		/* the raw "cris subset" data */
    int major_frame_start;	/* 1 if beginning of major frame */
    int seq;			/* sequence no. of this minor frm in major fr */
    u_char hsk[3];		/* housekeeping bytes (including SYNC, CMD,
				 * and DIAG bits */
    u_char cmd[SC_CRIS_TELEMBYTES-3];	/* cmd string */
    int diag;			/* 1 if this is diagnostic data */
    u_char diagdata[SC_CRIS_TELEMBYTES-3];	/* the diag data */
    int nevt;			/* no. cris normal evts in this minor frame */
    /* struct sc_crisEvt e[10];	/* the cris normal evt data */
};

int sc_tlm_next_minor(struct sc_minorFrame *f, FILE *fp);
int sc_tlm_sync_major(FILE *fp);

/* buffer for storing normal evt which extends over more than 1 minor frm */
static u_char Norm_buf[1000];
static u_char *Norm_bufp;

int
main(argc, argv)
    int argc;
    char *argv[];
{
    FILE *fp;
    int i;
    struct sc_minorFrame m;
    int majorcnt = 0;
    int ret;

    if (argc == 2) {
	fp = fopen(argv[1], "r");
	if (fp == NULL) {
	    fprintf(stderr,"%s: can't open '%s' (%s)\n", argv[0], argv[1],
		strerror(errno));
	    exit(1);
	}
    } else {
	fp = stdin;
    }

    ret = sc_tlm_sync_major(fp);
    fprintf(stderr,"sync after %d frames\n", ret);
    majorcnt++;
    printf("Major %d\n", majorcnt);
    while (1) {
	ret = sc_tlm_next_minor(&m, fp);
	if (ret == 0) {
	    exit(0);
	} else if (ret == -2 || ret == -3) {
	    fprintf(stderr,"ret = %d at %d.%d\n", ret, majorcnt, m.seq);
	    continue;
	} else if (ret == -1 || ret == -5) {
	    if (feof(fp)) {
		fprintf(stderr,"EOF at %d.%d\n", majorcnt, m.seq);
		exit(0);
	    } else if (ferror(fp)) {
		fprintf(stderr,"%s: error at %d.%d: %s\n", 
		    argv[0], strerror(errno), majorcnt, m.seq);
		exit(-1);
	    } else {
		fprintf(stderr,"ret = %d at %d.%d\n", ret, majorcnt, m.seq);
		exit(-1);
	    }
	} else {
	    if (m.major_frame_start) {
		majorcnt++;
		printf("Major %d\n", majorcnt);
	    }
	    printf("\tminor %d", m.seq);
	    if (m.hsk[0] & SC_TLM_SYNC) {
		printf("\tSYNC");
	    }
	    if (m.hsk[0] & SC_TLM_CMD) {
		printf("\tCMD");
		printf(" %s", m.cmd);
	    }
	    if (m.diag) {
		printf("\tDIAG");
	    } else {
		printf("\tnorm %d", m.nevt);
		/****
		for (i=0; i<m.nevt; i++) {
		    printf(" %d", m.e[i].nbytes);
		}
		****/
	    }
	    printf("\n");
	}
    }
}

/* sc_tlm_sync_major - read through the data until we come to the start of
 * a new major frame, which is indicated by transition of the SYNC bit from
 * 0 to 1. Returns 1 when we detect the start of the major frame; 0 on EOF;
 * a negative return value indicates an error from sc_tlm_next_minor(). */
int
sc_tlm_sync_major(fp)
    FILE *fp;
{
    int cnt = 0;
    struct sc_minorFrame f;
    int ret;

    while (1) {
	cnt++;
	ret = sc_tlm_next_minor(&f, fp);
	if (ret == 0) {
	    break;
	} else if (ret < 0) {
	    break;
	} else if (f.major_frame_start) {
	    ret = cnt;
	    break;
	}
    }
    return ret;
}


/* Reads the next minor frame data from the open file pointer fp and fills 
 * in the values pointed to by the sc_minorFrame pointer f. Returns 1 if
 * all went well; 0 for EOF; -1 on read error; -2 if the command reponse
 * length is too long; -3 if a normal event length is too long; -4 on error
 * from sc_cris_unpack(); -5 if the event doesn't start with the string
 * "next frm" (which is inserted by the cris_test_extr program). */
int
sc_tlm_next_minor(f, fp)
    struct sc_minorFrame *f;
    FILE *fp;
{
    int amt;
    int amtleft;	/* no. bytes left in minor frame */
    int cmdlen;		/* length of command response (bytes) */
    static sc_crisSubset cur;
    static u_char cur_minor_frame[SC_CRIS_TELEMBYTES];
    u_char cursync;
    int nbytes_cur_frame;	/* no. bytes in current minor frame */
    static int firsttime = 1;
    int i;
    static int in_normal = 0;	/* 1 if we are processing a long normal evt */
    int j;
    u_char k;
    static u_char lastsync;
    static int len;	/* length of normal evt left to read (bytes) */
    int n;
    sc_crisSubset next;
    int ret;
    static int seq = 0;		/* sequence no. in major frame */
    char startbuf[8];

    /* We have to have read in two minor frames, the current one and the
     * "next" one. This annoying detail is necessary because the first 4
     * bytes of a minor frame is actually data belonging to the previous
     * minor frame. I think this even applies to the first minor frame in a
     * major frame.*/

    if (firsttime) {
	/* read in the first "current" frame */
	firsttime = 0;
	ret = fread(startbuf, 1, 8, fp);
	if (ret != 8 || memcmp(startbuf, "next frm", 8) != 0) {
	    return -5;
	}
	ret = fread((char *)&cur, 1, sizeof(sc_crisSubset), fp);
	if (ret != sizeof(sc_crisSubset)) {
	    if (feof(fp)) {
		return 0;
	    } else if (ferror(fp)) {
		return -1;
	    }
	}
    }

    /* read in the next frame */
    ret = fread(startbuf, 1, 8, fp);
    if (ret != 8 || memcmp(startbuf, "next frm", 8) != 0) {
	return -5;
    }
    ret = fread((char *)&next, 1, sizeof(sc_crisSubset), fp);
    if (ret != sizeof(sc_crisSubset)) {
	if (feof(fp)) {
	    return 0;
	} else if (ferror(fp)) {
	    return -1;
	}
    }

    memcpy(&f->s, &cur, sizeof(sc_crisSubset));
    f->hsk[0] = cur.CrisData[4];
    f->hsk[1] = cur.CrisData[5];
    f->hsk[2] = cur.CrisData[6];

    cursync = f->hsk[0] & SC_TLM_SYNC;
    if (!cursync && lastsync) {
	f->major_frame_start = 1;
	seq = 0;
    } else {
	f->major_frame_start = 0;
	seq++;
    }
    f->seq = seq;
    lastsync = cursync;

    cmdlen = 0;
    if (f->hsk[0] & SC_TLM_CMD) {
	/* command response present */
	cmdlen = (int)cur.CrisData[7];
	if (cmdlen > SC_CRIS_TELEMBYTES - 3) {
	    return -2;
	}
	for (j=8,i=0; i<cmdlen; i++,j++) {
	    k = cur.CrisData[j];
	    if (k == '\n') {
		f->cmd[i] = '\\';
		i++;
		f->cmd[i] = 'n';
	    } else if (k == '\r') {
		f->cmd[i] = '\\';
		i++;
		f->cmd[i] = 'r';
	    } else {
		f->cmd[i] = k;
	    }
	}
	f->cmd[i] = '\0';
    }

    i = (cmdlen == 0) ? 7 : (8 + cmdlen);
    amt = SC_CRIS_TELEMBYTES - i;
    memcpy(cur_minor_frame, cur.CrisData+i, amt);
    memcpy(cur_minor_frame + amt, next.CrisData, 4);
    nbytes_cur_frame = amt + 4;
    
    /* make "next" frame into "current" frame for next time through loop */
    memcpy(&cur, &next, SC_CRIS_TELEMBYTES);

    if (f->hsk[0] & SC_TLM_DIAG) {
	/* diagnostic data */
	memcpy(f->diagdata, cur_minor_frame, nbytes_cur_frame);
	f->nevt = 0;	/* no normal evts */
	f->diag = 1;
	return 1;
    } else {
	/* normal evts */
	if (f->major_frame_start) {
	    /* a new major frame terminates any pending normal evt */
	    in_normal = 0;
	}
	f->diag = 0;
	f->nevt = 0;
	i = n = 0;
	amtleft = nbytes_cur_frame;

	if (in_normal) {
	    /* working on normal evt continued from previous minor frame */
	    if (len > amtleft) {
		/* evt continues into next minor frame */
		memcpy(Norm_bufp, cur_minor_frame, amtleft);
		Norm_bufp += amtleft;
		len -= amtleft;
		return 1;
	    } else {
		/* evt ends in this minor frame */
		memcpy(Norm_bufp, cur_minor_frame, len);
		/**********
		ret = sc_cris_unpack(Norm_buf, &(f->e[0]), SC_CAM2);
		if (ret) {
		    return -4;
		}
		**********/
		i += len;
		n++;
		f->nevt++;
		in_normal = 0;
		amtleft -= len;
		/* continue below to see if more evts in this minor frame */
	    }
	}

	len = (int) cur_minor_frame[i];
	while (len > 0) {
	    if (len > amtleft) {
		/* evt continues into next minor frame */
		in_normal = 1;
		memcpy(Norm_buf, cur_minor_frame+i, amtleft);
		len -= amtleft;
		Norm_bufp = Norm_buf + amtleft;
		return 1;
	    }
	    /***
	    ret = sc_cris_unpack(cur_minor_frame + i, &(f->e[n]), SC_CAM2);
	    if (ret) {
		return -4;
	    }
	    ****/
	    n++;
	    f->nevt++;
	    i += len;
	    if (i >= nbytes_cur_frame) {
		break;
	    }
	    len = (int)cur_minor_frame[i];
	}
	return 1;
    }
}

/********************************
		Format of CRIS major and minor frames

Byte	Item	Count	Bits	Sum	Description


	A. Major frame = 256 minor frames

0	a	4	8	32	4 byte offset
4	b	128	58x8	xxx	minor frames w/SYNC=0
xxx	c	128	58x8	yyy	minor frames w/SYNC=1



	B. Minor frame = 58 bytes


		Housekeeping bytes

	(Note that these bytes really start at an offset of 4 bytes from
	the start of the data. Those 4 bytes contain the last 4 bytes of
	the previous minor frame.)

0	0	1	1	1	SYNC = 0 for 1st 128 minor frames,
					     = 1 for 2nd 128 minor frames
	1	1	1	2	CMD  = 1 if we have a cmd response
	2	1	1	3	DIAG = 1 for diagnostic mode
	3	1	5	8	housekeeping bits
1	4	1	8	16	more housekeeping bits
2	5	1	8	24	more housekeeping bits


		Command response (if item #1 [CMD bit] is true)

3	6	1	8	8	length L of response (bytes)
4	7	L	Lx8	8+(Lx8)	command echo (ascii chars)


		CRIS normal events (if item #2 [DIAG bit] is false)

	(May be continued from previous minor frame unless this is the
	start of a new major frame. Note that the event may have a
	length of 0, in which case the rest of the current minor frame
	should be skipped.)

3 or
4+L	8	N	X	Y	CRIS Normal events 


		DIAGNOSTIC mode data (if item #2 [DIAG bit] is true)

	(Can extend across minor and major frames. Can interrupt a CRIS
	normal event, which will then resume after Diag mode is done.)

3 or 
4+L	9	N	X	Y	Diagnostic mode data (Normal evt +
					stack data + SOFT image)
********************************/

