/* conv_msu - convert a SOFT MSU data stream from big- to little-endian
 *	or vice-versa.
 *
 * USAGE: conv_msu [-o outfile] [infile ...]
 *
 * Marty Olevitch, July 1997
 */

#include <stdio.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <string.h>

#include <local.h>
#include <softcris.h>
#include <msu.h>

int Debug = 0;
char *Progname;

char **Argvp;				/* scan pointer for nextfile() */
char *Nullargv[] = { "-", NULL };	/* dummy argv for case of no args */
char *Inname;				/* filename for messages etc. */
u_long Bytecnt;				/* line number for messages etc. */
FILE *In = NULL;			/* current input file */
FILE *Ofp;	/* output file */
char *Outname;

#ifdef MAXDATA
#undef MAXDATA
#endif
#define MAXDATA 200000

static u_char *Db;	/* data buffer */
static u_char *Swapbuf;	/* swapped data */

#ifdef UTZOOERR
extern char *mkprogname();
#else
#define	mkprogname(a)	(a)
#endif

char *nextfile(void);

/* main - parse arguments and handle options */
int
main(argc, argv)
    int argc;
    char *argv[];
{

    int c;
    int errflg = 0;
    extern int optind;
    extern char *optarg;
    void process(void);

    Progname = mkprogname(argv[0]);
    Ofp = stdout;

    while ((c = getopt(argc, argv, "do:")) != EOF) {
	switch (c) {
	case 'o':	/* set output file (default stdout) */
	    Outname = optarg;
	    Ofp = fopen(optarg, "w");
	    if (Ofp == NULL) {
		fatal("%s: can't open '%s'", Progname, optarg);
	    }
	    break;
	case 'd':	/* Debugging. */
	    Debug++;
	    break;
	case '?':
	default:
	    errflg++;
	    break;
	}
    }
    if (errflg) {
	fprintf(stderr, "USAGE: %s ", Progname);
	fprintf(stderr, "[-o outfile] [-d] [file] ...\n");
	exit(2);
    }

    Db = (u_char *)malloc((unsigned)MAXDATA);
    if (Db == NULL) {
	fatal("%s: bad Db malloc", Progname);
    }

    Swapbuf = (u_char *)malloc((unsigned)MAXDATA);
    if (Swapbuf == NULL) {
	fatal("%s: bad Swapbuf malloc", Progname);
    }

    if (optind >= argc) {
	Argvp = Nullargv;
    } else {
	Argvp = &argv[optind];
    }
    Inname = nextfile();
    if (Inname != NULL) {
	process();
    }
    exit(0);
}

/* nextfile - switch files */
char *				/* filename */
nextfile(void)
{
    char *namep;
    struct stat statbuf;

    if (In != NULL) {
	(void) fclose(In);
    }

    namep = *Argvp;
    if (namep == NULL) {	/* no more files */
	return(NULL);
    }
    Argvp++;

    if (STREQ(namep, "-")) {
	In = stdin;
	namep = "stdin";
    } else {
	In = efopen(namep, "r");
	if (fstat(fileno(In), &statbuf) < 0) {
	    errmesg("can't fstat `%s'", namep);
	}
	if ((statbuf.st_mode & S_IFMT) == S_IFDIR) {
	    errmesg("`%s' is directory!", namep);
	}
    }

    Bytecnt = 0;
    return(namep);
}

/* process - process input data */
void
process(void)
{
    u_char *cp;
    u_short exp_id;
    int nitems = sizeof(u_char);
    u_char *p;
    int ret;
    long size;
    u_short stmp;
    u_short type;
    long vidsize, vidsize2;

    while (1) {
	/* Eat byte until we get a HDR_HDR */
	while (1) {
	    if ((ret = fread(Db, 1, 1, In)) != 1)  {
		if (ferror(In)) {
		    errmesg("%s: bad read (1) file '%s'", Progname, Inname);
		}
		Inname = nextfile();
		if (Inname == NULL) {
		    exit(0);
		}
	    }
	    if (Db[0] == ALIGN_FILL) {
		continue;
	    }
	    if ((ret = fread(Db+1, 1, 1, In)) != 1)  {
		if (ferror(In)) {
		    errmesg("%s: bad read (2) file '%s'", Progname, Inname);
		}
		Inname = nextfile();
		if (Inname == NULL) {
		    exit(0);
		}
	    }
	    cp = Db;
	    stmp = *(u_short *)cp;
	    if (stmp != HDR_HDR)
		continue;
	    break;
	}
	if ((ret = fread(Db+2, 1, 2, In)) != 2)  {
	    if (ferror(In)) {
		errmesg("%s: sc_msu_next_evt: bad read (3)\n", Progname);
	    }
	    Inname = nextfile();
	    if (Inname == NULL) {
		exit(0);
	    }
	}
	cp = GET_EVT_HDR(Db);
	stmp = *(u_short *)cp;
	if (stmp == SPILL_HDR || stmp == b_SPILL_HDR) {
	    ret = fread(Db+4, 1, 4, In);
	    if (ret != 4) {
		if (ferror(In)) {
		    errmesg("%s: bad read (4) file '%s'", Progname, Inname);
		}
		Inname = nextfile();
		if (Inname == NULL) {
		    exit(0);
		}
		continue;
	    }
	    sc_msu_reorder_spill(Db, Swapbuf);
	    ret = fwrite(Swapbuf, sizeof(u_char), SPILLSIZ, Ofp);
	    if (ret != SPILLSIZ) {
		if (ferror(Ofp)) {
		    fatal("%s: bad write to output file '%s'", Progname, Outname);
		}
	    }
	} else if (stmp == EVT_HDR || stmp == b_EVT_HDR) {
	    /* read in std stuff */
	    ret = fread(Db+4, 1, STDSIZ - 4, In);
	    if (ret != STDSIZ - 4) {
		if (ferror(In)) {
		    errmesg("%s: bad read (5) file '%s'", Progname, Inname);
		}
		Inname = nextfile();
		if (Inname == NULL) {
		    exit(0);
		}
		continue;
	    }
	    ret = sc_msu_reorder_std(Db, Swapbuf);
	    if (ret) {
		errmesg2("%s: bad std stuff file '%s'", Progname, Inname);
		continue;
	    }
	    if (stmp == EVT_HDR) {
		p = GET_size(Db);
		size = *(long *)p;
		p = GET_type(Db);
		type = *(u_short *)p;
	    } else {
		p = GET_size(Swapbuf);
		size = *(long *)p;
		p = GET_type(Db);
		type = *(u_short *)p;
	    }
	    ret = fread(Db+STDSIZ, 1, size-STDSIZ, In);
	    if (ret != size-STDSIZ) {
		if (ferror(In)) {
		    errmesg("%s: bad read (6) file '%s'", Progname, Inname);
		}
		Inname = nextfile();
		if (Inname == NULL) {
		    exit(0);
		}
		continue;
	    }
	    if (type == T_REAL || type == T_CAL_V) {
		ret = sc_msu_reorder_real(Db, Swapbuf+STDSIZ, &size, &vidsize,
		    &vidsize2, &exp_id);
		if (ret) {
		    errmesg2("%s: bad real reorder file '%s'", Progname, Inname);
		    continue;
		}
	    } else if (type == T_REPLY) {
		ret = sc_msu_reorder_reply(Db, Swapbuf+STDSIZ);
		if (ret) {
		    errmesg2("%s: bad reply reorder file '%s'", Progname, Inname);
		    continue;
		}
	    } else if (type == T_PHA_ONLY) {
		ret = sc_msu_reorder_pha_only(Db, Swapbuf+STDSIZ, &size);
		if (ret) {
		    errmesg2("%s: bad pha_only reorder file '%s'", Progname, Inname);
		    continue;
		}
	    }
	    ret = fwrite(Swapbuf, sizeof(u_char), size, Ofp);
	    if (ret != size) {
		if (ferror(Ofp)) {
		    fatal("%s: bad real write file '%s'", Progname, Outname);
		}
	    }
	}
    }
}
