/* blob.c
 *
 *  Routines to format the video data from a hodoscope camera into
 *  blob structures, and calculate centroids for each blobs.
 *
 *  Marty Olevitch, with Mike Westphall, Spring 1993, Summer 1993
 *	Jan 1994 (softlib version)
 *	May 1996 (simplified version for cris images)
 *	Sep 1996 (other changes)
 *	Nov 1996
 */

static char cvsid[] = "$Id: blob.c,v 1.13 1998/09/09 18:25:26 marty Exp $";

#include <stdio.h>
#include <string.h>
#include <math.h>

#include "softcris.h"
#include "msu.h"
#include "traj.h"

/* Phony exp_id to flag cris-generated images. Should fit into a u_short. */
#define CRIS_BLOBIFY_ID	0xccc	

#define MAX_IMAGE_SIZE	60000	/* must correspond with CRIS software */
static int Blob_init;		/* nonzero if cris blob module initialized */

/* Pix_discrim - any pixel value below this will be ignored in the
 * phase of collecting pixels into blobs in blobify(). Set by
 * sc_blob_set_discrim(). */
static int Pix_discrim = 0;

/* Blob_pix_ptr: a pointer to an array that will store pixel values for
 * pixels that are part of blobs. The memory should be allocated in the
 * application program and is assumed to be correctly sized to hold all the
 * data. See the blob_pix_ptr parameter to blobify(). This is a kludge to
 * allow us to see pixel values in the blobs. */
static u_short *Blob_pix_ptr;

/* Blob_pix_per_blob: ptr to an array of ints that will give us the number
 * of pixels in each blob. */
static int *Blob_pix_per_blob;
static int *Blob_npix;
static long *Blob_intensity;
static int Totpix;	/* total number of pixels in last blobify evt */

/* To enable blob algorithm debugging output, make BLOB_DEBUG = 1. The
 * debugging output will be found in the output file. But this really slows
 * down the data acquisition a lot. */
#define BLOB_DEBUG 0
#define BDFILENAME "blobdbg.out"
static int Blob_debug;
static FILE *bdf;	/* blob_debug output file */

static long abs_evtno;

/* An historic note: */
#define	ACE_NXCOL	550	/* old ACE (pre93) video frame width */
#define SOF_NXCOL	388	/* SOFIE and ACE93 video frame width */

#define NXCOL		SC_SOFT_WIDTH	/* no. of video frame columns */
#define YROWS		SC_SOFT_HEIGHT	/* no. of video frame rows */

#define MIN_NUM_PIX	5	/* default minimum # of pix to form a blob */
static int min_num_pix;		/* min # pix to form blob (set in blobify()) */

#define MAX_NUM_PIX	1000	/* maximum num of pix allowed in a blob */

#define MINYLINES	2	/* minimum # of ylines to form a blob */
#define MAX_SEG		16	/* maximum # of segments that any given
				 * yline in a blob may be broken into */

/* cuts */
#define X_MIN_CUT	0  	/* set equal to 0 for no cuts */
#define X_MAX_CUT	NXCOL	/* set equal to NXCOL for no cuts */
#define Y_MIN_CUT	0  	/* set equal to 0 for no cuts */
#define Y_MAX_CUT	YROWS	/* set equal to YROWS for no cuts */

static int *x = NULL;	/* x-coords */
static int *y = NULL;	/* y-coords */
static int *pix = NULL;	/* pixel values */

static double *Dx_cen = NULL;	/* array of centroids (doubles) */
static double *Dy_cen = NULL;	/* array of centroids (doubles) */

static int *Can_assign;   /* can assign segment to existing blob */
static int *Combine_list; /* list of blobs that can be combined */

/* Maxblobs - the maximum number of blobs we expect to use. The default 
 * may be changed by running the routine sc_set_maxblobs(). The value 
 * of Maxblobs is used to size a number of arrays in this file and in 
 * traj.c via the sc_get_maxblobs() function. */
static int Maxblobs = SC_MSU_MAXBLOBS;

/* blob segment details - for each event, we store the number of blobs, 
 * the number of segments in each blob, and the total number of segments
 * for the event. This information can be retrieved by the
 * sc_blob_segments() function. When a new event is processed, it
 * overwrites the prior events information. */
static int Lastevt_nblobs;	/* no. blobs in last evt */
static int Lastevt_totblobseg;	/* total no. segs in good blobs only */
static int Lastevt_totseg;	/* total no. segs (including not in blobs) */
static int *Lastevt_blobsegs = NULL;	/* no. segs each good blob */

/* BLOB structure - information about each blob that we find */
struct blob {
    int idx[MAX_NUM_PIX];
    int y_line;		/* current y line read in */
    int prev_y_line;	/* previous y line */
    int npix;		/* Number of pixels in blob */
    int nlines;		/* Number of y lines in blob */
    int begin[MAX_SEG][2]; /* x-coord of begin of previous & current y_line */
    int end[MAX_SEG][2];   /* x-coord of end of previous & current y_line */
    int	segment[2];	/* # of segments in curr. and prev. yline */
    int totseg;		/* total number of segments in this blob */
};
typedef struct blob BLOB;
static BLOB *blob = NULL;

#define PREV	0	/* index of previous line begin, end in BLOB */
#define CURR	1	/* index of current line begin, end  in BLOB */

/* blob status */
#define BS_UNUSED   0   /* this BLOB is not in use */
#define BS_FILLING  1   /* BLOB currently being accumulated */
#define	BS_CLOSED   2   /* BLOB filled and finished */
#define BS_OVFLOW   3   /* BLOB w/too many pixels (can still accumulate) */
#define BS_ACTIVE(x)	((x == BS_FILLING) || (x == BS_OVFLOW))
static int *Blob_stat = NULL;

#define NBLOBERRS   7

/* blob error indices */
#define TOO_MANY_PIX	0   /* too many pixels in blob (combine()) */
#define SEG_OVFLOW	1   /* segment overflow (combine()) */
#define TOO_MANY_BLOBS	2   /* too many blobs found */
#define ASSIGN_ERR	3   /* couldn't do blob assignment */
#define BOGUS_NFIT	4   /* bogus value of nfits */
#define TOO_MANY_PIX_A	5   /* too many pixels in blob (assign()) */
#define SEG_OVFLOW_A	6   /* segment overflow (assign()) */

static int Bloberr_counts[NBLOBERRS];

static int Bloberr_masks[] = { 0x0001, 0x0002, 0x0004, 0x0008,
			0x0010, 0x0020, 0x0040 };

/* blob error phrases */
static char *bloberr_phrases[] = {
    "PixC",	/* too many pixels in blob (combine()) */
    "SegC",	/* segment overflow, flushing (combine()) */
    "Nblb",	/* too many blobs found */
    "Basm",	/* couldn't find correct blob assignment */
    "Nfit",	/* bogus value of nfits */
    "PixA",	/* too many pixels in blob (assign()) */
    "SegA",	/* too many segments in blob (assign()) */
};

static int blob_init(void);
static void find_blobs(int pix_tot, double *x_cen, double *y_cen,
	   int *num_cen, char *errormsg);
static void struct_init(BLOB *bp);
static int assign(BLOB *bp, int start, int total, int nblob);
static int combine(BLOB *b1, BLOB *b2, int n1, int n2);
static void calculate(BLOB *bp, int n, double *x_cen, double *y_cen);
static void update_blob_limits(int amt);


/* blob_init - initialize some data structures */
static int
blob_init()
{
    int i;

    Blob_init = 0;

    if (Blob_stat != NULL) {
	free(Blob_stat);
    }
    Blob_stat = (int *)calloc(Maxblobs, sizeof(int));
    if (Blob_stat == (int *)NULL) {
	return -1;
    }

    if (Lastevt_blobsegs != NULL) {
	free(Lastevt_blobsegs);
    }
    Lastevt_blobsegs = (int *)calloc(Maxblobs, sizeof(int));
    if (Lastevt_blobsegs == (int *)NULL) {
	return -1;
    }

    if (blob != NULL) {
	free(blob);
    }
    blob = (BLOB *)calloc(Maxblobs, sizeof(BLOB));
    if (blob == (BLOB *)NULL) {
	return -1;
    }
    Blob_debug = BLOB_DEBUG;
    if (Blob_debug) {
	bdf = fopen(BDFILENAME, "w");
	if (bdf == (FILE *)NULL) {
	    fprintf(stderr, "blob_init: can't open blob debug file '%s'\n",
		 BDFILENAME);
	    Blob_debug = 0;
	} else {
	    fprintf(bdf, 
	    	"WARNING: using blob debugging. Expect slow data rate!\n");
	    fprintf(bdf, "Start\n");
	}
    }

    /* allocate the x, y, and pix arrays */
    if (x != NULL) {
	free(x);
    }
    x = (int *)calloc(MAX_IMAGE_SIZE, sizeof(int)*2);
    if (x == (int *)NULL) {
    	fprintf(stderr, "blob_init: can't alloc x array\n");
	return(-1);
    }

    if (y != NULL) {
	free(y);
    }
    y = (int *)calloc(MAX_IMAGE_SIZE, sizeof(int)*2);
    if (y == (int *)NULL) {
	fprintf(stderr, "blob_init: can't alloc y array\n");
	return(-1);
    }

    if (pix != NULL) {
	free(pix);
    }
    pix = (int *)calloc(MAX_IMAGE_SIZE, sizeof(int)*2);
    if (pix == (int *)NULL) {
	fprintf(stderr, "blob_init: can't alloc pix array\n");
	return(-1);
    }

    if (Dx_cen != NULL) {
	free(Dx_cen);
    }
    Dx_cen = (double *)calloc(Maxblobs, sizeof(double));
    if (Dx_cen == (double *)NULL) {
	return -1;
    }

    if (Dy_cen != NULL) {
	free(Dy_cen);
    }
    Dy_cen = (double *)calloc(Maxblobs, sizeof(double));
    if (Dy_cen == (double *)NULL) {
	return -1;
    }

    if (Can_assign != NULL) {
	free(Can_assign);
    }
    Can_assign = (int *)calloc(Maxblobs, sizeof(int));
    if (Can_assign == (int *)NULL) {
	return -1;
    }

    if (Combine_list != NULL) {
	free(Combine_list);
    }
    Combine_list = (int *)calloc(Maxblobs, sizeof(int));
    if (Combine_list == (int *)NULL) {
	return -1;
    }

    Lastevt_nblobs = 0;
    Lastevt_totseg = 0;
    for (i=0; i<Maxblobs; i++) {
	Lastevt_blobsegs[i] = 0;
    }

    Blob_init = 1;
    return 0;
}

/* sc_msu_blobify - parse the video data into x, y, and pix arrays for the
 * find_blobs() routine. Returns the number of centroids found if all went
 * well, else returns -1. The x- and y- coords are returned in the arrays
 * pointed to by the x and y arguments. Note that the centroid coordinates
 * are in half-pixel units, both the integer and floating point values. The
 * integer values have been rounded to the nearest integer (not
 * truncated!). */
int
sc_msu_blobify(e, x_cen, y_cen, errormsg, blob_pix_ptr, blob_npix,
	blob_int, pix_per_blob, min_pix, Fx_cen, Fy_cen, vidsys)
    struct sc_msuEvt *e;	/* ptr to sc_msuEvt structure for this event (in) */
    int *x_cen;    	/* ptr to array of centroid x-coords (out) */
    int *y_cen;		/* ptr to array of centroid y-coords (out) */
    char *errormsg;	/* ptr to string that can hold a message (out) */
    /* blob_pix_ptr: ptr to shorts that will hold pixel values that are in 
     * blobs. If NULL, no pixel values are saved. */
    u_short *blob_pix_ptr;
    int *blob_npix;	/* no. of pixels in blobs found */
    long *blob_int;	/* ptr to blob intensity values (long) */
    int *pix_per_blob;	/* ptr to no. of pixels in each blob (int) */
    int min_pix;	/* min. # pixels to form a blob - use default if < 0 */
    double *Fx_cen;	/* ptr to array of centroid x-coords - floating point */
    double *Fy_cen;	/* ptr to array of centroid y-coords - floating point */
    int vidsys;		/* vidsys: which camera to use; either 1 or 2 */ 
{
    u_char *cp;
    int curx, cury;
    u_short *dwp;
    int i;
    u_int isize;
    int max_blobs;	/* Maxblobs (see below) */
    int num_cen;	/* number of centroids found by find_blobs() */
    u_char *p;		/* ptr to start of raw video data (input) */
    u_short stmp;
    int val;

    if (!Blob_init) {
	if (blob_init()) {
	    return -1;
	}
    }

    Blob_pix_ptr = blob_pix_ptr;	/* used in calculate() below */
    Blob_pix_per_blob = pix_per_blob;	/* used in calculate() below */
    Blob_npix = blob_npix;
    if (Blob_npix != NULL) {
	*Blob_npix = 0;
    }
    Blob_intensity = blob_int;

    if (min_pix < 0) {
	min_num_pix = MIN_NUM_PIX;
    } else {
	min_num_pix = min_pix;
    }

    if (vidsys == SC_CAM1) {
	p = e->vidptr;
    } else if (vidsys == SC_CAM2) {
	p = e->vidptr2;
    } else {
	sprintf(errormsg,
	    "sc_msu_blobify: vidsys must be 1 or 2, not %d\n", vidsys);
	fprintf(stderr, "%s", errormsg);
	return -1;
    }

    max_blobs = Maxblobs;
    if (e->exp_id == CRIS_BLOBIFY_ID) {
	/* cris-generated image; make up phony abs_evtno for err msgs. */
	abs_evtno = 0L;
    } else {
	/* not a cris-generated image; use real abs_evtno. */
	abs_evtno = e->abs;
    }

    Totpix = 0;

    /* p should be pointing at the start of video data */
    dwp = (u_short *)p;
    
    stmp = *dwp++;
    i = 0;
    while (stmp != SC_END_HDR) {
	if (SC_IS_VIDEO_HDR(stmp)) {
	    if (SC_IS_Y_HDR(stmp)) {
		cury = SC_Y_VAL(stmp);
	    } else if (SC_IS_X_HDR(stmp)) {
		curx = SC_X_VAL(stmp);
	    }
	} else {    /* data */
	    val = SC_VID_DATA(stmp);
#ifdef NOTDEF
	    val >>= 4;	    /* 8-bit value */
#endif
	    if (val >= Pix_discrim) {
		y[i] = cury;
		x[i] = curx;
		pix[i] = val;
		Totpix++;
		if (i++ >= MAX_IMAGE_SIZE * 2) {
		    sprintf(errormsg, "sc_msu_blobify: image too large");
		    fprintf(stderr, "%s", errormsg);
		    return -1;
		}
		curx++;
	    }
	}
	stmp = *dwp++;
    }
    find_blobs(i, Dx_cen, Dy_cen, &num_cen, errormsg);
    if (num_cen > max_blobs) {
	/* should not happen */
	sprintf(errormsg, "sc_msu_blobify: num_cen (%d) > max blobs (%d)",
	    num_cen, max_blobs);
	fprintf(stderr, "%s", errormsg);
	return -4;
    }
    for (i=0; i<num_cen; i++) {
	/* Multiplied by 2 to get half-pixel units like CRIS blobs */
	x_cen[i] = (int)((Dx_cen[i] * 2.0) + .5);
	y_cen[i] = (int)((Dy_cen[i] * 2.0) + .5);
	if (Fx_cen != (double *)NULL) {
	    Fx_cen[i] = Dx_cen[i] * 2.0;
	}
	if (Fy_cen != (double *)NULL) {
	    Fy_cen[i] = Dy_cen[i] * 2.0;
	}
    }

    /* record the segment details about this event */
    Lastevt_nblobs = 0;
    Lastevt_totblobseg = 0;
    for (i=0; i<Maxblobs; i++) {
	if (Blob_stat[i] == BS_CLOSED) {
	    /* blob in good standing */
	    Lastevt_blobsegs[Lastevt_nblobs] = blob[i].totseg;
	    Lastevt_totblobseg += Lastevt_blobsegs[Lastevt_nblobs];
	    Lastevt_nblobs++;
	}
    }

    if (Lastevt_nblobs != num_cen) {
	/* should not happen */
	return -2;
    }

    return (num_cen);
}
    
/*	find_blobs(pix_tot, x_cen, y_cen, num_cen, errormsg)
 *	    int pix_tot;    ( total number of pixels (input))
 *	    double *x_cen;  ( x centroids (out))
 *	    double *y_cen;  ( y centroids (out))
 *	    int *num_cen;   ( number of centroids (out))
 *	    char *errormsg   ( ptr to space for error messages (out))
 *
 *	The number of blobs (or centroids) is returned via the pointer
 *	to num_cen and the actual coordinates of the centroids are returned
 *	in the arrays x_cen and y_cen.
 *
 *  The basic procedure is to collect the pixel values into contiguous
 *  segments on a row (or yline). Any segments that are connected to each other
 *  are grouped together as a blob. After all blobs have been identified,
 *  subject to certain cuts such as having a minimum amount of pixels, a
 *  centroid is calculated and the results returned.
 *
 *  Mike Westphall, Spring, 1989
 *  Marty Olevitch, Spring 1992 (much reorganization for speedup)
 *  Mike Westphall, Summer, 1992 (much reorganized to make functional)
 *  Marty Olevitch, Spring 1993 (minor changes to integrate into ace93) */
static void
find_blobs(pix_tot, x_cen, y_cen, num_cen, errormsg)
    int pix_tot;	/* total number of pixels */
    double *x_cen;	/* x centroids */
    double *y_cen;	/* y centroids */
    int *num_cen;	/* number of centroids */
    char *errormsg;
{
    int blobs_used; /* Number of blob structures that are ever occupied */
    int blob1;
    int bpix;	    /* Beginning pixel of a line */
    int epix;	    /* Ending pixel of a line */
    int i,j;
    int max;
    int nfits;	    /* no. of blobs the current line of pixels will fit into */
    int npix;	    /* Number of pixels examined */
    int npix_line;  /* Number of pixels on current line */
    int prevyline;  /* Previous camera y line */
    int retcode = 0;
    int x_max_cut;  /* limits size of pixel array viewed */
    int yline;	    /* Current camera y line */
    char strtmp[80];/* temporary string */

    x_max_cut = X_MAX_CUT;
    blobs_used = 0;
    prevyline = -1;
    npix = 0;
    Lastevt_totseg = 0;	/* total no. of segments */

    /* initialize blob structures */
    for (i = 0; i < Maxblobs; i++) {
	struct_init(&blob[i]);
	Blob_stat[i] = BS_UNUSED;
    }

    for (i=0; i<NBLOBERRS; i++) {
	Bloberr_counts[i] = 0;
    }

    while(npix <= pix_tot) {
	/* first collect pixels into a contiguous segment */
	bpix = npix;
	yline = y[bpix];			
	if(yline != prevyline) {	/* check for new y line */
	    update_blob_limits(blobs_used);
	    prevyline = yline;
	}

	/* we are working on the same segment if the x increases by 1 and
	 * the the y stays the same */
	max = pix_tot - npix;
	for(i=0; i<max; i++) {
	    if((x[bpix+i+1] - x[bpix+i] != 1) || (y[bpix+i+1] != yline))
		break;
	}
	npix_line = i + 1;
	epix = bpix + i;
	npix += npix_line;

	Lastevt_totseg++;	/* got a segment; increment */

	/* if the segment fails the cut; don't add it to any blobs */
	if((x[bpix] < X_MIN_CUT) || (x[epix] > x_max_cut) ||
		(yline < Y_MIN_CUT) || (yline > Y_MAX_CUT)) {
	    continue;
	}

	/* reset some stuff for the next part */
	nfits = 0;
	for(i=0; i<Maxblobs; i++) {
	    Can_assign[i] = 0;
	    Combine_list[i] = -1;
	}

	/* Check to see if the segment belongs to any current blobs */
	for (i=0; i<blobs_used; i++) {
	    if (BS_ACTIVE(Blob_stat[i])) {
		/* check to see if pixels belong to this blob */
		for(j=0; j<blob[i].segment[PREV]; j++){
		    if(x[bpix] <= blob[i].end[j][PREV] &&
				x[epix] >= blob[i].begin[j][PREV] &&
				yline-1 == blob[i].prev_y_line) {
			nfits++;
			Can_assign[i] = 1;
			Combine_list[nfits-1] = i;
			break;
		    }
		}
	    }
	}

	if (nfits == 0) {		/* put it in new blob */
	    int found_unused_blob = 0;
	    for (i = 0; i < Maxblobs; i++) {
		if (Blob_stat[i] == BS_UNUSED) {
		    if(Blob_debug) {
			fprintf(bdf, "---new blob %d (%d-%d) y:%d\n",
				i, x[bpix], x[epix],yline);
		    }
		    retcode = assign(&blob[i], bpix, npix_line, i);
		    if (retcode & Bloberr_masks[TOO_MANY_PIX_A]) {
			Bloberr_counts[TOO_MANY_PIX_A]++;
		    }
		    if (retcode & Bloberr_masks[SEG_OVFLOW_A]) {
			Bloberr_counts[SEG_OVFLOW_A]++;
		    }
		    /* Blob_stat could be set in assign() */
		    if (Blob_stat[i] == BS_UNUSED) {
			Blob_stat[i] = BS_FILLING;
		    }
		    if ((i + 1) > blobs_used)
			blobs_used = i + 1;
		    found_unused_blob = 1;
		    break;
		}
	    }
	    if (found_unused_blob == 0) {	/* error: num. of blobs
						 * greater than declared */
		Bloberr_counts[TOO_MANY_BLOBS]++;
	    }
	} else if(nfits == 1) {	/* assign it to an existing blob */
	    for(i=0; i<blobs_used; i++) {
		/* find which blob to assign it to */
		if(Can_assign[i] == 1) {
		    if(Blob_debug) {
			fprintf(bdf, "---old blob %d (%d-%d) y:%d\n",
				i, x[bpix],x[epix],yline);
		    }
		    retcode = assign(&blob[i], bpix, npix_line, i);
		    if (retcode & Bloberr_masks[TOO_MANY_PIX_A]) {
			Bloberr_counts[TOO_MANY_PIX_A]++;
		    }
		    if (retcode & Bloberr_masks[SEG_OVFLOW_A]) {
			Bloberr_counts[SEG_OVFLOW_A]++;
		    }
		    break;
		}
	    }
	    if(i >= blobs_used) {
		Bloberr_counts[ASSIGN_ERR]++;
	    }
	} else if(nfits >= 2) {	/* combine blobs, then assign */
	    for(i=0; i<blobs_used; i++) {
		/* get first blob that can be combined... */
		if(Combine_list[i] != -1) {
		    blob1 = Combine_list[i];
		    break;
		}
	    }
	    for(i++; i<blobs_used; i++) {
		/* ...and combine the other blobs to it... */
		if(Combine_list[i] != -1) {
		    if(Blob_debug) {
			fprintf(bdf, "---com %d -> %d\n",
				Combine_list[i], blob1);
		    }
		    retcode = combine(&blob[blob1], &blob[Combine_list[i]],
				      blob1, Combine_list[i]);
		    if (retcode & Bloberr_masks[TOO_MANY_PIX]) {
			Bloberr_counts[TOO_MANY_PIX]++;
		    }
		    if (retcode & Bloberr_masks[SEG_OVFLOW]) {
			Bloberr_counts[SEG_OVFLOW]++;
		    }
		}
	    }
	    /* ...then assign the fragment to it. */
	    if(Blob_debug) {
		fprintf(bdf, "---oldcom blob %d (%d-%d) y:%d\n",
			 blob1, x[bpix],x[epix],yline);
	    }
	    retcode = assign(&blob[blob1], bpix, npix_line, blob1);
	    if (retcode & Bloberr_masks[TOO_MANY_PIX_A]) {
		Bloberr_counts[TOO_MANY_PIX_A]++;
	    }
	    if (retcode & Bloberr_masks[SEG_OVFLOW_A]) {
		Bloberr_counts[SEG_OVFLOW_A]++;
	    }
	} else {
	    Bloberr_counts[BOGUS_NFIT]++;
	}

	/* Check for closed, too-small, too-big blobs. A blob is closed if
	 * we have moved more than 1 yline beyond it and it has neither too few
	 * or too many pixels or ylines. */
	for (i = 0; i < blobs_used; i++) {
	    if (BS_ACTIVE(Blob_stat[i])) {
		if (((yline - blob[i].y_line) > 1) || (npix >= pix_tot)) {
		    if ((blob[i].nlines < MINYLINES) ||
				(Blob_stat[i] == BS_OVFLOW)) {
			/* too many ylines or overflow */
			struct_init(&blob[i]);
			Blob_stat[i] = BS_UNUSED;
			if(Blob_debug) {
			    fprintf(bdf,
			     "--- delete blob %d (too few lines or overflow)\n",
				    i);
			}
		    } else {
			/* more than the minimum no. of ylines... */
			if (blob[i].npix < min_num_pix) {
			    /* ...but too few pixels */
			    struct_init(&blob[i]);
			    Blob_stat[i] = BS_UNUSED;
			    if(Blob_debug) {
				fprintf(bdf,
				    "--- delete blob %d (too few pixels)\n",
				    i);
			    }
			} else {
			    /* A blob in good standing! */
			    Blob_stat[i] = BS_CLOSED;
			    if(Blob_debug) {
				fprintf(bdf, "--- close blob %d\n", i);
			    }
			}
		    }
		}
	    }
	}
	if(Blob_debug) {
	    fprintf(bdf, "-------------blob summary\n");
	    for(i=0; i<blobs_used; i++) {
		fprintf(bdf, "---------------b%d (%d-%d)(%d) stat:%d\n",
			i, blob[i].begin[PREV], blob[i].end[PREV],
			blob[i].y_line, Blob_stat[i]);
	    }
	    fprintf(bdf, "================\n");
	}
    }	/* End of main 'for' loop */

    /* Final Check for closed, too-small, too-big blobs. A blob is closed if
     * if it has neither too few or too many pixels or ylines. */
    for (i = 0; i < blobs_used; i++) {
	if (BS_ACTIVE(Blob_stat[i])) {
	    if((blob[i].nlines<MINYLINES)||(Blob_stat[i]==BS_OVFLOW)){
		/* too many ylines or overflow */
		struct_init(&blob[i]);
		Blob_stat[i] = BS_UNUSED;
		if(Blob_debug) {
		    fprintf(bdf,
			    "+++ delete blob %d (too few lines or overflow)\n",
			    i);
		}
	    } else {
		/* more than the minimum no. of ylines... */
		if (blob[i].npix < min_num_pix) {
		    /* ...but too few pixels */
		    struct_init(&blob[i]);
		    Blob_stat[i] = BS_UNUSED;
		    if(Blob_debug) {
			fprintf(bdf,
				"+++ delete blob %d (too few pixels)\n", i);
		    }
		} else {
		    /* A blob in good standing! */
		    Blob_stat[i] = BS_CLOSED;
		    if(Blob_debug) {
			fprintf(bdf, "+++ close blob %d\n", i);
		    }
		}
	    }
	}
    }

    /* prepare blob info */
    *num_cen = 0;
    for (i = 0; i < blobs_used; i++) {
	if (Blob_stat[i] == BS_CLOSED) {
	    calculate(&blob[i], (*num_cen), x_cen, y_cen);
	    (*num_cen)++;
	}
    }

    /* deal with the error messages here - errormsg better be long enough */
    errormsg[0] = '\0';
    for (i=0; i<NBLOBERRS; i++) {
	if (Bloberr_counts[i] > 0) {
	    sprintf(strtmp, "%s+%d ", bloberr_phrases[i],Bloberr_counts[i]);
	    strcat(errormsg, strtmp);
	}
    }
    if (errormsg[0] != '\0') {
	sprintf(strtmp, " (%ld)", abs_evtno);
	strcat(errormsg, strtmp);
    }
    if (Blob_debug) {
	fflush(bdf);
    }
}


/* struct_init - initialize a blob structure. */
static void
struct_init(bp)
    BLOB *bp;
{
    memset((char *)bp, '\0', sizeof(BLOB));
    bp->y_line = -1;
    bp->prev_y_line = -1;
    bp->segment[PREV] = 1;
    bp->end[0][PREV] = NXCOL + 1;
    bp->totseg = 0;
}

/* assign - This subroutine takes a BLOB structure and assigns to it more
 * pixel values. It also updates other BLOB elements. */
static int
assign(bp, start, total, nblob)
    BLOB *bp;	/* pointer to blob structure */
    int start;	/* beginnig pixel number */
    int total;	/* total number of pixels to be added */
    int nblob;	/* blob number */
{
    int i;
    int adjust;
    int commence;
    int halt;
    int y_line_temp;
    char strtmp[80];
    int ret = 0;

    commence = bp->npix;
    bp->npix += total;
    adjust = start - commence;
    halt = commence + total;

    if (halt >= MAX_NUM_PIX) {	/* error check for more than max num of pix */
	Blob_stat[nblob] = BS_OVFLOW;
	ret |= Bloberr_masks[TOO_MANY_PIX_A];
	y_line_temp = y[(halt-1) + adjust];
    } else {
	for (i = commence; i < halt; i++) {	    /* add the new pixels */
	    bp->idx[i] = i + adjust;
	    y_line_temp = y[i + adjust];
	}
    }

    bp->totseg++; 
    if (bp->y_line != y_line_temp) {	/* starting new y line */
	bp->y_line = y_line_temp;
	bp->nlines++;
	bp->segment[CURR] = 1;
	bp->begin[0][CURR] = x[start];  /* set begin and end */
	bp->end[0][CURR] = x[start+total-1];
    } else {				/* continuing same y line */
	if(bp->segment[CURR] > MAX_SEG) {
	    Blob_stat[nblob] = BS_OVFLOW;
	    ret |= Bloberr_masks[SEG_OVFLOW_A];  /* was "Big Puke" error mesg */
	} else {
	    bp->segment[CURR]++;		/* define additional segment */
	    bp->begin[(bp->segment[CURR]-1)][CURR] = x[start];
	    bp->end[(bp->segment[CURR]-1)][CURR] = x[start+total-1];
	}
    }
    return ret;
}


/* combine - combines two blob structures into one. The 2nd one becomes part
 * of the first. Returns: a mask:
 *	bit 1: too many pixels in blob	    TOO_MANY_PIX
 *	bit 2: segment overflow, flushing   SEG_OVFLOW
 */
static int 
combine(b1, b2, n1, n2)
    BLOB *b1;	/* pointer to first blob structure */
    BLOB *b2;	/* pointer to second blob structure */
    int n1;	/* blob number of 1st blob */
    int n2;	/* blob number of 2nd blob */
{
    int i,j;
    int adj;
    int ret = 0;

    adj = b1->npix;
    
    b1->npix += b2->npix;
    if(b1->npix >= MAX_NUM_PIX) {
	Blob_stat[n1] = BS_OVFLOW;
	ret |= Bloberr_masks[TOO_MANY_PIX];
    } else {
	for(i=0; i < b2->npix; i++) {
	    b1->idx[adj+i] = b2->idx[i];
	}
    }

    b1->totseg += b2->totseg; 
    if(b1->y_line == b2->y_line){
	if((b1->segment[CURR] + b2->segment[CURR]) > MAX_SEG){
	    ret |= Bloberr_masks[SEG_OVFLOW];
	    Blob_stat[n1] = BS_OVFLOW;
	} else {
	    for(j=0;j<b2->segment[CURR];j++){
		b1->segment[CURR]++;
		b1->begin[(b1->segment[CURR]-1)][CURR] = b2->begin[j][CURR];
		b1->end[(b1->segment[CURR]-1)][CURR] = b2->end[j][CURR];
	    }
	}
    }

    if(b2->y_line > b1->y_line){
	b1->y_line = b2->y_line;
	b1->prev_y_line = b2->prev_y_line;
	b1->segment[CURR] = b2->segment[CURR];
	for(j=0;j<b1->segment[CURR];j++){
	    b1->begin[j][CURR] = b2->begin[j][CURR];
	    b1->end[j][CURR] = b2->end[j][CURR];
	}
    }
	
    if((b1->segment[PREV] + b2->segment[PREV]) > MAX_SEG){
	Blob_stat[n1] = BS_OVFLOW;
	ret |= Bloberr_masks[SEG_OVFLOW];
    } else {
	for(j=0;j<b2->segment[PREV];j++){
	    b1->segment[PREV]++;
	    b1->begin[(b1->segment[PREV]-1)][PREV] = b2->begin[j][PREV];
	    b1->end[(b1->segment[PREV]-1)][PREV] = b2->end[j][PREV];
	}
    }
	
    if(b1->nlines < b2->nlines)
	b1->nlines = b2->nlines;

    struct_init(b2);
    Blob_stat[n2] = BS_UNUSED;
    return ret;
}

/* calculate - determine the centroid of the blob */
static void
calculate(bp, n, x_cen, y_cen)
    BLOB *bp;
    int n;
    double *x_cen;
    double *y_cen;
{
    int i;
    long totval, ysum, xsum;
    long v;	/* temporary variable for pixel value */
    int t;	/* temp value for index into x,y,pix arrays */

    if (bp->npix >= MAX_NUM_PIX) {
	fprintf(stderr, "calculate: npix = %d  (abs:%ld)\n", 
	    bp->npix, abs_evtno);
    }

    totval = xsum = ysum = 0L;
    if (Blob_pix_per_blob != NULL) {
	*Blob_pix_per_blob = bp->npix;
	Blob_pix_per_blob++;
    }
    
    for(i=0; i<bp->npix; i++) {
	t = bp->idx[i];
	v = (long)pix[t];
	if (Blob_pix_ptr != NULL) {
	    /* save pixel values for the application if desired */
	    *Blob_pix_ptr = (u_short)(pix[t] & 0x0fff);
	    Blob_pix_ptr++;
	}
	totval += v;
	xsum += ((long)x[t] * v);
	ysum += ((long)y[t] * v);
    }
    if (Blob_npix != NULL) {
	*Blob_npix += bp->npix;
    }
    if (Blob_intensity != (long *)NULL) {
	*Blob_intensity = totval;
	Blob_intensity++;
    }
    if(totval != 0L) {
	x_cen[n] = (double)xsum / (double)totval;
	y_cen[n] = (double)ysum / (double)totval;
    } else {
	x_cen[n] = y_cen[n] = -1.0;
    }
}

/* update_blob_limits - change the values of each blob's previous begin and
 * end because we have started a new camera y line. */
static void
update_blob_limits(amt)
    int amt;	/* number of blobs used */
{
    int i,j;
    for(i=0; i<amt; i++) {
	if (BS_ACTIVE(Blob_stat[i])) {
	    for (j=0; j<blob[i].segment[CURR]; j++){
		blob[i].begin[j][PREV] = blob[i].begin[j][CURR];
		blob[i].end[j][PREV] = blob[i].end[j][CURR];
	    }
	    blob[i].segment[PREV] = blob[i].segment[CURR];
	    blob[i].prev_y_line = blob[i].y_line;
	}
    }
}

/* sc_cris_blobify - used to parse the video data from a CRIS-generated
 * soft image. There is no event data, just the video frame. This routine
 * is really just a front-end for sc_msu_blobify(). Returns the number of
 * centroids found. The video frame data is put in the cam 1 pointer no
 * matter which camera it really came from, hence the digit "1" in the call
 * to sc_msu_blobify(). */
int
sc_cris_blobify(dataptr, x_cen, y_cen, errormsg, blob_pix_ptr, blob_npix, 
	blob_int, pix_per_blob, min_pix, Fx_cen, Fy_cen)
    u_char *dataptr;	/* ptr to image data (in) */
    int *x_cen;		/* ptr to array of centroid x-coords (out) */
    int *y_cen;		/* ptr to array of centroid y-coords (out) */
    char *errormsg;	/* ptr to string that can hold a message (out) */
    u_short *blob_pix_ptr;	/* ptr to shorts that will hold pixel values 
    				 * that are in blobs. If NULL, no pixel 
				 * values are saved. */
    int *blob_npix;	/* no. of pixels in blobs found */
    long *blob_int;	/* ptr to blob intensity values (long) */
    int *pix_per_blob;	/* ptr to no. of pixels in each blob (int) */
    int min_pix;	/* min. # pixels to form a blob - use default if < 0 */
    double *Fx_cen;	/* ptr to array of centroid x-coords - floating point */
    double *Fy_cen;	/* ptr to array of centroid y-coords - floating point */
{
    struct sc_msuEvt e;
    int ret;

    e.exp_id = CRIS_BLOBIFY_ID;	/* tells sc_msu_blobify() that this is a
    				 * cris-generated image. */
    e.vidptr = dataptr;
    ret = sc_msu_blobify(&e, x_cen, y_cen, errormsg, blob_pix_ptr, blob_npix,
    		blob_int, pix_per_blob, min_pix, Fx_cen, Fy_cen, 1);
    return ret;	
}

/* sc_blob_totpix - return the total number of pixels found in the last
 * blobify event */
int
sc_blob_totpix(void)
{
    return(Totpix);
}

/* sc_blob_segments - return information about the number of blob segments
 * in the last event processed. Note that this routine returns the number
 * of segments in each valid blob. Segments in overflowed or otherwise
 * rejected blobs are not counted here. Same for total number of segments.
 * The calling function should allocate enough space for the blobsegs
 * array. It can be a maximum of Maxblobs long. */
void
sc_blob_segments(nblobs, blobsegs, totblobseg, totseg)
    int *nblobs;	/* ptr to no. blobs in last event */
    int *blobsegs;	/* array of nblobs ints w/no. of segments per blob */
    int *totblobseg;	/* total no. of segs in good blobs only in last evt */
    int *totseg;	/* total no. of segments (including not in blobs) */
{
    int i;
    *nblobs = Lastevt_nblobs;
    for (i=0; i<Lastevt_nblobs; i++) {
	*(blobsegs + i) = Lastevt_blobsegs[i];
    }
    *totblobseg = Lastevt_totblobseg;
    *totseg = Lastevt_totseg;
}

/* values for the states in sc_check_soft_video() */
#define WANT_START	0
#define WANT_Y		1
#define WANT_X		2
#define WANT_1ST_DATA	3
#define WANT_DATA	4

/* sc_check_soft_video - look for errors in the structure of the soft video
 * data. Note that some code is commented out in each of the states. This
 * is code that could be used for more detailed checks than we do at the
 * present. A return value of 0 indicates that no problems were found. A
 * negative return value indicates what kind of error was found. */
int sc_check_soft_video(ptr, maxwds)
    u_short *ptr;	/* pointer to start of soft video data */
    short maxwds;	/* max no. of words to deal with; use value of -1
			 * to get the default value */
{
    /* u_short d;	/* data value */
    u_short *p;
    short state;
    u_short wdcnt;
    /* u_short x;	/* x coord */
    /* u_short y;	/* y coord */

    if (ptr == NULL) {
	return -1;
    }

    if (maxwds == -1) {
	maxwds = (SC_SOFT_MAXSIZE / sizeof(u_short));
    }
    
    state = WANT_START;
    for (p=ptr, wdcnt=1; wdcnt <= maxwds; p++, wdcnt++) {
	switch (state) {
	    case WANT_START:
	    	if (*p != SC_START_HDR) {
		    return -2;
		}
		state = WANT_Y;
		break;
	    case WANT_Y:
		if (*p == SC_END_HDR) {
		    return 0;
		} else if (!SC_IS_Y_HDR(*p)) {
		    return -3;
		} else {
		    state = WANT_X;
		}
		break;
	    case WANT_X:
	    	if (!SC_IS_X_HDR(*p)) {
		    return -4;
		} else {
		    state = WANT_1ST_DATA;
		}
		break;
	    case WANT_1ST_DATA:
	    	/* demand some data */
	    	if (SC_IS_VIDEO_HDR(*p)) {
		    return -5;
		} else {
		    state = WANT_DATA;
		}
		break;
	    case WANT_DATA:
	    	if (SC_IS_VIDEO_HDR(*p)) {
		    if (*p == SC_END_HDR) {
			return 0;
		    } else if (SC_IS_Y_HDR(*p)) {
			state = WANT_X;
		    } else {
			/* unexpected header */
			return -6;
		    }
		} else {
		    /* data - OK */
		    ;
		}
		break;
	    default:
	    	/* state machine out of whack - should not happen */
		return -7;
		break;
	}
    }
    if (wdcnt > maxwds) {
	return -8;
    }
    return -10;	/* should not happen */
}

/* sc_too_many_blobs - see if the last soft frame blobified had too many
 * blobs. Returns the number of such errors. */
int
sc_too_many_blobs(void)
{
    return (Bloberr_counts[TOO_MANY_BLOBS]);
}

/* sc_set_maxblobs - can be used the value of Maxblobs for routines in this
 * file and in traj.c. It must be run before any calls to blobify. Return code
 * is -1 if you gave a value of n <= 0, or 0 if all went well. */
int
sc_set_maxblobs(n)
    int n;
{
    if (n <= 0) {
	return -1;
    }

    Blob_init = 0;	/* force re-allocation of all those buffers */
    traj_reset();	/* tell traj modules to reinitialize */
    Maxblobs = n;
    return 0;
}

int
sc_get_maxblobs(void)
{
    return Maxblobs;
}

/* sc_set_pix_discrim() - set the software discriminator. Returns the old
 * value of the discriminator, or -1 if there is a problem. */
int
sc_set_pix_discrim(val)
    int val;
{
    int old;
    if (val < 0) {
	return -1;
    }
    old = Pix_discrim;
    Pix_discrim = val;
    return old;
}

int
sc_get_pix_discrim(void)
{
    return Pix_discrim;
}

int
sc_msu_fix_blobs(e)
    struct sc_msuEvt *e;
{
    long blob_int[SC_MSU_MAXBLOBS];
    int blob_npix;
    char errormesg[1000];
    int i;
    int ret;
    int retval;
    int x_cen[SC_MSU_MAXBLOBS];
    int y_cen[SC_MSU_MAXBLOBS];

    retval = 0;

    /* camera 1 */
    ret = sc_msu_blobify(e, x_cen, y_cen, errormesg, NULL,
	&blob_npix, blob_int, NULL, 5, NULL, NULL, SC_CAM1);
    if (ret < 0 || ret >= SC_MSU_MAXBLOBS) {
	retval = 0x1;
    } else {
	for (i=0; i<ret; i++) {
	    /* for cam 1, eliminate any centroids in the bright spot */
	    if (263 < x_cen[i] && x_cen[i] < 279 && 
		    442 < y_cen[i] && y_cen[i] < 452) {
		continue;
	    }

	    /* replace the bad pre-calculated centroids with good ones */
	    e->blobx[i] = x_cen[i];
	    e->bloby[i] = y_cen[i];
	    e->blobI[i] = blob_int[i];
	}
	e->nblobs = ret;
    }

    /* camera 2 */
    ret = sc_msu_blobify(e, x_cen, y_cen, errormesg, NULL,
	&blob_npix, blob_int, NULL, 5, NULL, NULL, SC_CAM2);
    if (ret < 0 || ret >= SC_MSU_MAXBLOBS) {
	retval |= 0x2;
    } else {
	for (i=0; i<ret; i++) {
	    /* replace the bad pre-calculated centroids with good ones */
	    e->blobxb[i] = x_cen[i];
	    e->blobyb[i] = y_cen[i];
	    e->blobIb[i] = blob_int[i];
	}
	e->nblobsb = ret;
    }
    return retval;
}
