/* cris.c - parse a CRIS event
 *
 *	Marty Olevitch, Aug 1996
 */

static char cvsid[] = "$Id: cris.c,v 1.10 1998/10/09 15:50:58 marty Exp $";

#include "softcris.h"

static char *Cris_det_name[SC_CRIS_NDET] = {
     "E1A",  "E1B", "E2AB", "E3AB", "E4AB", "E5AB", "E6AB", "E7AB",
    "E8AB", "E9AB", "G2AB", "G3AB", "G4AB", "G5AB", "G6AB", "G7AB",
     "E1C",  "E1D", "E2CD", "E3CD", "E4CD", "E5CD", "E6CD", "E7CD",
    "E8CD", "E9CD", "G2CD", "G3CD", "G4CD", "G5CD", "G6CD", "G7CD",
};

/* The X centroid value has an offset subtracted from it, which must be
 * added back. Unfortunately, the two cameras have different values and
 * we have no way of knowing which camera's data we are receiving. Since
 * camera 2 is the better one and we will be using it more, we use the
 * camera 2 value as the default below in cris_unpack(). */
#define CAM1_XOFFSET	44
#define CAM2_XOFFSET	48

/* intensity_code - a table of values giving the intensity values to
 * our resolution. The algorithm for generating the values is at the end
 * of this file. */
static unsigned long Intensity_code[32] = {
	0L,		4L,		8L,		12L,	
	16L,		24L,		32L,		48L,
	64L,		96L,		128L,		192L,
	256L,		384L,		512L,		768L,
	1024L,		1536L,		2048L,		3072L,
	4096L,		6144L,		8192L,		12288L,
	16384L,		24576L,		32768L,		49152L,
	65536L,		98304L,		131072L,	196608L,
};

static void extract_ph(u_char *srcp, short *dp1, short *dp2);
static void extract_guard(u_char *srcp, u_short *php, int *addrp);

/* MIN_NORMAL is the minimum size in bytes of a "normal" cris normal event.
 * This consists of the 6 byte "event header" plus 5 bytes for the 2 E1
 * pulse heights. Most data should also have 2 bytes of "tag bits", but
 * some early data taken at GSI '96 did not have them. */
#define MIN_NORMAL	11

/* Allow_gsi_1_byte_evts is a kludge to deal with the GSI '96 data which
 * had CRIS normal events with an advertised length of 1 byte. The value
 * can be set/unset using the sc_cris_allow_1_byte_evts() routine. */
static int Allow_gsi_1_byte_evts = 0;

static int Cam = SC_CAM2;

/* Secs1996 - current value for seconds since 1996 - used by
 * sc_cris_unpack() to set secs1996 in sc_crisEvt struct. Set by
 * sc_cris_set_secs1996() */
static unsigned long Secs1996 = 0;

static long uncompress_npix(u_char npix);

int
sc_cris_allow_1_byte_evts(set)
    int set;
{
    int ret;
    ret = Allow_gsi_1_byte_evts;
    if (set) {
	Allow_gsi_1_byte_evts = 1;
    } else {
	Allow_gsi_1_byte_evts = 0;
    }
    return ret;
}

int
sc_cris_get_cam(void)
{
    return(Cam);
}

int
sc_cris_set_cam(cam)
    int cam;
{
    int tmp;
    tmp = Cam;
    if (cam == SC_CAM1 || cam == SC_CAM2) {
	Cam = cam;
	return tmp;
    } else {
	return -1;
    }
}

void
sc_cris_set_secs1996(unsigned long sec)
{
    Secs1996 = sec;
}

int
sc_cris_unpack_cam(p, c)
    u_char *p;
    struct sc_crisEvt *c;
{
    return(sc_cris_unpack(p, c, Cam));
}

int
sc_cris_unpack(p, c, cam)
    u_char *p;
    struct sc_crisEvt *c;
    int cam;
{
    int cnt;
    u_char *cp, *cp2;
    char *destp;
    int i;
    int idx;
    int ret;
    char *srcp;
    int tag;	/* 0 if no tags bits item */
    short val;
    int x, y;
    int xoffset;

    tag = 1;
    ret = sc_chk_cris_norm(p, c);
    if (ret < 0) {
	return -1;
    } else if (ret == 1) {
	/* assume no tag bits in this evt */
	tag = 0;
    } else if (ret == 2) {
	/* GSI '96 1-byte evt */
	return 2;
    }	
    
    if (cam == 1) {
	xoffset = CAM1_XOFFSET;
    } else {
	xoffset = CAM2_XOFFSET;
    }

    c->cam = cam;
    if (c->got_2nd_tel) {
	cp = p + 7;
    } else {
	cp = p + 6;
    }

    for (i=0; i < c->nblobs; i++) {
	cp2 = cp + 2;
	idx = (int)(*cp2 & 0x1f);
	c->hodo[i].I = Intensity_code[idx];
	y = (int)(*cp2 & 0xe0) >> 5;
	cp2 = cp + 1;
	y |= (int)((*cp2 & 0x3f) << 3);
	c->hodo[i].y = y;
	x = (int)(*cp2 & 0xc0) >> 6;
	cp2 = cp;
	x |= (int)(*cp2 << 2);
	c->hodo[i].x = x + xoffset;
	cp += 3;
    }

    if (tag) {
	/* Towards the end of the GSI '96 accelerator run, the "tag bits"
	 * data item was inserted into the data stream. There is no easy
	 * way to tell if an event has it or not. In flight, it should be
	 * there. */
	c->prim_tag = *cp << 8;
	cp++;
	c->prim_tag |= *cp;
	cp++;
    }

    if (c->R1 % 2) {
	cnt = c->R1 + 1;
    } else {
	cnt = c->R1;
    }

    for (i=0; i<SC_CRIS_NCTRPH; i++) {
	c->prim_ctr_ph[i] = -1;
	c->sec_ctr_ph[i] = -1;
    }

    for (i=0; i<SC_CRIS_NGUARD; i++) {
	c->prim_guard_ph[i] = -1;
	c->sec_guard_ph[i] = -1;
    }

    for (i=0; i<cnt; i+=2) {
	extract_ph(cp, &(c->prim_ctr_ph[i]), &(c->prim_ctr_ph[i+1]));
	cp += 3;
    }

    if (c->R1 % 2 == 0) {
	extract_ph(cp, &(c->prim_ctr_ph[i]), NULL);
	cp += 2;
    }

    if (c->got_guard) {
	u_short ph;
	int addr;
	for (i=0; i < c->G1; i++) {
	    extract_guard(cp, &ph, &addr);
	    c->prim_guard_ph[addr-2] = ph;
	    cp += 2;
	}
    }

    if (c->got_2nd_tel) {
	if (tag) {
	    c->sec_tag = *cp << 8;
	    cp++;
	    c->sec_tag |= *cp;
	    cp++;
	}

	if (c->R2 % 2) {
	    cnt = c->R2 + 1;
	} else {
	    cnt = c->R2;
	}

	for (i=0; i<cnt; i+=2) {
	    extract_ph(cp, &(c->sec_ctr_ph[i]), &(c->sec_ctr_ph[i+1]));
	    cp += 3;
	}

	if (c->R2 % 2 == 0) {
	    extract_ph(cp, &(c->sec_ctr_ph[i]), NULL);
	    cp += 2;
	}

	if (c->got_guard) {
	    u_short ph;
	    int addr;
	    for (i=0; i < c->G2; i++) {
		extract_guard(cp, &ph, &addr);
		c->sec_guard_ph[addr-2] = ph;
		cp += 2;
	    }
	}
    }

    if (c->stim) {
	c->stimdata.guard_pulsers = (*cp & 0xf0) >> 4;
	c->stimdata.ctr_pulsers = *cp & 0x0f;
	cp++;
	c->stimdata.guard_dac_level = *cp;
	cp++;
	c->stimdata.ctr_dac_level = *cp;
	cp++;
	c->stimdata.guard_hi_gain = (*cp & 0x80) >> 7;
	c->stimdata.ctr_hi_gain = (*cp & 0x40) >> 6;
	c->stimdata.soft_led_pulse = *cp & 0x3f;
    } else {
	c->stimdata.guard_pulsers = 255;
	c->stimdata.ctr_pulsers = 255;
	c->stimdata.guard_dac_level = 255;
	c->stimdata.ctr_dac_level = 255;
	c->stimdata.guard_hi_gain = 255;
	c->stimdata.ctr_hi_gain = 255;
	c->stimdata.soft_led_pulse = 255;
    }

    c->secs1996 = Secs1996;
    c->evtno = 0L;

    return 0;
}

static void
extract_ph(srcp, dp1, dp2)
    u_char *srcp;
    short *dp1;
    short *dp2;
{
    u_char *p1, *p2, *p3;
    u_short val;

    p1 = srcp;
    p2 = srcp + 1;
    p3 = srcp + 2;

    if (dp1 != NULL) {
	val = *p1;
	val <<= 4;
	val |= (*p2 >> 4);
	*dp1 = val;
    }

    if (dp2 != NULL) {
	val = *p2 & 0x0f;
	val <<= 8;
	val |= *p3;
	*dp2 = val;
    }
}

static void
extract_guard(srcp, php, addrp)
    u_char *srcp;
    u_short *php;
    int *addrp;
{
    u_char *p1, *p2;

    p1 = srcp;
    p2 = srcp + 1;

    *php = *p1;
    *php <<= 4;
    *php |= (*p2 >> 4);

    *addrp = *p2 & 0x07;
}

/* sc_cris_unpack_raw - unpack raw cris stack data into the pointed-to
 * sc_crisRawStack structure. The upper 4 bits are masked out here since
 * they are unused. It is assumed that the data is properly aligned.  */
void 
sc_cris_unpack_raw(dp, p)
    u_short *dp;		/* pointer to raw data */
    struct sc_crisRawStack *p;	/* ptr to sc_crisRawStack structure */
{
    int i; 

    for (i=0; i<4; i++) {
	p->tags[i] = dp[i];
    } 

    for (i=0; i<32; i++) {
	p->ph[i] = dp[i+4] & 0x0fff;
    }
}

/* sc_cris_unpack_stack - for those who want to deal with the raw stack
 * data, it is unpacked here into a more understandable form. Code based on
 * R. Radocinski (getGroupDiagnosticEvent_ce()) */
void
sc_cris_unpack_stack(struct sc_crisRawStack *s, struct sc_crisRawEvent *e)
{
    u_short tagwd;

    tagwd = s->tags[0];
    e->tag.stim_puls_tag = (u_char) (tagwd & 0x0001);
    tagwd >>= 1;
    e->tag.stim_off_tag = (u_char) (tagwd & 0x0001);
    tagwd >>= 1;
    e->ab.tag.z1 = (u_char) (tagwd & 0x0001);
    tagwd >>= 1;
    e->ab.tag.z2 = (u_char) (tagwd & 0x0001);
    tagwd >>= 1;
    e->ab.tag.z_gt_2 = (u_char) (tagwd & 0x0001);
    tagwd >>= 1;
    e->cd.tag.z1 = (u_char) (tagwd & 0x0001);
    tagwd >>= 1;
    e->cd.tag.z2 = (u_char) (tagwd & 0x0001);
    tagwd >>= 1;
    e->cd.tag.z_gt_2 = (u_char) (tagwd & 0x0001);
    tagwd >>= 1;
    e->tag.hazard = (u_char) (tagwd & 0x0001);
    tagwd >>= 1;
    e->tag.ghor = (u_char) (tagwd & 0x0001);
    tagwd >>= 1;
    e->tag.mor = (u_char) (tagwd & 0x0001);
    tagwd >>= 1;
    e->ab.tag.adc2 = (u_char) (tagwd & 0x0001);
    tagwd >>= 1;
    e->ab.tag.min_range = (u_char) (tagwd & 0x0001);
    tagwd >>= 1;
    e->cd.tag.adc2 = (u_char) (tagwd & 0x0001);
    tagwd >>= 1;
    e->cd.tag.min_range = (u_char) (tagwd & 0x0001);

    tagwd = s->tags[1];
    e->ab.tag.e1_0 = (u_char) (tagwd & 0x0001);
    tagwd >>= 1;
    e->ab.tag.e1_1 = (u_char) (tagwd & 0x0001);
    tagwd >>= 1;
    e->ab.tag.e2 = (u_char) (tagwd & 0x0001);
    tagwd >>= 1;
    e->ab.tag.e3 = (u_char) (tagwd & 0x0001);
    tagwd >>= 1;
    e->ab.tag.e4 = (u_char) (tagwd & 0x0001);
    tagwd >>= 1;
    e->ab.tag.e5 = (u_char) (tagwd & 0x0001);
    tagwd >>= 1;
    e->ab.tag.e6 = (u_char) (tagwd & 0x0001);
    tagwd >>= 1;
    e->ab.tag.e7 = (u_char) (tagwd & 0x0001);
    tagwd >>= 1;
    e->ab.tag.e8 = (u_char) (tagwd & 0x0001);
    tagwd >>= 1;
    e->ab.tag.e9 = (u_char) (tagwd & 0x0001);
    tagwd >>= 1;
    e->ab.tag.g2 = (u_char) (tagwd & 0x0001);
    tagwd >>= 1;
    e->ab.tag.g3 = (u_char) (tagwd & 0x0001);
    tagwd >>= 1;
    e->ab.tag.g4 = (u_char) (tagwd & 0x0001);
    tagwd >>= 1;
    e->ab.tag.g5 = (u_char) (tagwd & 0x0001);
    tagwd >>= 1;
    e->ab.tag.g6 = (u_char) (tagwd & 0x0001);
    tagwd >>= 1;
    e->ab.tag.g7 = (u_char) (tagwd & 0x0001);

    tagwd = s->tags[2];
    e->cd.tag.e1_0 = (u_char) (tagwd & 0x0001);
    tagwd >>= 1;
    e->cd.tag.e1_1 = (u_char) (tagwd & 0x0001);
    tagwd >>= 1;
    e->cd.tag.e2 = (u_char) (tagwd & 0x0001);
    tagwd >>= 1;
    e->cd.tag.e3 = (u_char) (tagwd & 0x0001);
    tagwd >>= 1;
    e->cd.tag.e4 = (u_char) (tagwd & 0x0001);
    tagwd >>= 1;
    e->cd.tag.e5 = (u_char) (tagwd & 0x0001);
    tagwd >>= 1;
    e->cd.tag.e6 = (u_char) (tagwd & 0x0001);
    tagwd >>= 1;
    e->cd.tag.e7 = (u_char) (tagwd & 0x0001);
    tagwd >>= 1;
    e->cd.tag.e8 = (u_char) (tagwd & 0x0001);
    tagwd >>= 1;
    e->cd.tag.e9 = (u_char) (tagwd & 0x0001);
    tagwd >>= 1;
    e->cd.tag.g2 = (u_char) (tagwd & 0x0001);
    tagwd >>= 1;
    e->cd.tag.g3 = (u_char) (tagwd & 0x0001);
    tagwd >>= 1;
    e->cd.tag.g4 = (u_char) (tagwd & 0x0001);
    tagwd >>= 1;
    e->cd.tag.g5 = (u_char) (tagwd & 0x0001);
    tagwd >>= 1;
    e->cd.tag.g6 = (u_char) (tagwd & 0x0001);
    tagwd >>= 1;
    e->cd.tag.g7 = (u_char) (tagwd & 0x0001);

    tagwd = s->tags[3];
    e->tag.soft_trig0 = (u_char) (tagwd & 0x0001);
    tagwd >>= 1;
    e->tag.soft_trig1 = (u_char) (tagwd & 0x0001);
    tagwd >>= 1;
    e->tag.hy_en = (u_char) (tagwd & 0x0001);
    tagwd >>= 1;
    e->tag.he_en = (u_char) (tagwd & 0x0001);
    tagwd >>= 1;
    e->CycleCount = (u_char) (tagwd & 0x00ff);

    e->ab.e1_0 = s->ph[0];
    e->ab.e1_1 = s->ph[1];
    e->ab.e2   = s->ph[2];
    e->ab.e3   = s->ph[3];
    e->ab.e4   = s->ph[4];
    e->ab.e5   = s->ph[5];
    e->ab.e6   = s->ph[6];
    e->ab.e7   = s->ph[7];
    e->ab.e8   = s->ph[8];
    e->ab.e9   = s->ph[9];
    e->ab.g2   = s->ph[10];
    e->ab.g3   = s->ph[11];
    e->ab.g4   = s->ph[12];
    e->ab.g5   = s->ph[13];
    e->ab.g6   = s->ph[14];
    e->ab.g7   = s->ph[15];

    e->cd.e1_0 = s->ph[16];
    e->cd.e1_1 = s->ph[17];
    e->cd.e2   = s->ph[18];
    e->cd.e3   = s->ph[19];
    e->cd.e4   = s->ph[20];
    e->cd.e5   = s->ph[21];
    e->cd.e6   = s->ph[22];
    e->cd.e7   = s->ph[23];
    e->cd.e8   = s->ph[24];
    e->cd.e9   = s->ph[25];
    e->cd.g2   = s->ph[26];
    e->cd.g3   = s->ph[27];
    e->cd.g4   = s->ph[28];
    e->cd.g5   = s->ph[29];
    e->cd.g6   = s->ph[30];
    e->cd.g7   = s->ph[31];
}

/* sc_cris_det_name - give its numerical ID, fill in the pointed-to string
 * with the name a cris detector (such as E1A, E2CD, etc). Returns 0 if all
 * went well, otherwise it sets the string to "(none)" and returns -1.  */
int
sc_cris_det_name(n, s)
    int n;
    char *s;
{
    if (n < 0 || n >= SC_CRIS_NDET) {
	strcpy(s, "(none)");
	return -1;
    }
    strcpy(s, Cris_det_name[n]);
    return 0;
}

int
sc_diag_unpack_cam(p, d)
    u_char *p;
    struct sc_diagEvt *d;
{
    return (sc_diag_unpack(p, d, Cam));
}

/* sc_diag_unpack - unpack a diagnostic event into the pointed-to
 * sc_diagEvt struct. Returns 0 if all went well, else -1 if there was an
 * error from sc_cris_unpack(). A diagnostic event consists of the
 * following: the raw stack data (72 bytes), the SOFT data (variable
 * length), and last, the cris normal event. We have to search through the
 * SOFT data looking for the SC_END_HDR until we get to the cris normal
 * evt. Returns -1 for sc_cris_unpack() error, -2 for no SC_START_HDR in
 * SOFT data, -3 for no SC_END_HDR, or 0 if all went well. */
int
sc_diag_unpack(p, d, cam)
    u_char *p;		/* pointer to diagnostic data */
    struct sc_diagEvt *d;
    int cam;
{
    u_char *crisp;	/* ptr to start of cris data */
    int i;
    int needswap;
    u_char stackbuf[SC_STACKSIZE];
    u_char *stackp;
    int ret;
    u_short softbuf[SC_SOFT_MAXSIZE/2];
    u_short *softstart;
    u_short *sp;
    u_short end_hdr;

    /* check the SOFT start header to see if we have data from the correct
     * -endianness. If not, reorder bytes. */
    softstart = sp = (u_short *)(p + SC_STACKSIZE);
    if (*sp == SC_START_HDR) {
	needswap = 0;
	end_hdr = SC_END_HDR;
    } else if (*sp == SC_ALT_START_HDR) {
	needswap = 1;
	end_hdr = SC_ALT_END_HDR;
    } else {
	/* bad SOFT data */
	return -2;
    }

    if (needswap) {
	swab(p, stackbuf, SC_STACKSIZE);
	stackp = stackbuf;
    } else {
	stackp = p;
    }

    sc_cris_unpack_raw((u_short *)stackp, &d->s);

    /* Determine size of the SOFT data by searching through it until we 
     * come to the end. */
    for (i=1; i<SC_SOFT_MAXSIZE/2; i++) {
	if (*sp == end_hdr) {
	    break;
	}
	sp++;
    }
    if (i >= SC_SOFT_MAXSIZE/2) {
	/* got to end of data, but no SC_END_HDR - indicates truncated data */
	return -3;
    }

    if (needswap) {
	ret = sc_reorder_soft(softstart, softbuf, i);
	if (ret) {
	    return -4;
	}
	memcpy((u_char *)d->softp, softbuf, i*2);
    } else {
	memcpy((u_char *)d->softp, softstart, i*2);
    }
    d->softwds = (u_short)i;

    sp++;
    crisp = (u_char *)sp;
    ret = sc_cris_unpack(crisp, &d->e, cam);
    if (ret) {
	return ret;
    }

    return 0;
}

/* sc_search_norm_start - find a place in the data where we think the next
 * cris event starts. Returns: -1 if we couldn't find the place in max
 * bytes; else a byte offset from the start of the data */
int
sc_search_norm_start(p, max)
    u_char *p;
    int max;
{
    u_char *cp;
    struct sc_crisEvt e;
    int i;
    int ret;

    for (i=0, cp=p; i<max; i++, cp++) {
	ret = sc_chk_cris_norm(cp, &e);
	if (ret == 0) {
	    break;
	}
    }

    if (i >= max) {
	return -1;
    }
    return i;
}

/* sc_chk_cris_norm - make sure the advertised length of the given cris 
 * normal event makes sense. Also fills in some of the sc_cris_evt
 * structure. Returns: 0 if length is as advertised, 1 if length would be
 * correct with no "tag bits" elements, or if the lenght is incorrect,
 * returns the negative of the correct value. */
int
sc_chk_cris_norm(p, c)
    u_char *p;		/* ptr to start of raw data for cris normal evt */
    struct sc_crisEvt *c;
{
    int altsum;	/* correct length for evt without tag bits */
    int sum;	/* correct lenght for evt with tag bits */
    u_char comp_npix;	/* compressed (raw) npix */

    if (p == NULL) {
	return -1;
    }
    memset(c, '\0', sizeof(struct sc_crisEvt));

    c->p = p;
    c->nbytes = p[0];
    if (c->nbytes < MIN_NORMAL) {
	if (Allow_gsi_1_byte_evts && c->nbytes == 1) {
	    /* we have a GSI '96 1-byte evt */
	    return 2;
	}
	return -1;
    }

    c->stim = SC_CRIS_IS_STIM(p);
    c->prim_tel_pair = SC_CRIS_PRIM_TEL_PAIR(p);
    c->buffer_id = SC_CRIS_EVT_BUF_ID(p);

    c->guard_hi_or = SC_CRIS_GUARD_HI_OR(p);
    c->hazard = SC_CRIS_HAZARD(p);
    c->G1 = SC_CRIS_G1(p);
    c->R1 = SC_CRIS_R1(p);

    c->got_2nd_tel = SC_CRIS_GOT_TEL2(p);
    c->got_guard = SC_CRIS_GOT_GUARD(p);
    c->centroid_err = SC_CRIS_CENTROID_ERR(p);
    c->nblobs = SC_CRIS_SOFT_NBLOBS(p);
    
    comp_npix = *(p+4);
    c->npix = uncompress_npix(comp_npix);

    c->age = *(p+5) & 0x3f;
    c->sec_tel_act = (*(p+5) & 0xc0) >> 6;

    c->G2 = SC_CRIS_G2(p);
    c->R2 = SC_CRIS_R2(p);

    /* determine what the length of the event should be */
    if (c->got_2nd_tel) {
	sum = 7;
    } else {
	sum = 6;
    }

    if (c->stim) {
	sum += 4;
    }

    sum += 2;	/* primary tag bits; altsum doesn't have this */

    if (c->got_guard) {
	sum += (c->G1 * 2);
    }

    if (c->R1 % 2 == 0) {
	/* r1 even */
	sum += (((c->R1 + 1) * 12) + 4) / 8;
    } else {
	/* r1 odd */
	sum += ((c->R1 + 1) * 12) / 8;
    }

    sum += (c->nblobs * 3);	/* SOFT blobs */

    if (c->got_2nd_tel) {
	sum += 2;	/* 2nd tag bits; altsum doesn't have this */
	if (c->got_guard) {
	    sum += (c->G2 * 2);
	}
	if (c->R2 % 2 == 0) {
	    /* r2 even */
	    sum += (((c->R2 + 1) * 12) + 4) / 8; 
	} else {
	    sum += ((c->R2 + 1) * 12) / 8; 
	}
    }

    altsum = sum;
    altsum -= 2;		/* primary tag bits */
    if (c->got_2nd_tel) {
	altsum -= 2;	/* secondary tag bits */
    }
	
    if (c->nbytes == sum) {
	/* length correct for evt with tag bits */
	return 0;
    } else if (c->nbytes == altsum) {
	/* length correct for evt without tag bits */
	return 1;
    } else {
	/* incorrect length */
	return (-sum);
    }
}

/* Note from Bob Radocynski regarding the expansion of the npix item:
 * 
 * Attached is a copy of my algorithm for decompressing the "npix" value 
 * (with downward rounding).  My algorithm returns the largest integer 
 * less than the mean of the range of the values which can be represented 
 * by the compressed value.  Your algorithm returns the smallest integer 
 * which is in the range of represented by the compressed value.
 * 
 * The rounding issue is not important.  The issue of whether you want to 
 * be near the mean of the range or the minimum of the range is more 
 * important. */

/* uncompress_npix - Bob Radocynski version which gives an expanded npix
 * closer to the mean of the range specified by the compressed value. */
static long
uncompress_npix(u_char npix)
{
   u_char exponent;
   long fraction;

   if (npix < 32) {
       return ((long)npix);
   }
   exponent = (npix >> 4) - 2;
   fraction = (long) (((npix & 0x0f) << 1) | 0x21);
   return ((fraction << exponent) - 1L);
}

#ifdef NOTDEF
/* uncompress_npix - original version */
static long
uncompress_npix(u_char npix)
{
    long mantissa;	/* "mantissa" for soft npix */
    long exponent;	/* "exponent" for soft npix */

    mantissa = (long)(comp_npix & 0x0f);
    exponent = (long)((comp_npix & 0xf0) >> 4);
    if (exponent == 0) {
	c->npix = 0;
    } else {
	c->npix = (mantissa + 16L) << (exponent - 1L);
    }
}
#endif

/* Here is the algorithm used to generate the coded ("compressed")
 * intensity values. Based on a discussion with Brownlee Gauld, 
 * 6 May 1996, describing the effect of his forth code.
 * 
 * 	(Intensity I is a value that fits in 18 bits at this point)
 * 	I = I / 4;	(get it down to 16 bits)
 * 	nshift = 0;
 * 	while (I > 4) {
 * 	    I >>= 1;
 * 	    nshift++;
 * 	}
 * 	val = (nshift * 2) + I
 * 
 * Here is the code used to generate the intensity_codes values above.
 * 
 *     int i;
 *     int I, val, prev;
 *     int nshift;
 *     int ans[32];
 * 
 *     prev = -1;
 *     for (i=0; i<65535; i++) {
 * 	I = i;
 * 	nshift = 0;
 * 	while (I > 4) {
 * 	    I >>= 1;
 * 	    nshift++;
 * 	}
 * 	val = (nshift * 2) + I;
 * 	if (val != prev) {
 * 	    prev = val;
 * 	    ans[val] = i * 4;
 * 	}
 *     }
 *     printf("long intensity_code[32] = {\n");
 *     for (i=0; i<4; i++) {
 * 	printf("\t%d,\t", ans[i]);
 *     }
 *     printf("}\n");
 *     for (i=4; i<8; i++) {
 * 	printf("\t%d,\t", ans[i]);
 *     }
 *     printf("}\n");
 *     for (i=8; i<12; i++) {
 * 	printf("\t%d,\t", ans[i]);
 *     }
 *     printf("}\n");
 *     for (i=12; i<16; i++) {
 * 	printf("\t%d,\t", ans[i]);
 *     }
 *     printf("}\n");
 *     for (i=16; i<20; i++) {
 * 	printf("\t%d,\t", ans[i]);
 *     }
 *     printf("}\n");
 *     for (i=20; i<24; i++) {
 * 	printf("\t%d,\t", ans[i]);
 *     }
 *     printf("}\n");
 *     for (i=24; i<28; i++) {
 * 	printf("\t%d,\t", ans[i]);
 *     }
 *     printf("}\n");
 *     for (i=28; i<32; i++) {
 * 	printf("\t%d,\t", ans[i]);
 *     }
 *     printf("}\n");
 */
