/* gsi_ch2.c - routines for dealing with chapter 2 data from gsi '96 runs
 *
 *	Marty Olevitch, Aug 1996
 */

static char cvsid[] = "$Id: gsi_ch2.c,v 1.3 1999/05/27 17:47:08 marty Exp $";

#include "softcris.h"
#include <errno.h>

static u_char *D;
#define DBUFSIZ	32000	/* size of data buffer */
#define RDAMT1	186	/* no of u_chars in initial read */

static u_long Bytecnt;
static int Initialized;

static int gsi_init(void);

static int
gsi_init(void)
{
    Bytecnt = 0L;
    D = (u_char *)malloc(DBUFSIZ);
    if (D == (u_char *)NULL) {
	fprintf(stderr, "gsi_init: can't malloc (%s)\n", strerror(errno));
	return SC_GSI_FAIL;
    }
    (void)sc_cris_allow_1_byte_evts(1);
    Initialized = 1;
    return SC_GSI_OK;
}

void
sc_gsi_finito()
{
    if (D != (u_char *)NULL) {
	free(D);
    }
    (void)sc_cris_allow_1_byte_evts(0);
    Initialized = 0;
}

/* sc_gsi_resync - attempt to find the next sync word in the data. Returns
 * number of bytes read. */
int
sc_gsi_resync(fp, filename)
    FILE *fp;		/* input file */
    char *filename;
{
    u_char *cp1;	/* points to start of value */
    u_char *cp2;	/* points to last (4th) byte of value */
    int nrd;
    int ret;
    int totrd = 0;
    u_long val;

    if (Initialized == 0) {
	ret = gsi_init();
	if (ret != SC_GSI_OK) {
	    return ret;
	}
    }

    cp1 = (u_char *)&val;
    cp2 = cp1 + 3;
    nrd = fread(cp1, sizeof(u_long), 1, fp);
    totrd = nrd * sizeof(u_long);
    if (nrd != 1) {
	if (ferror(fp)) {
	    fprintf(stderr, "sc_gsi_resync: read err on file '%s' (%s)\n",
		filename, strerror(errno));
	} else if (feof(fp)) {
	    fprintf(stderr, "sc_gsi_resync: EOF on file '%s'\n", filename);
	} else {
	    fprintf(stderr,
	    	"sc_gsi_resync: error on file '%s' (wanted 1, got %d)\n",
	    	filename, nrd);
	}
	return (totrd);
    }
    while (val != SC_CH2_SYNCWORD) {
	val <<= 8;
	nrd = fread(cp2, sizeof(u_char), 1, fp);
	totrd += sizeof(u_char);
	if (nrd != 1) {
	    if (ferror(fp)) {
		fprintf(stderr, "sc_gsi_resync: read err on file '%s' (%s)\n",
		    filename, strerror(errno));
	    } else if (feof(fp)) {
		fprintf(stderr, "sc_gsi_resync: EOF on file '%s'\n", filename);
	    } else {
		fprintf(stderr, "sc_gsi_resync: err on file '%s' (wanted 1, got %d)\n",
		    filename, nrd);
	    }
	    return (totrd);
	}
    }
    return (totrd);
}

int
sc_gsi_next_evt(fp, ch2, d, filename)
    FILE *fp;			/* input file */
    struct sc_ch2 *ch2;		/* chapter 2 data */
    struct sc_diagEvt *d;	/* diagnostic evt data */
    char *filename;
{
    u_char *cp;
    int bad_key;
    int bad_sync;
    int bad_user_data;
    short dsize;
    int got_leading_zeroes;
    int i;
    int j;
    int nrd;
    u_short *rawstackp;
    int ret;
    u_short *sp, *sp2;

    if (Initialized == 0) {
	ret = gsi_init();
	if (ret != SC_GSI_OK) {
	    return ret;
	}
    }

    if (D == (u_char *)NULL) {
	fprintf(stderr, "sc_gsi_next_evt: data buffer not allocated!");
	return SC_GSI_FAIL;
    }

    bad_key = bad_sync = bad_user_data = got_leading_zeroes = 0;
    
    /* read the chap 2 stuff up to the user data */
    nrd = fread(D, sizeof(u_char), RDAMT1, fp);
    if (nrd != RDAMT1) {
	if (ferror(fp)) {
	    fprintf(stderr, 
	    	"sc_gsi_next_evt: read error in file '%s' at bytecnt=%lu (%s)",
		filename, Bytecnt, strerror(errno));
	    return SC_GSI_READ_ERR;
	}
	if (feof(fp)) {
	    return SC_GSI_EOF;
	}
	fprintf(stderr, 
	    "sc_gsi_next_evt: read err on '%s' cnt=%lu (wanted %d, got %d)\n",
	    filename, Bytecnt, RDAMT1, nrd);
	return SC_GSI_FAIL;
    }

    Bytecnt += nrd * sizeof(u_char);

    cp = D;
    ch2->p = D;

    ch2->key = *(short *)cp;		cp += sizeof(short);
    if (ch2->key != 2) {
	bad_key = 1;
	fprintf(stderr,
	    "sc_gsi_next_evt: chap 2 != 2 file '%s' (bytecnt=%lu)\n", 
	    filename, Bytecnt);
    }
    ch2->spare = *(short *)cp;		cp += sizeof(short);
    ch2->tags1 = *(short *)cp;		cp += sizeof(short);
    ch2->tags0 = *(short *)cp;		cp += sizeof(short);
    ch2->ext_tags = *cp;		cp++;
    ch2->cal_dac_hi_bytes = *cp;	cp++;
    ch2->cal_dac_lo_bytes = *(short *)cp; cp += sizeof(short);
    ch2->time_seq = *(u_long *)cp;	cp += sizeof(u_long);
    ch2->time = *(u_long *)cp;		cp += sizeof(u_long);
    ch2->live_time = *(u_long *)cp;	cp += sizeof(u_long);
    for (i=0; i<3; i++) {
	for (j=0; j<24; j++) {
	    ch2->base_peak_time[j][i] = *(short *)cp;
	    cp += sizeof(short);
	}
    }
    ch2->user_data_words = *(short *)cp; cp += sizeof(short);
    if (ch2->user_data_words < 0) {
	fprintf(stderr,
	  "sc_gsi_next_evt: bad user_data_words val=%d file '%s' bytecnt=%lu\n",
	  ch2->user_data_words, filename, Bytecnt);
	return SC_GSI_FAIL;
    }
    for (i=0; i<8; i++) {
	ch2->stage[i] = *(short *)cp;
	cp += sizeof(short);
    }

    dsize = 4 + (ch2->user_data_words * sizeof(short));
    dsize -= (dsize % 4);
    dsize -= 18;    /* 16 bytes stage + 2 b. user_data_words in Verse 4 */
    nrd = fread(cp, sizeof(u_char), dsize, fp);
    if (nrd != dsize) {
	if (feof(fp)) {
	    fprintf(stderr,
	    	"sc_gsi_next_evt: unexpected EOF on file '%s' at bytecnt=%lu\n",
		filename, Bytecnt);
	    return SC_GSI_EOF;
	}
	if (ferror(fp)) {
	    fprintf(stderr, 
	  	"sc_gsi_next_evt: read err on file '%s' at bytecnt=%lu (%s)\n",
		filename, Bytecnt, strerror(errno));
	    return SC_GSI_READ_ERR;
	}
	fprintf(stderr, 
	    "sc_gsi_next_evt: read err on '%s' cnt=%lu (wanted %d, got %d)\n",
	    filename, Bytecnt, RDAMT1, nrd);
	return SC_GSI_FAIL;
    }
    ch2->userp = cp;
    if (*cp == 0x0) {
	/* check that we don't have an event with a bunch of zero 
	 * bytes inserted */
	got_leading_zeroes = 1;
	ch2->userp = sc_gsi_skip_zeroes(cp, dsize);
	if (ch2->userp != NULL) {
	    fprintf(stderr,
	    	"sc_gsi_next_evt: %d leading 0s userdata file '%s' cnt=%lu\n",
	    	ch2->userp-cp, filename, Bytecnt);
	} else {
	    fprintf(stderr, 
	    	"sc_gsi_next_evt: only got 0s in %d bytes file '%s' cnt=%lu\n",
	    	dsize, filename, Bytecnt);
	    bad_user_data = 1;
	}
    }
    Bytecnt += (nrd*sizeof(u_char));
    cp += (nrd*sizeof(u_char));

    nrd = fread(cp, sizeof(long), 2, fp);
    if (nrd != 2) {
	if (feof(fp)) {
	    fprintf(stderr,
	    	"sc_gsi_next_evt: unexpected EOF (1) file '%s' bytecnt=%lu\n",
		filename, Bytecnt);
	    return SC_GSI_EOF;
	}
	if (ferror(fp)) {
	    fprintf(stderr, 
	    	"sc_gsi_next_evt: read err (1) file '%s' bytecnt=%lu (%s)\n", 
		filename, Bytecnt, strerror(errno));
	    return SC_GSI_READ_ERR;
	}
	fprintf(stderr,
	    "sc_gsi_next_evt: read err file '%s' cnt=%lu (want %d, got %d)\n",
	    filename, Bytecnt, RDAMT1, nrd);
	return SC_GSI_FAIL;
    }

    ch2->seq_no = *(u_long *)cp;
    cp += sizeof(u_long);
    ch2->sync = *(u_long *)cp;
    cp += sizeof(u_long);
    if (ch2->sync != SC_CH2_SYNCWORD) {
	fprintf(stderr, 
	   "sc_gsi_next_evt: bad syncwd want=%x got=%x file '%s' bytecnt=%lu\n",
	   SC_CH2_SYNCWORD, ch2->sync, filename, Bytecnt);
	bad_sync = 1;
    }
    
    if (bad_key) {
	return SC_GSI_BAD_KEY;
    } else if (bad_sync) {
	return SC_GSI_BAD_SYNC;
    } else if (bad_user_data) {
	return SC_GSI_BAD_USER_DATA;
    } else if (got_leading_zeroes) {
	return SC_GSI_ZEROES;
    }

    memset(d, '\0', sizeof(struct sc_diagEvt));
    ret = sc_cris_unpack_cam(ch2->userp, &d->e);
    if (ret != 0 && ret != 2) {
	return SC_GSI_BAD_NORM;
    }

    rawstackp = (u_short *)(ch2->userp + d->e.nbytes + (d->e.nbytes % 2));
    sc_cris_unpack_raw(rawstackp, &d->s);
    sp = d->softp;
    sp2 = (u_short *)(ch2->userp + 
    	d->e.nbytes + (d->e.nbytes % 2) + SC_STACKSIZE);

    d->softwds = 0;
    for (i=0; i<SC_SOFT_MAXSIZE/2; i++) {
	*sp++ = *sp2;
	d->softwds++;
	if (*sp2 == SC_END_HDR) {
	    break;
	}
	sp2++;
    }
    if (i >= SC_SOFT_MAXSIZE/2) {
	return SC_GSI_BAD_SOFT;
    }

    return SC_GSI_OK;
}

/* sc_gsi_search_for_soft() checks byte-by-byte from the pointer startp, looking
 * for the 3-byte signature 0x10, 0x00, 0x14 that identifies the beginning
 * of the soft data. Checks up to a maximum of maxwds 2-byte words. Places
 * the value of the location found in loc or NULL if soft not found. */
u_char *
sc_gsi_search_for_soft(startp, maxwds, bytesrd)
    u_char *startp;	/* search starting location */
    int maxwds;		/* max no. of words to search */
    int *bytesrd;	/* no. of bytes read to find soft data */
{
    u_char *cp;
    int i;
    int maxbytes;

    maxbytes = 2 * maxwds;
    for (i=0; i<maxbytes; i++) {
	cp = startp + i;
	if (*cp == 0x10 && *(cp+1) == 0x00 && *(cp+2) == 0x14) {
	    *bytesrd = i;
	    return cp;
	}
    }
    *bytesrd = 0;
    return NULL;
}

/* sc_gsi_skip_zeroes() - skip up to max bytes with value 0. Returns a pointer
 * to the first non-zero bytes or NULL if only zeroes are found. */
u_char *
sc_gsi_skip_zeroes(startp, max)
    u_char *startp;	/* start location */
    int max;		/* max. no. of bytes to skip */
{
    u_char *cp;
    int i;

    for (cp=startp, i=0; i<max; cp++, i++) {
	if (*cp != 0x0) {
	    return cp;
	}
    }
    return NULL;
}

