/* level1.c - routines to interface between softlib and the level-I data
 *
 * Marty Olevitch, July 1998
 */

#ifndef NO_LEVEL1	/* don't want (have) level1 routines */

static char cvsid[] = "$Id: level1.c,v 1.11 1998/12/02 18:15:47 marty Exp $";

#include <string.h>	/* memcpy */
#include "softcris.h"

static int next_day(void);

/* Reject_bad_diag_normal_evt - if nonzero, then we reject diagnostic
 * events which have a mangled normal event part. Default is to reject such
 * diagnostic events. The value may be changed by using the routine
 * sc_reject_diags_with_bad_normal_evts(). */
static int Reject_bad_diag_normal_evt = 1;

/* MAX_SOFTDATA_EVTS - maximum number of L1CrisDiagnosticEvent structs that
 * can hold the soft video data for a single diagnostic events. For
 * example, a value of 3 means the original struct, followed by two
 * continuation structs. */
#define MAX_SOFTDATA_EVTS	3

/* Day1996 - day we are currently working on (since Jan 1, 1996) */
static int Day1996 = -1;
static int Endday;

static u_long Curdaysec;	/* sec1996 value for start of Day1996 */
static u_long Nextdaysec;	/* sec1996 value for day after Day1996 */

/* Startsec - the first event returned by sc_L1_next_cris_evt_sec() must have
 * a secs1996 value greater than this. */
static int Startsec;

/* Endsec - the last event returned by sc_L1_next_cris_evt_sec() must have
 * a secs1996 value less than this. */
static int Endsec;

/* Just_set_day - nonzero if the day has just been set via sc_L1_set_day */
static u_short Just_set_day = 0;

#define CRIS_DAY	0x0001
#define DIAG_DAY	0x0002
#define CMD_DAY		0x0004
#define HSK_DAY		0x0008
#define HIPRI_DAY	0x0010
#define LOPRI_DAY	0x0020

/* sc_L1diag_cvt - convert a level-I cris diagnostic event into softlib
 * diagnostic event format. d1 should point to an L1CrisDiagnosticEvent
 * struct, and d should point to the sc_diagEvt struct. If d1 has
 * the soft data continued in the next event, then d2 should point to that
 * next event. Otherwise d2 should be NULL. 
 *
 * Return codes: 0 if all went well, -1 for errors. The error code is set
 * as follows: 500 = NULL pointers, 501 if the cris normal event is mangled
 * in some way, and 502 if the soft data is continued into too many events.
 * nevts will point to the number of continuation events that were
 * encountered. */
int
sc_L1_diag_cvt(struct L1CrisDiagnosticEvent *d1, 
    struct sc_diagEvt *d, int *nevts)
{
    int i;
    u_char *p;
    int ret = 0;
    u_short *sp, *sp2;

    *nevts = 0;

    if (d1 == NULL || d == NULL) {
	set_err_code(500, "diag_cvt: null parameters");
	return -1;
    }

    /* unpack the cris normal evt portion of the diag event */
    sc_cris_set_secs1996(d1->Second1996MinorFrame0);
    ret = sc_cris_unpack_cam(d1->NormalEvent, &d->e);
    if (ret) {
	if (Reject_bad_diag_normal_evt) {
	    /* rejecting diags with mangled normal event */
	    set_err_code(501, "diag_cvt: bad normal event");
	    return -1;
	} else {
	    /* not rejecting diags with bad normal evts; just zero the
	     * normal event part */
	    memset((char *)&d->e, '\0', sizeof(struct sc_crisEvt));
	}
    }

    /* now the raw stack portion */
    sc_cris_unpack_raw(d1->RawStack, &d->s);

    p = (u_char *)d->softp;
    d->softwds = 0;

    for (i=0; d1->ContinuationFlag; i++) {
	/* take care of evts with the continuation flag set */
	if (i >= MAX_SOFTDATA_EVTS - 1) {
	    set_err_code(503, 
	    	"diag_cvt: soft data continued into more than %d structs", 
		MAX_SOFTDATA_EVTS);
	    return -1;
	}
	(*nevts)++;
	memcpy(p, (u_char *)(d1->SoftEvent), NUM_CRIS_RAW_SOFT_EVT_BYTES);
	p += NUM_CRIS_RAW_SOFT_EVT_BYTES;
	d->softwds += NUM_CRIS_RAW_SOFT_EVT_BYTES/2;
	d1++;
    }
    
    sp = (u_short *)p;
    sp2 = d1->SoftEvent;
    for (i=0; i < NUM_CRIS_RAW_SOFT_EVT_BYTES/2; i++) {
	*sp++ = *sp2;
	d->softwds++;
	if (*sp2 == SC_END_HDR) {
	    break;
	}
	sp2++;
    }
    if (i >= NUM_CRIS_RAW_SOFT_EVT_BYTES/2) {
	set_err_code(504, "diag_cvt: no end header");
	return -1;
    }

    return 0;
}

/* sc_L1_set_day - set the day of which we want to get data */
void
sc_L1_set_day(int startday, int endday)
{
    Day1996 = startday;
    Curdaysec = sc_day96_to_sec96(Day1996);
    Nextdaysec = sc_day96_to_sec96(Day1996+1);
    Endday = endday;
    Just_set_day = 0xffff;
}

/* next_day - increment Day1996. If we have exceeded the Endday, meaning
 * that there are no more days to process, return 0. Otherwise return 1. */
static int
next_day(void)
{
    Day1996++;
    Curdaysec = sc_day96_to_sec96(Day1996);
    Nextdaysec = sc_day96_to_sec96(Day1996+1);
    if (Day1996 > Endday) {
	/* no more days */
	return 0;
    } else {
	return 1;
    }
}

/* sc_L1_eof - return 1 if we have reached the end of data for the
 * sc_L1_next... routines. This can be used to distinguish between the end
 * of data and an error condition for a single event. */
int
sc_L1_eof(void)
{
    if (Day1996 > Endday) {
	return 1;
    } else {
	return 0;
    }
}

int
sc_L1_get_day(void)
{
    return Day1996;
}

/* sc_L1_next_cris_evt - fill in the memory pointed to by p with the next
 * sc_crisEvt in the range of days set by sc_L1_set_day(). 
 *
 * Return codes:
 *	-1	no more data left in range or error in passed params
 *	 0	next event available */
int
sc_L1_next_cris_evt(struct sc_crisEvt *p, int *day1996)
{
    static struct sc_crisEvt *Evtp;	/* events for current second */
    static uint32 *Times;		/* seconds1996 for current day */
    static int Tidx;			/* index into Times array */

    if (p == NULL || day1996 == NULL) {
	set_err_code(510, "next_cris_evt: called with NULL parameters\n");
	return -1;
    }

    if (Day1996 == -1) {
	/* day not set */
	set_err_code(511, "next_cris_evt: sc_L1_set_day() not run yet\n");
	return -1;
    }

    if (Just_set_day & CRIS_DAY) {
	/* new day was set */
	Just_set_day ^= CRIS_DAY;
	Times = getRequestTimesForDaysEvents_c(Day1996, 1);
	Tidx = 0;
	while (Times[Tidx] == 0) {
	    /* no more events on this day */
	    if (next_day()) {
		Times = getRequestTimesForDaysEvents_c(Day1996, 1);
		Tidx = 0;
	    } else {
		/* no more days left */
		set_err_code(512, "next_cris_evt: last evt\n");
		return -1;
	    }
	}
	Evtp = getEvents_c(Times[Tidx]);
    }
    while (Evtp->secs1996 == 0) {
	/* no more events in this second */
	Tidx++;
	while (Times[Tidx] == 0) {
	    /* no more events on this day */
	    if (next_day()) {
		Times = getRequestTimesForDaysEvents_c(Day1996, 1);
		Tidx = 0;
	    } else {
		/* no more days left */
		set_err_code(513, "next_cris_evt: last evt (B)\n");
		return -1;
	    }
	}
	Evtp = getEvents_c(Times[Tidx]);
    }
    memcpy(p, Evtp, sizeof(struct sc_crisEvt));
    Evtp++;
    *day1996 = Day1996;
    return 0;
}

/* sc_L1_next_diag - fill in the memory pointed to by p with the next
 * sc_diagEvt in the range of days set by sc_L1_set_day(). 
 *
 * Return codes:
 *	-1	no more data left in range or error in passed params
 *	 0	next event available */
int
sc_L1_next_diag(struct sc_diagEvt *p, int *day1996)
{
    static struct L1CrisDiagnosticEvent *Evtp;	/* level1 diags for day */
    int ret;
    int ncont;	/* no. continuation evts encountered in diag_cvt() */

    if (p == NULL || day1996 == NULL) {
	set_err_code(520, "next_diag: called with NULL parameters\n");
	return -1;
    }

    if (Day1996 == -1) {
	/* day not set */
	set_err_code(521, "next_diag: sc_L1_set_day() not run yet\n");
	return -1;
    }

    if (Just_set_day & DIAG_DAY) {
	/* new day was set */
	Just_set_day ^= DIAG_DAY;
	Evtp = getDaysDiagnosticEvents_c(Day1996, 1);
    }
    while (Evtp->Second1996MinorFrame0 == 0) {
	/* end of events in day */
	if (next_day()) {
	    Evtp = getDaysDiagnosticEvents_c(Day1996, 1);
	} else {
	    /* no more days left */
	    set_err_code(522, "next_diag: last evt\n");
	    return -1;
	}
    }

    ret = sc_L1_diag_cvt(Evtp, p, &ncont);

    Evtp += ncont;

    if (ret) {
	/* zero the diag struct on conversion error */
	memset((char *)p, '\0', sizeof(struct sc_diagEvt));
	fprintf(stderr, "%s (%d)\n", sc_get_err_mesg(), sc_get_err_code());
    }

    Evtp++;
    *day1996 = Day1996;
    return 0;
}

/* sc_L1_next_hsk_raw - fill in the memory pointed to by p with the next
 * L1CrisHskp in the range of days set by sc_L1_set_day(). Note that these
 * are the housekeeping events with the raw units, not the "enginering"
 * units.
 *
 * Return codes:
 *	-1	no more data left in range or error in passed params
 *	 0	next event available */
int
sc_L1_next_hsk_raw(struct L1CrisHskp *p, int *day1996)
{
    static struct L1CrisHskp *Evtp;	/* level1 hsks for day */
    int ret;

    if (p == NULL || day1996 == NULL) {
	set_err_code(530, "next_hsk_raw: called with NULL parameters\n");
	return -1;
    }

    if (Day1996 == -1) {
	/* day not set */
	set_err_code(531, "next_hsk_raw: sc_L1_set_day() not run yet\n");
	return -1;
    }
    if (Just_set_day & HSK_DAY) {
	/* new day was set */
	Just_set_day ^= HSK_DAY;
	Evtp = getDaysHskps_c(Day1996, 1);
    }
    while (Evtp->Second1996 == 0) {
	/* end of events in day */
	if (next_day()) {
	    Evtp = getDaysHskps_c(Day1996, 1);
	} else {
	    /* no more days left */
	    return -1;
	}
    }
    memcpy(p, Evtp, sizeof(struct L1CrisHskp));
    Evtp++;
    *day1996 = Day1996;
    return 0;
}

/* sc_L1_next_hsk - fill in the memory pointed to by p with the next
 * L1_1CrisHskp in the range of days set by sc_L1_set_day(). Note that
 * these are the housekeeping events with "engineering" units.
 *
 * Return codes:
 *	-1	no more data left in range or error in passed params
 *	 0	next event available */
int
sc_L1_next_hsk(struct L1_1CrisHskp *p, int *day1996)
{
    static struct L1_1CrisHskp *Evtp;	/* level1 hsks for day */
    int ret;

    if (p == NULL || day1996 == NULL) {
	set_err_code(570, "next_hsk: called with NULL parameters\n");
	return -1;
    }

    if (Day1996 == -1) {
	/* day not set */
	set_err_code(571, "next_hsk: sc_L1_set_day() not run yet\n");
	return -1;
    }
    if (Just_set_day & HSK_DAY) {
	/* new day was set */
	Just_set_day ^= HSK_DAY;
	Evtp = getGroupHskps_ce(Curdaysec, Nextdaysec, 1);
    }
    while (Evtp->Second1996 == 0) {
	/* end of events in day */
	if (next_day()) {
	    Evtp = getGroupHskps_ce(Curdaysec, Nextdaysec, 1);
	} else {
	    /* no more days left */
	    return -1;
	}
    }
    memcpy(p, Evtp, sizeof(struct L1_1CrisHskp));
    Evtp++;
    *day1996 = Day1996;
    return 0;
}

/* sc_L1_next_cmd - fill in the memory pointed to by p with the next
 * L1CrisCommandEcho in the range of days set by sc_L1_set_day().
 *
 * Return codes:
 *	-1	no more data left in range or error in passed params
 *	 0	next event available */
int
sc_L1_next_cmd(struct L1CrisCommandEcho *p, int *day1996)
{
    static struct L1CrisCommandEcho *Evtp;	/* level1 cmd echoes for day */
    int ret;

    if (p == NULL || day1996 == NULL) {
	set_err_code(540, "next_cmd: called with NULL parameters\n");
	return -1;
    }

    if (Day1996 == -1) {
	/* day not set */
	set_err_code(541, "next_cmd: sc_L1_set_day() not run yet\n");
	return -1;
    }

    if (Just_set_day & CMD_DAY) {
	/* new day was set */
	Just_set_day ^= CMD_DAY;
	Evtp = getDaysCmdEchoes_c(Day1996, 1);
    }

    while (Evtp->Second1996 == 0) {
	/* end of events in day */
	if (next_day()) {
	    Evtp = getDaysCmdEchoes_c(Day1996, 1);
	} else {
	    /* no more days left */
	    return -1;
	}
    }

    memcpy(p, Evtp, sizeof(struct L1CrisCommandEcho));
    Evtp++;
    *day1996 = Day1996;
    return 0;
}

/* sc_L1_next_hipri_rate - fill in the memory pointed to by p with the next
 * L1CrisHighPriorityRate in the range of days set by sc_L1_set_day().
 *
 * Return codes:
 *	-1	no more data left in range or error in passed params
 *	 0	next event available */
int 
sc_L1_next_hipri_rate(struct L1CrisHighPriorityRate *p, int *day1996)
{
    static struct L1CrisHighPriorityRate *Evtp;	/* level1 day's hi pri rates */
    int ret;

    if (p == NULL || day1996 == NULL) {
	set_err_code(550, "next_hipri: called with NULL parameters\n");
	return -1;
    }

    if (Day1996 == -1) {
	/* day not set */
	set_err_code(551, "next_hipri: sc_L1_set_day() not run yet\n");
	return -1;
    }

    if (Just_set_day & HIPRI_DAY) {
	/* new day was set */
	Just_set_day ^= HIPRI_DAY;
	Evtp = getDaysHiRates_c(Day1996, 1);
    }
    while (Evtp->Second1996 == 0) {
	/* end of events in day */
	if (next_day()) {
	    Evtp = getDaysHiRates_c(Day1996, 1);
	} else {
	    /* no more days left */
	    return -1;
	}
    }
    memcpy(p, Evtp, sizeof(struct L1CrisHighPriorityRate));
    Evtp++;
    *day1996 = Day1996;
    return 0;
}

/* sc_L1_next_lopri_rate_raw - fill in the memory pointed to by p with the
 * next L1CrisLowPriorityRate in the range of days set by sc_L1_set_day().
 * Note that this routine uses raw units, not "engineering" units.
 *
 * Return codes:
 *	-1	no more data left in range or error in passed params
 *	 0	next event available */
int 
sc_L1_next_lopri_rate_raw(struct L1CrisLowPriorityRate *p, int *day1996)
{
    static struct L1CrisLowPriorityRate *Evtp;	/* level1 day's lo pri rates */
    int ret;

    if (p == NULL || day1996 == NULL) {
	set_err_code(560, "next_lopri_raw: called with NULL parameters\n");
	return -1;
    }

    if (Day1996 == -1) {
	/* day not set */
	set_err_code(561, "next_lopri_raw: sc_L1_set_day() not run yet\n");
	return -1;
    }

    if (Just_set_day & LOPRI_DAY) {
	/* new day was set */
	Just_set_day ^= LOPRI_DAY;
	Evtp = getDaysLoRates_c(Day1996, 1);
    }
    while (Evtp->Second1996 == 0) {
	/* end of events in day */
	if (next_day()) {
	    Evtp = getDaysLoRates_c(Day1996, 1);
	} else {
	    /* no more days left */
	    return -1;
	}
    }
    memcpy(p, Evtp, sizeof(struct L1CrisLowPriorityRate));
    Evtp++;
    *day1996 = Day1996;
    return 0;
}

/* sc_L1_next_lopri_rate - fill in the memory pointed to by p with the next
 * L1_1CrisLowPriorityRate in the range of days set by sc_L1_set_day().
 * Note that this routine uses "engineering" units.
 *
 * Return codes:
 *	-1	no more data left in range or error in passed params
 *	 0	next event available */
int 
sc_L1_next_lopri_rate(struct L1_1CrisLowPriorityRate *p, int *day1996)
{
    static struct L1_1CrisLowPriorityRate *Evtp;
    int ret;

    if (p == NULL || day1996 == NULL) {
	set_err_code(580, "next_lopri: called with NULL parameters\n");
	return -1;
    }

    if (Day1996 == -1) {
	/* day not set */
	set_err_code(581, "next_lopri: sc_L1_set_day() not run yet\n");
	return -1;
    }

    if (Just_set_day & LOPRI_DAY) {
	/* new day was set */
	Just_set_day ^= LOPRI_DAY;
	Evtp = getGroupLoRates_ce(Curdaysec, Nextdaysec, 1);
    }
    while (Evtp->Second1996 == 0) {
	/* end of events in day */
	if (next_day()) {
	    Evtp = getGroupLoRates_ce(Curdaysec, Nextdaysec, 1);
	} else {
	    /* no more days left */
	    return -1;
	}
    }
    memcpy(p, Evtp, sizeof(struct L1_1CrisLowPriorityRate));
    Evtp++;
    *day1996 = Day1996;
    return 0;
}

/* sc_L1_set_start_end_sec() - set the values of Startsec and Endsec used
 * by the sc_L1_next_cris_evt_sec() and related routines. */
void
sc_L1_set_start_end_sec(int startsec, int endsec)
{
    Startsec = startsec;
    Endsec = endsec;
}

/* sc_L1_next_cris_evt_sec() - return the next event in the range of days
 * (set by sc_L1_set_day()) such that Startsec <= c->secs1996 <= Endsec.
 * Startsec and Endsec are set by using the the sc_L1_set_start_end_sec()
 * function. Return codes as for sc_L1_next_cris_evt(). */
int
sc_L1_next_cris_evt_sec(struct sc_crisEvt *p, int *day1996)
{
    int ret;

    while (1) {
	ret = sc_L1_next_cris_evt(p, day1996); 
	if (ret != 0) {
	    return ret;
	}
	if (Startsec <= p->secs1996 && p->secs1996 <= Endsec) {
	    return 0;
	}
	if (p->secs1996 > Endsec) {
	    return -1;
	}
    }
}

/* sc_L1_next_diag_sec() - return the next diagnostic event in the range of
 * days (set by sc_L1_set_day()) such that Startsec <= c->e.secs1996 <=
 * Endsec. Startsec and Endsec are set by using the the
 * sc_L1_set_start_end_sec() function. Return codes same as for
 * sc_L1_next_diag(). */
int
sc_L1_next_diag_sec(struct sc_diagEvt *p, int *day1996)
{
    int ret;

    while (1) {
	ret = sc_L1_next_diag(p, day1996);
	if (ret != 0) {
	    return ret;
	}
	if (Startsec <= p->e.secs1996 && p->e.secs1996 <= Endsec) {
	    return 0;
	}
	if (p->e.secs1996 > Endsec) {
	    return -1;
	}
    }
}

/* sc_L1_next_hsk_raw_sec() - return the next L1CrisHskp in the range of days
 * (set by sc_L1_set_day()) such that Startsec <= c->Second1996 <= Endsec.
 * Startsec and Endsec are set by using the the sc_L1_set_start_end_sec()
 * function. Return codes same as for sc_L1_next_hsk(). */
int
sc_L1_next_hsk_raw_sec(struct L1CrisHskp *p, int *day1996)
{
    int ret;

    while (1) {
	ret = sc_L1_next_hsk_raw(p, day1996);
	if (ret != 0) {
	    return ret;
	}
	if (Startsec <= p->Second1996 && p->Second1996 <= Endsec) {
	    return 0;
	}
	if (p->Second1996 > Endsec) {
	    return -1;
	}
    }
}

/* sc_L1_next_hsk_sec() - return the next L1_1CrisHskp in the range of days
 * (set by sc_L1_set_day()) such that Startsec <= c->Second1996 <= Endsec.
 * Startsec and Endsec are set by using the the sc_L1_set_start_end_sec()
 * function. Return codes same as for sc_L1_next_hsk(). */
int
sc_L1_next_hsk_sec(struct L1_1CrisHskp *p, int *day1996)
{
    int ret;

    while (1) {
	ret = sc_L1_next_hsk(p, day1996);
	if (ret != 0) {
	    return ret;
	}
	if (Startsec <= p->Second1996 && p->Second1996 <= Endsec) {
	    return 0;
	}
	if (p->Second1996 > Endsec) {
	    return -1;
	}
    }
}

/* sc_L1_next_cmd_sec() - return the next L1CrisCommandEcho in
 * the range of days (set by sc_L1_set_day()) such that Startsec <=
 * c->Second1996 <= Endsec. Startsec and Endsec are set by using the the
 * sc_L1_set_start_end_sec() function. Return codes same as for
 * sc_L1_next_cmd(). */
int
sc_L1_next_cmd_sec(struct L1CrisCommandEcho *p, int *day1996)
{
    int ret;

    while (1) {
	ret = sc_L1_next_cmd(p, day1996);
	if (ret != 0) {
	    return ret;
	}
	if (Startsec <= p->Second1996 && p->Second1996 <= Endsec) {
	    return 0;
	}
	if (p->Second1996 > Endsec) {
	    return -1;
	}
    }
}

/* sc_L1_next_hipri_rate_sec() - return the next L1CrisHighPriorityRate in
 * the range of days (set by sc_L1_set_day()) such that Startsec <=
 * c->Second1996 <= Endsec. Startsec and Endsec are set by using the the
 * sc_L1_set_start_end_sec() function. Return codes same as for
 * sc_l1_next_hipri_rate(). */
int 
sc_L1_next_hipri_rate_sec(struct L1CrisHighPriorityRate *p, int *day1996)
{
    int ret;

    while (1) {
	ret = sc_L1_next_hipri_rate(p, day1996);
	if (ret != 0) {
	    return ret;
	}
	if (Startsec <= p->Second1996 && p->Second1996 <= Endsec) {
	    return 0;
	}
	if (p->Second1996 > Endsec) {
	    return -1;
	}
    }
}

/* sc_L1_next_lopri_rate_raw_sec() - return the next L1CrisLowPriorityRate
 * in the range of days (set by sc_L1_set_day()) such that Startsec <=
 * c->Second1996 <= Endsec. Startsec and Endsec are set by using the the
 * sc_L1_set_start_end_sec() function. Return codes same as for
 * sc_l1_next_lopri_rate(). */
int
sc_L1_next_lopri_rate_raw_sec(struct L1CrisLowPriorityRate *p, int *day1996)
{
    int ret;

    while (1) {
	ret = sc_L1_next_lopri_rate_raw(p, day1996);
	if (ret != 0) {
	    return ret;
	}
	if (Startsec <= p->Second1996 && p->Second1996 <= Endsec) {
	    return 0;
	}
	if (p->Second1996 > Endsec) {
	    return -1;
	}
    }
}

/* sc_L1_next_lopri_rate_sec() - return the next L1_1CrisLowPriorityRate in
 * the range of days (set by sc_L1_set_day()) such that Startsec <=
 * c->Second1996 <= Endsec. Startsec and Endsec are set by using the the
 * sc_L1_set_start_end_sec() function. Return codes same as for
 * sc_l1_next_lopri_rate(). */
int
sc_L1_next_lopri_rate_sec(struct L1_1CrisLowPriorityRate *p, int *day1996)
{
    int ret;

    while (1) {
	ret = sc_L1_next_lopri_rate(p, day1996);
	if (ret != 0) {
	    return ret;
	}
	if (Startsec <= p->Second1996 && p->Second1996 <= Endsec) {
	    return 0;
	}
	if (p->Second1996 > Endsec) {
	    return -1;
	}
    }
}

/* sc_day_of_year_96 - given the year and day-of-year, return the number of
 * days from Jan. 1, 1996 (inclusive). The year may be in the form 1998,
 * 1999, 2000, 2001, ... or in the form 98, 99, 0, 1 ... Years less than
 * 1996 or greater than 2050 are rejected and -1 is returned. */
int
sc_day_of_year_96(int yr, int doy)
{
    int d;
    int nleap;	/* no. of leap days since 1996 */
    int nyr;

    if (yr < 50) {
	/* assume it is 00, 01, 02... for years >= 2000 */
	yr += 2000;
    } else if (yr < 100) {
	/* 96, 97, 98 ... */
	yr += 1900;
    } else if (yr < 1996 | yr > 2050) {
	/* don't want to handle these */
	return -1;
    }

    if (yr == 1996) {
	return doy;
    }

    nleap = ((yr - 1997) / 4) + 1;
    nyr = yr - 1996;
    d = (365 * nyr) + nleap + doy;
    return d;
}

/* sc_day96_to_sec96 - given a day-of-year-since-1996 return the number of
 * seconds since 1996. The first day is number 1, but the first second is
 * number 0. When more leap seconds are declared, they must be added
 * after the current ones. */
u_long
sc_day96_to_sec96(int day)
{
    u_long sec;
 
    sec = (day - 1) * 86400L;
 
    if (day >= 548) {   /* July 1997 leap second */
        sec++;
    }
 
    return sec;
}

void 
sc_reject_diags_with_bad_normal_evts(int doit)
{
    if (doit) {
	Reject_bad_diag_normal_evt = 1;
    } else {
	Reject_bad_diag_normal_evt = 0;
    }
}

#endif /* ! NO_LEVEL1 */
