/* lookup - use the lookup table to convert from pixel to fiber coords.
 *
 *  For SOFT data
 *
 * Some of the code here is dependent on the particular set of
 * fibers. Search for the string "FIBER DEPENDENT" to find such places.
 *
 *  Marty Olevitch, May 1994
 *	(based on Dave Crary's work for previous MSU runs.
 *	See his program get_ctabs3.c in one of his cosray directories.)
 */

static char cvsid[] = "$Id: lookup.c,v 1.2 1999/02/25 19:41:37 marty Exp $";

/* Format of the lookup table (LUT):
 *  - lines beginning with a '#' character are ignored (comments)
 *  - the data for a particular layer must begin with a "layer" statement:
 *	layer LL [keyword ....]
 *	examples:
 *	    layer X3
 *	    layer Y2 top-to-bottom
 *	    layer X1 right-to-left top-to-bottom cutoff 51.2
 *	(only the "layer" and LL keywords are required. LL gives the "name"
 *	of the layer. The possible keywords are:
 *		top-to-bottom (default)
 *		bottom-to-top
 *		left-to-right (default)
 *		right-to-left
 *		cutoff xxx.xx
 *	the cutoff is determined from the data itself if no "cutoff" keyword
 *	is specified.)
 *  - the layer statement is followed by data lines, which
 *    consist of 2 floating point values separated by white
 *    space: an x value and a y value which represent reference points
 *    along a tab. There should be SC_NLUTPTS reference points per tab.
 *    We have the reference points for one tab, sorted in order of their
 *    x-coordinate, followed by another tab, etc.
 */

#define LOOKUP_VERSION	"1.15"

#ifdef NOTDEF
/*
    Revision history

    (Now merged into softcrislib - see the Changes file for any new changes)

    1.15    24 Jan 1996	    able to deal with multiple LUTs simultaneously
    1.14    23 Jan 1996	    chk_pointy() routine will also check y value
    1.13    22 Jan 1996	    lu_pos() did not check return value from
    			    chk_point().
    1.12    14 Mar 1995	    added scale factors to DIST distance calculation
			     so that result in is in units of microns rather
			     than x- or y-pixels
    1.11     6 Mar 1995	    more sophisticated position determination
    1.10     1 Mar 1995	    added lu_pos_extra() routine. Changed lu_pos() to
			     use new find_pos() routine.
    1.9	    13 Dec 1994	    cutoff now has a value for each ref pt on tab -
			     needed because layers are much closer together
			     for these fibers; bug in lu_pos(): needed <= in
			     "pixels go from right to left" code.
    1.8	    14 Oct 1994	    changes for MSU Oct 94 accelerator run: NTABS is
			     now 66; read "Reflimits" from lut file.
    1.7	    5 Jul 1994	    lu_pix_to_real_pos() function.
    1.6	    1 Jul 1994	    lu_write_binary_table() puts in a version no. at
			     the beginning of the table.
			     lu_read_binary_table() checks this and issues a
			     warning, but continues, processing, if it
			     doesn't match.
    1.5	    1 Jul 1994	    prefix public functions with "lu_". Changed some
			     function names. More "device independence".
    1.4	    30 Jun 1994	    read cutoffs from "layer" stmt or derive from
			     data; other stuff to make calculating
			     quantities more "device independent".
    1.3	    27 Jun 1994	    specify layer info in lut file
    1.2	    23 Jun 1994	    Binary lut file stuff
    1.1	    23 Jun 1994	    Speed-ups in get_tab() and getpos()
    1.0	    May 1994	    Initial working version
*/
#endif

#include <stdio.h>
#include <math.h>
#include <assert.h>
#include <string.h>
#include <errno.h>

#include "misc.h"
#include "msu.h"
#include "softcris.h"

#define XSCALE	16.0	/* scale factor to convert x-pixels to microns */
#define YSCALE	27.0	/* scale factor to convert y-pixels to microns */

#ifndef SQR
#define SQR(x)	((x)*(x))
#endif

#define DIST(x1, y1, x2, y2)	(sqrt(SQR((XSCALE*(x2))-(XSCALE*(x1))) + \
				      SQR((YSCALE*(y2))-(YSCALE*(y1)))))
				      
#define LINLEN	80  /* max. line length in lut file */

#define TOPBOT	1	/* tabs top-to-bottom */
#define BOTTOP	-1	/* tabs bottom-to-top */
#define LEFTRT	1	/* tabs left-to-right */
#define RTLEFT	-1	/* tabs right-to-left */

#define CUTDEF	-1.0    /* initial (unset) cutoff value */
#define NO_CUTOFF(u,l,r) (Lyr[u][l].cutoff[r] == CUTDEF) /* is cutoff unset? */

#define LNL 10	    /* Layer Name Length */

/* layerinfo - details about each of layer of the detector. The topbot and
 * leftrt stuff is for the MSU Dec 1993 run. Hopefully, future detectors
 * won't have those flaws (Alas, this didn't occur. 24 Jan 96). The
 * defaults are changed by reading in a "layer" statement (for cutoff,
 * topbot, or leftrt) or data (for cutoff, start, or end) from the LUT
 * file. (cutoff in the layer statement overrides the data).  The defaults
 * are set at the beginning of lu_read_table(). */

static struct layerinfo {
    double cutoff[SC_NLUTPTS];  /* approx end of tab in Y in pixel space (see
			     * lu_tab()) */
    int start;		    /* starting tab */
    int end;		    /* ending tab */
    int topbot;		    /* 1 if tabs go top to bottom, else -1 */
    int leftrt;		    /* 1 if pixels go left to right, else -1 */
    char name[LNL];	    /* "Y3", "X2", etc */
} Lyr[SC_NLUTS][SC_NLAYERS];

#define TOOFAR	1.0 /* if a point is more than this many units from a y
		     * intercept, it is considered to not be on the line.
		     * (FIBER DEPENDENT possibly) */

#define BIGNUM	9999.0	    /* big number */
#define TOKSEP	" \t"	    /* token separators for "layer" lines */

/* INTERP - given the tab t, reference point r, and x value, do the
 * interpolation to find a y value. Used in lu_tab(). */
#define INTERP(u, t, r, x) \
    (((Lut[u].ylut[(t)][(r)+1] - Lut[u].ylut[(t)][(r)]) /   \
      (Lut[u].xlut[(t)][(r)+1] - Lut[u].xlut[(t)][(r)]) *   \
      ((x) - Lut[u].xlut[(t)][(r)])) +               \
     (Lut[u].ylut[(t)][(r)]))

/* lutinfo - information about each lookup table (LUT) 
 * Reflimit is used to determine between which 2 reference pts a point
 * lies. Can be read from lut file with the reflimit keyword or from the
 * default values. 
 * sectionlength array: length of each section betw ref pts in each
 * tab; sectionlength[0] holds the distance from refpt 0 to refpt 1;
 * sectionlength[1] the dist from refpt 1 to refpt 2, etc. Note that
 * the last element of the array isn't used, but it is useful to have
 * it there for a couple of the loops in init_reflen(). */
static struct lutinfo {
    double xlut[SC_NTABS][SC_NLUTPTS];	/* X-coords of reference points */
    double ylut[SC_NTABS][SC_NLUTPTS];	/* Y-coords of reference points */
    double tablength[SC_NTABS];		/* length of each tab */
    double ylo, yhi;			/* lowest and highest y values */
    double reflimit[SC_NLUTPTS-2];		/* see comment above */
    double sectionlength[SC_NTABS][SC_NLUTPTS];  /* see comment above */
    double reflen[SC_NTABS][SC_NLUTPTS];	/* total length in layer to this
					 * reference point */
    int ntabs;   			/* actual no. of tabs found */
} Lut[SC_NLUTS];

static int Nluts = 0;	/* actual no. of lookup tables in use */

static double get_length(int lut, int tab);
static void init_reflen(int lut);

#ifdef NOTDEF /* not used at present */
static double find_maxy(int tab);
static double find_miny(int tab);
#endif

static double find_pos(int lut, int *xrefval, int layer, int tab,
		    double x, double y);
static int find_prev_next_vals(int lut, int *ptab, int *player, double *plen,
	            double *px, double *py,
		    int *ntab, int *nlayer, double *nlen,
	            double *nx, double *ny,
		    int tab, int layer, int xref, double x);
static double xdist(int lut, double *y, double x, int tab, 
		    int xref1, int xref2);
static int chk_point(int lut, int layer, int tab, double x);
static int chk_pointy(int lut, int layer, int tab, double x, double y);


/* sc_lu_read_table - read in a lookup table into the xlut and ylut arrays.
 * Any line beginning with a '#' is a comment line. Non-comment lines
 * consist of an x- and y-coordinate of a reference point for a tab.
 * Returns a non-negative LUT id if there are no problems, else -1. */
int
sc_lu_read_table(char *lutfile)
{
    char *cp;
    FILE *fp;
    char *field;		/* field ptr from strtok() */
    int firstline;		/* 1 if first data line in a layer */
    char *fret;
    int gotlayer = 0;		/* bitmask telling which layers we've got */
    int i, j;
    char layername[LINLEN];	/* name of layer read from "layer" stmt */
    int layerval;		/* SC_Y3, SC_X3, etc */
    char line[LINLEN];
    int curlut;			/* current lut id */
    double oldy;
    int tab;
    double x0, y0;

    fp = fopen(lutfile, "r");
    if (fp == (FILE *)NULL) {
	fprintf(stderr, "sc_lu_read_table: can't open file '%s' (%s)\n",
		lutfile, strerror(errno));
	return -1;
    }
    
    if (Nluts >= SC_NLUTS) {
	fprintf(stderr, "sc_lu_read_table: too many LUTs in use. Sorry\n");
	return -1;
    }
    curlut = Nluts;
    Nluts++;

    /* initialize Lyr structures */
    for (i=0; i<SC_NLAYERS; i++) {
	for (j=0; j<SC_NLUTPTS; j++) {
	    Lyr[curlut][i].cutoff[j] = CUTDEF;
	}
	Lyr[curlut][i].start = 0;
	Lyr[curlut][i].end = 0;
	Lyr[curlut][i].topbot = TOPBOT;
	Lyr[curlut][i].leftrt = LEFTRT;
	strcpy(Lyr[curlut][i].name, "??");
    }

    Lut[curlut].ylo = BIGNUM;
    Lut[curlut].yhi = 0;
    Lut[curlut].reflimit[0] = 100.0;
    Lut[curlut].reflimit[1] = 200.0;
    Lut[curlut].reflimit[2] = 300.0;
    Lut[curlut].reflimit[3] = 400.0;

    i = 0;
    tab = 0;
    oldy = 0.0;
    firstline = 0;
    strcpy(layername, "null");
    layerval = SC_NOLAYER;
    while (fgets(line, LINLEN, fp) != NULL) {
	CHOP(line, '\0');
	/* strip leading white space */
	for (cp=line; *cp == ' ' || *cp == '\t'; cp++)
	    ;
	if (*cp == '#') {
	    continue;	/* comment line */
	}
	if (SOFT_STREQN(cp, "reflimit", strlen("reflimit"))) {
	    int k;
	    field = strtok(cp, TOKSEP);	    /* "reflimit" */
	    for (k=0; k<SC_NLUTPTS-2; k++) {
		field = strtok(NULL, TOKSEP);
		if (field != NULL) {
		    Lut[curlut].reflimit[k] = atof(field);
		}
	    }
	    continue;
	}
	if (SOFT_STREQN(cp, "layer", strlen("layer"))) {
	    if (layerval == SC_NOLAYER) {
		layerval = 0;
	    } else {
		layerval++;
	    }
	    field = strtok(cp, TOKSEP);
	    /* this is a "layer" statement in the LUT file */
	    field = strtok(NULL, TOKSEP);
	    strcpy(layername, field);
	    strncpy(Lyr[curlut][layerval].name, layername, LNL);
	    if (gotlayer & (1 << layerval)) {
		fprintf(stderr,"sc_lu_read_table: duplicate layer! "
				"(no. %d, \"%s\")\n", layerval, layername);
		return -1;
	    }
	    gotlayer |= (1 << layerval);
	    while ((field = strtok(NULL, TOKSEP)) != NULL) {
		if (SOFT_STREQ(field, "top-to-bottom")) {
		    Lyr[curlut][layerval].topbot = TOPBOT;
		} else if (SOFT_STREQ(field, "bottom-to-top")) {
		    Lyr[curlut][layerval].topbot = BOTTOP;
		} else if (SOFT_STREQ(field, "left-to-right")) {
		    Lyr[curlut][layerval].leftrt = LEFTRT;
		} else if (SOFT_STREQ(field, "right-to-left")) {
		    Lyr[curlut][layerval].leftrt = RTLEFT;
		} else if (SOFT_STREQ(field, "cutoff")) {
		    field = strtok(NULL, TOKSEP);
		    if (field == NULL) {
			fprintf(stderr,
				"sc_lu_read_table: no cutoff val (layer %s)\n",
				layername);
			return -1;
		    } else {
			double f;
			f = atof(field);
			for (j=0; j<SC_NLUTPTS; j++) {
			    Lyr[curlut][layerval].cutoff[j]= f;
			}
		    }
		} else {
		    fprintf(stderr,"sc_lu_read_table: bad field \"%s\"\n",
			    field);
		    return -1;
		}
	    }
	    Lyr[curlut][layerval].start = tab;
	    continue;
	}
	if (layerval == SC_NOLAYER) {
	    /* didn't get a "layer" statement in the LUT file */
	    fprintf(stderr,
		    "sc_lu_read_table: no \"layer\" statement in '%s'."
		    " Suspect obsolete LUT file?\n", lutfile);
	    return -1;
	}
	/* this must be a data line */
	sscanf(cp, "%lf%lf", &x0, &y0);
	Lut[curlut].xlut[tab][i] = x0;
	Lut[curlut].ylut[tab][i] = y0;
	if (y0 < Lut[curlut].ylo) {
	    Lut[curlut].ylo = y0;
	}
	if (y0 > Lut[curlut].yhi) {
	    Lut[curlut].yhi = y0;
	}
	if (i == 0) {
	    /* check to make sure this tab starts off at a higher
	     * value than the previous tab */
	    if (y0 < oldy) {
		fprintf(stderr, "sc_lu_read_table: bad lut file '%s'!\n", lutfile);
		fprintf(stderr, " (points not in order near x=%f, y=%f)\n",
			x0, y0);
		fclose(fp);
		Nluts--;
		return -1;
	    }
	    oldy = y0;
	}
	Lyr[curlut][layerval].end = tab;    /* keep updating end tab */
	i++;
	if (i >= SC_NLUTPTS) {
	    i = 0;
	    tab++;
	}
    }
    fclose(fp);
    Lut[curlut].ntabs = tab;

    /* See if we need to set the cutoff for any tabs. We need to do this if
     * the cutoff wasn't explicity set by the "cutoff" keyword in a "layer"
     * statement in the LUT file. */
    for (i=0; i<SC_NLAYERS-1; i++) {
	double this, next;
	for (j=0; j<SC_NLUTPTS; j++) {
	    if (NO_CUTOFF(curlut, i, j)) {
		/* average the y of the tab at this reference point in this
		 * layer with the y at this reference point of the first tab
		 * of the next layer. */
		this = Lut[curlut].ylut[Lyr[curlut][i].end][j];
		next = Lut[curlut].ylut[Lyr[curlut][i+1].start][j];
		Lyr[curlut][i].cutoff[j] = (this + next) / 2.0;
	    }
	}
    }
    /* Now do the last layer. */
    for (j=0; j<SC_NLUTPTS; j++) {
	if (NO_CUTOFF(curlut, i, j)) {
	    /* add a bit to y of last tab in this layer (no next layer) */
	    Lyr[curlut][i].cutoff[j] = 
	    	Lut[curlut].ylut[Lyr[curlut][i].end][j] + (TOOFAR * 3.0);
	}
    }
    
    /* add a little to the ylo and yhi values */
    Lut[curlut].ylo -= (TOOFAR * 3.0);
    Lut[curlut].yhi += (TOOFAR * 3.0);

    /* calculate the length in pixels of each tab */
    for (i=0; i < Lut[curlut].ntabs; i++) {
	Lut[curlut].tablength[i] = get_length(curlut, i);
    }
    
    /* at this point, calculate any corrections to the lookup table
     * and initialize the reflen array */
    init_reflen(curlut);

    return (curlut);
}

/* get_length - return the length in pixels of a tab (in a particular LUT) 
 * and calculate the sectionlength array for that tab. */
static double
get_length(int lut, int tab)
{
    int x1, y1, x2, y2;
    int i;
    double sectlen;
    double len = 0.0;

    assert (lut >= 0 && lut < Nluts);

    for (i=1; i<SC_NLUTPTS; i++) {
	x1 = Lut[lut].xlut[tab][i-1];
	y1 = Lut[lut].ylut[tab][i-1];
	x2 = Lut[lut].xlut[tab][i];
	y2 = Lut[lut].ylut[tab][i];
	sectlen = DIST(x1, y1, x2, y2);
	Lut[lut].sectionlength[tab][i-1] = sectlen;
	len += sectlen;
    }
    return (len);
}

/* sc_lu_tab - return the tab number for the given x, y. Return -1 if
 * the point is not on a mapped tab. */
int
sc_lu_tab(int lut, double x, double y)
{
    int i;
    int ref;	/* reference point in tab */
    int t;	/* tab */
    double yint[3];	/* interpolated y values */
    double dint[3];	/* abs. value of diff. betw. interpolated-y and y */
    int ret;

    if (lut < 0 || lut >= Nluts) {
	fprintf(stderr, "lookup: sc_lu_tab: bad lut value = %d. Aborting\n", 
		lut);
	return -1;
    }

    /* By knowing x, we can tell which reference points the interpolated
     * value will lie between. ref is the leftmost one.  (FIBER DEPENDENT
     * (can be set in lut file with the "reflimit" keyword)). */
    if (x <= Lut[lut].reflimit[0]) {
	ref = 0;
    } else if (x <= Lut[lut].reflimit[1]) {
	ref = 1;
    } else if (x <= Lut[lut].reflimit[2]) {
	ref = 2;
    } else if (x <= Lut[lut].reflimit[3]) {
	ref = 3;
    } else {
	ref = 4;
    }

    /* The tab to start testing with can be found knowing the y.  This
     * version gets the info from either the "layer" statement ("cutoff"
     * entry) in the lut file or from the lut data itself. */
    for (i=0; i<SC_NLAYERS; i++) {
	if (y < Lyr[lut][i].cutoff[ref]) {
	    t = Lyr[lut][i].start;
	    break;
	}
    }
    if (i >= SC_NLAYERS) {
	t = Lyr[lut][SC_NLAYERS-1].start;
    }

    /* Loop through the tabs, starting with the one selected above,
     * until a tab is found which has a y coord greater than that of
     * the y pixel value. */
    for ( ; t < Lut[lut].ntabs; t++) {
	 if (Lut[lut].ylut[t][ref] >= Lut[lut].ylut[t][ref+1]) {
	    /* points up */
	    if (y < Lut[lut].ylut[t][ref+1]) {
		break;
	    }
	 } else {
	    /* points down */
	    if (y < Lut[lut].ylut[t][ref]) {
		break;
	    }
	 }
    }

    /* Now find the interpolated y values for the nearest tabs. In
     * the normal case, we have to look at the current tab, the
     * previous tab, and the tab before that because if the slope of
     * the section is extreme it is possible that the point goes with
     * any of those tabs. */
    if (t == 0) {
	yint[2] = BIGNUM;
	yint[1] = BIGNUM;
	yint[0] = INTERP(lut, 0, ref, x);
    } else if (t == 1) {
	yint[2] = BIGNUM;
	yint[1] = INTERP(lut, 0, ref, x);
	yint[0] = INTERP(lut, 1, ref, x);
    } else if (t >= Lut[lut].ntabs) {
	t = Lut[lut].ntabs - 1;
	yint[2] = BIGNUM;
	yint[1] = INTERP(lut, t-1, ref, x);
	yint[0] = INTERP(lut, t, ref, x);
    } else {
	yint[2] = INTERP(lut, t-2, ref, x);
	yint[1] = INTERP(lut, t-1, ref, x);
	yint[0] = INTERP(lut, t, ref, x);
    }

    /* find the differences */
    for (i=0; i<3; i++) {
	dint[i] = fabs(yint[i] - y);
    }

    /* find the smallest difference - this should be the closest tab */
    if (dint[0] < dint[1]) {
	if (dint[0] < dint[2]) {
#ifdef NOTDEF
	    if (dint[0] > TOOFAR) {
		ret = -1;
	    } else {
		ret = t;
	    }
#else
	    ret = t;
#endif
	} else {
#ifdef NOTDEF	    
	    if (dint[2] > TOOFAR) {
		ret = -1;
	    } else {
		ret = t-2;
	    }
#else
	    ret = t-2;
#endif
	}
    } else if (dint[1] < dint[2]) {
#ifdef NOTDEF	
	if (dint[1] > TOOFAR) {
	    ret = -1;
	} else {
	    ret = t-1;
	}
#else
	ret = t-1;
#endif
    } else {
#ifdef NOTDEF	
        if (dint[2] > TOOFAR) {
	    ret = -1;
	} else {
	    ret = t-2;
	}
#else
	ret = t-2;
#endif
    }
    return ret;
}

/* sc_lu_layer - return the layer that corresponds to the specified tab */
int
sc_lu_layer(int lut, int tab)
{
    int layer;
    int i;

    if (lut < 0 || lut >= Nluts) {
	fprintf(stderr, "lookup: sc_lu_layer: bad lut value = %d. Aborting\n", 
		lut);
	return -1;
    }

    if (tab < Lyr[lut][0].start || tab > Lyr[lut][SC_NLAYERS-1].end) {
	layer = SC_NOLAYER;
    } else {
	for (i=0; i<SC_NLAYERS; i++) {
	    if (tab <= Lyr[lut][i].end) {
		layer = i;
		break;
	    }
	}
    }
    return (layer);
}

/* sc_lu_pos - return in *pos the coordinate postition within a layer
 * by first getting the position of the nearest reference point (from
 * reflen array) and then adding on the remaining distance to the
 * point.  Return 0 if all went well, else -1. */
int
sc_lu_pos(int lut, double *pos, int layer, int tab, double x, double y)
    /* lut - which lookup table to use */
    /* pos - coordinate position within layer */
    /* layer - SC_Y3, SC_X3 , etc */
    /* tab - absolute tab as returned by sc_lu_tab() */
    /* x, y - coords of the blob centroid in pixel space */
{
    int ret;
    
    if (lut < 0 || lut >= Nluts) {
	fprintf(stderr, "lookup: sc_lu_pos: bad lut value = %d. Aborting\n", 
		lut);
	return -1;
    }

    ret = chk_pointy(lut, layer, tab, x, y);
    if (ret != 0) {
	return ret;
    }
    *pos = find_pos(lut, (int *)NULL, layer, tab, x, y);
    return 0;
}

/* sc_lu_pos_extra - return in *pos the coordinate postition within a layer by
 * first getting the position of the nearest reference point (from reflen
 * array) and then adding on the remaining distance to the point.  Return 0 if
 * all went well, else -1. Also, find the position, tab, and layer for the
 * previous and next tabs. */
int
sc_lu_pos_extra(int lut, double *pos, int layer, int tab, double x, double y,
       double *prevpos, int *prevlayer, int *prevtab, double *px, double *py,
       double *nextpos, int *nextlayer, int *nexttab, double *nx, double *ny)
    /* lut - which LUT to use */
    /* pos - coordinate position within layer */
    /* layer - SC_Y3, SC_X3 , etc */
    /* tab - absolute tab as returned by sc_lu_tab() */
    /* x, y - coords of the blob centroid in pixel space */
{
    int xref;
    int ret;
    
    if (lut < 0 || lut >= Nluts) {
	fprintf(stderr, "lookup: sc_lu_pos_extra: bad lut value = %d. Aborting\n", 
		lut);
	return -1;
    }

    ret = chk_pointy(lut, layer, tab, x, y);
    if (ret != 0) {
	return ret;
    }
    *pos = find_pos(lut, &xref, layer, tab, x, y);

    find_prev_next_vals(lut, prevtab, prevlayer, prevpos, px, py,
			nexttab, nextlayer, nextpos, nx, ny,
			tab, layer, xref, x);

    return 0;
}

static double
find_pos(int lut, int *xrefval, int layer, int tab, double x, double y)
{
    double len = 0.0;
    int xref;

    assert (lut >= 0 && lut < Nluts);

    if (Lyr[lut][layer].leftrt == LEFTRT) {
        /* pixels go from left to right */
        for (xref=0; xref < SC_NLUTPTS; xref++) {
            if (Lut[lut].xlut[tab][xref] > x) {
                break;
            }
        }
	len = xdist(lut, (double *)NULL, x, tab, xref, xref-1);
    } else {
        /* pixels go from right to left */
        for (xref=SC_NLUTPTS-1; xref >= 0; xref--) {
            if (Lut[lut].xlut[tab][xref] <= x) {
                break;
            }
        }
	len = xdist(lut, (double *)NULL, x, tab, xref, xref+1);
    }
    if (xrefval != (int *)NULL) {
	*xrefval = xref;
    }
    return(len);
}

static int
find_prev_next_vals(int lut, 
    int *ptab, int *player, double *plen, double *px, double *py,
    int *ntab, int *nlayer, double *nlen, double *nx, double *ny,
    int tab, int layer, int xref, double x)
{
    double tmpy;	/* dummy y value of prev, next point */
    int leftrt;
    double a1, a2;
    int ret;
    
    assert (lut >= 0 && lut < Nluts);

    *nx = *px = x;

    /* previous tab values */
    tmpy = -1.0;
    *ptab = tab - 1;
    if (*ptab < 0) {
	/* tab was the very first one - return dummy values*/
	*player = -1;
	*plen = -1.0;
    } else {
	if (tab == Lyr[lut][layer].start) {
	    *player = layer-1;  /* this tab is starting tab of this layer,
				 * so prev tab must be in a different layer */
	} else {
	    *player = layer;    /* prev tab is still in this layer */
	}
	ret = chk_point(lut, *player, *ptab, x);
	if (ret != 0) {
	    *plen = -1.0;
	} else {
	    leftrt = Lyr[lut][layer].leftrt;
	    if (leftrt == LEFTRT) {
		if (Lyr[lut][*player].leftrt == leftrt) {
		    /* both tabs same direction */
		    a1 = xref;
		    a2 = xref - 1;
		} else {
		    a1 = xref - 1;
		    a2 = xref;
		}
	    } else {
		if (Lyr[lut][*player].leftrt == leftrt) {
		    /* both tabs same direction */
		    a1 = xref;
		    a2 = xref + 1;
		} else {
		    a1 = xref + 1;
		    a2 = xref;
		}
	    }
	    *plen = xdist(lut, &tmpy, x, *ptab, a1, a2);
	}
    }
    *py = tmpy;

    /* next tab values */
    tmpy = -1.0;
    *ntab = tab + 1;
    if (*ntab >= SC_NTABS) {
	/* tab was the very last one - return dummy values*/
	*nlayer = -1;
	*nlen = -1.0;
    } else {
	if (tab == Lyr[lut][layer].end) {
	    *nlayer = layer+1;  /* this tab is ending tab of this layer,
				 * so next tab must be in a different layer */
	} else {
	    *nlayer = layer;    /* next tab is still in this layer */
	}
	ret = chk_point(lut, *nlayer, *ntab, x);
	if (ret != 0) {
	    *nlen = -1.0;
	} else {
	    leftrt = Lyr[lut][layer].leftrt;
	    if (leftrt == LEFTRT) {
		if (Lyr[lut][*nlayer].leftrt == leftrt) {
		    /* both tabs same direction */
		    a1 = xref;
		    a2 = xref - 1;
		} else {
		    a1 = xref - 1;
		    a2 = xref;
		}
	    } else {
		if (Lyr[lut][*nlayer].leftrt == leftrt) {
		    /* both tabs same direction */
		    a1 = xref;
		    a2 = xref + 1;
		} else {
		    a1 = xref + 1;
		    a2 = xref;
		}
	    }
	    *nlen = xdist(lut, &tmpy, x, *ntab, a1, a2);
	}
    }
    *ny = tmpy;
}


char *
sc_lu_report_layer(int lut, int layerval)
{
    static char layertext[80];

    if (lut < 0 || lut >= Nluts) {
	fprintf(stderr, 
		"lookup: sc_lu_report_layer: bad lut value = %d. Aborting\n", 
		lut);
	sprintf(layertext, "No such LUT");
    }

    if (layerval == SC_NOLAYER) {
	sprintf(layertext, "SC_NOLAYER");
    } else if (0 <= layerval && layerval < SC_NLAYERS) {
	sprintf(layertext, Lyr[lut][layerval].name);
    } else {
	sprintf(layertext, "Unknown layer = %d (max is %d)",
		layerval, SC_NLAYERS-1);
    }
    return layertext;
}

/* init_reflen - initialize the reflen array, which gives the
 * position of each reference point in the the lookup table. */
static void
init_reflen(int lut)
{
    double len;
    int t, p;
    int tab, layer;

    assert (lut >= 0 && lut < Nluts);

    for (tab=0; tab < Lut[lut].ntabs; tab++) {
	layer = sc_lu_layer(lut, tab);
	if (layer == SC_NOLAYER) {
	    continue;
	}

	/* first add up the lengths of all the tabs up to this one */
    	len = 0.0;
	if (Lyr[lut][layer].topbot == TOPBOT) {
	    /* tabs go from top to bottom */
	    for (t=Lyr[lut][layer].start; t<tab; t++) {
		len += Lut[lut].tablength[t];
	    }
	} else {
	    /* tabs go from bottom to top */
	    for (t=Lyr[lut][layer].end; t>tab; t--) {
		len += Lut[lut].tablength[t];
	    }
	}

    	/* now add on the distance to each ref point in this tab */
	if (Lyr[lut][layer].leftrt == LEFTRT) {
	    /* pixels go from left to right */
	    for (p=0; p<SC_NLUTPTS; p++) {
		Lut[lut].reflen[tab][p] = len;
		len += Lut[lut].sectionlength[tab][p];
	    }
	} else {
	    /* from right to left go the pixels */
	    for (p=SC_NLUTPTS-1; p>=0; p--) {
		Lut[lut].reflen[tab][p] = len;
		len += Lut[lut].sectionlength[tab][p-1];
	    }
	}
    }
}

#define VERLEN	20	/* max length of version string */

/* sc_lu_write_binary_table - write out the essential data structures
 * to a file in binary format */
int
sc_lu_write_binary_table(int lut, char *filename)
{
    FILE *wb;
    char version[VERLEN];

    if (lut < 0 || lut >= Nluts) {
	fprintf(stderr, 
		"lookup: sc_lu_write_binary_table: bad lut value = %d. Aborting\n", 
		lut);
	return -1;
    }

    wb = fopen(filename, "wb");
    if (wb == (FILE *)NULL) {
	fprintf(stderr, "sc_lu_write_binary_table: can't open file '%s' (%s)\n",
		filename, strerror(errno));
	return -1;
    }
    strncpy(version, LOOKUP_VERSION, VERLEN);
    fwrite(version, sizeof(char), VERLEN, wb);
    fwrite((char *)&Lut[lut], sizeof(Lut[0]), 1, wb);
    fwrite((char *)Lyr[lut], sizeof(Lyr[0][0]), SC_NLAYERS, wb);
    fclose(wb);
    return 0;
}

/* sc_lu_read_binary_table - read the essential data structures from
 * the binary lut file. Returns a non-negative LUT id, or -1 if there
 * was a problem. */
int
sc_lu_read_binary_table(char *filename)
{
    struct layerinfo tmplyr[SC_NLAYERS];
    struct lutinfo tmplut; 
    FILE *rb;
    char version[VERLEN];

    if (Nluts >= SC_NLUTS) {
	fprintf(stderr, "lookup: sc_lu_read_binary_table: too many LUTs\n");
	return -1;
    }

    rb = fopen(filename, "rb");
    if (rb == (FILE *)NULL) {
	fprintf(stderr, "sc_lu_read_binary_table: can't open file '%s' (%s)\n",
		filename, strerror(errno));
	return -1;
    }
    fread(version, sizeof(char), VERLEN, rb);
    if (!SOFT_STREQ(version, LOOKUP_VERSION)) {
	/* not the same version - issue warning */
	fprintf(stderr, "WARNING: sc_lu_read_binary_table: different versions!"
			" (is %s, but should be %s)\n",
			version, LOOKUP_VERSION);
    }
    
    fread((char *)&tmplut, sizeof(tmplut), 1, rb);
    fread((char *)tmplyr, sizeof(tmplyr[0]), SC_NLAYERS, rb);
    fclose(rb);

    memcpy(&Lut[Nluts], &tmplut, sizeof(tmplut));
    memcpy(Lyr[Nluts], tmplyr, sizeof(tmplyr[0]) * SC_NLAYERS);
    Nluts++;
    return (Nluts-1);
}

#ifdef NOTDEF	/* not used at present */
static double
find_maxy(int lut, int tab)
{
    double max;
    int i;
    if (lut < 0 || lut >= Nluts) {
	fprintf(stderr, "lookup: find_maxy: bad lut value = %d. Aborting\n", 
		lut);
	return -1;
    }

    max = Lut[lut].ylut[tab][0];
    for (i=1; i<SC_NLUTPTS; i++) {
	if (Lut[lut].ylut[tab][i] > max) {
	    max = Lut[lut].ylut[tab][i];
	}
    }
    return max;
}

static double
find_miny(int tab)
{
    double min;
    int i;
    if (lut < 0 || lut >= Nluts) {
	fprintf(stderr, "lookup: find_miny: bad lut value = %d. Aborting\n", 
		lut);
	return -1;
    }

    min = yLut[lut].lut[tab][0];
    for (i=1; i<SC_NLUTPTS; i++) {
	if (Lut[lut].ylut[tab][i] < min) {
	    min = Lut[lut].ylut[tab][i];
	}
    }
    return min;
}
#endif	/* NOTDEF */

/* sc_lu_pix_to_real_pos() - given an (x,y) in pixel space, find the tab,
 * layer, and position in real space. Returns 0 if all goes well, else -1.
 * The values of tab and layer are filled in as a side-effect. */
int
sc_lu_pix_to_real_pos(int lut, double x, double y, int *tab, 
    int *layer, double *rpos)
{
    int ret;
    if (lut < 0 || lut >= Nluts) {
	fprintf(stderr, 
	    "lookup: sc_lu_pix_to_real_pos: bad lut value = %d. Aborting\n", 
	    lut);
	return -1;
    }

    *tab = sc_lu_tab(lut, x, y);
    if (*tab == -1) {
	return -1;
    }
    *layer = sc_lu_layer(lut, *tab);
    if (*layer == SC_NOLAYER) {
	return -1;
    }
    ret = sc_lu_pos(lut, rpos, *layer, *tab, x, y);
    if (ret != 0) {
	return -1;
    }
    return 0;
}

void
sc_lu_print_reflen(int lut)
{
    int t;
    int p;
    for (t=0; t<SC_NTABS; t++) {
	printf("tab %d\t", t);
	for (p=0; p<SC_NLUTPTS; p++) {
	    printf("%.2lf\t", Lut[lut].reflen[t][p]);
	}
	printf("\n");
    }
}

/* xdist - find the "corrected" distance from the given point to the given
 * segment. */
static double
xdist(int lut, double *y, double x, int tab, int xref1, int xref2)
    /* lut = which lut to use 
     * y = returned y coord
     * x = x coord of centroid
     * tab =fiber tab
     * xref1, xref2 = end pts of nearest segment in tab */
{
    double m;	/* slope of segment */
    double x1, y1, x2, y2;
    double ny;
    double len;
    
    assert (lut >= 0 && lut < Nluts);

    x1 = Lut[lut].xlut[tab][xref1];
    y1 = Lut[lut].ylut[tab][xref1];
    x2 = Lut[lut].xlut[tab][xref2];
    y2 = Lut[lut].ylut[tab][xref2];

    if (x2 - x1 == 0.0) {
	fprintf(stderr, "lookup(xdist) denom=0. Shouldn't happen. Aborting\n");
	return -1;
    }
    m = (y2 - y1) / (x2 - x1);
    ny = (m * (x - x1)) + y1;
    len = Lut[lut].reflen[tab][xref1] - 
	    DIST(Lut[lut].xlut[tab][xref1], Lut[lut].ylut[tab][xref1], x, ny);

    if (y != (double *)NULL) {
	*y = ny;
    }
    return (len);
}

/* chk_point - some sanity checks */
static int
chk_point(int lut, int layer, int tab, double x)
{
    assert (lut >= 0 || lut < Nluts);
    if (layer < 0 || layer > SC_NLAYERS) {
        return -1;
    }
    if (tab < Lyr[lut][layer].start || tab > Lyr[lut][layer].end) {
        return -2;
    }
    if (x < Lut[lut].xlut[tab][0] || x > Lut[lut].xlut[tab][SC_NLUTPTS-1]) {
        return -3;
    }
    return 0;
}

/* chk_pointy - same as chk_point, but also make sure the y value is okay */
static int
chk_pointy(int lut, int layer, int tab, double x, double y)
{
    int ret;
    assert (lut >= 0 || lut < Nluts);
    if (ret = chk_point(lut, layer, tab, x)) {
	return ret;
    }
    if (y < Lut[lut].ylo || y > Lut[lut].yhi) {
	return -4;
    }
    return 0;
}
