/* minorframe.c - process the cris minor frames
 *
 *	Routines to handle the cris minor frames. See the sc_crisSubset
 *	struct in softcris.h for details about what is contained in the
 *	minor frame. The description of the CrisData part of the minor
 *	frame (consisting of housekeeping, command responses, normal and
 *	diagnostic data) is found at the end of this file.
 *	
 *	Note that the input for these routines is the raw telemetry data 
 *	which has been processed by cris_test_extr, a program which is not
 *	exactly part of the soft library. The source code of this program
 *	and Robert Radocynski's downlink "library" is found in the
 *	directory Util/Downlink.
 *
 * Marty Olevitch, Jan 1997
 */

static char cvsid[] = "$Id: minorframe.c,v 1.19 1998/07/22 21:09:02 marty Exp $";

#include <stdio.h>
#include <stdlib.h>	/* calloc */
#include <string.h>	/* memcpy */
#include <errno.h>
#include <time.h>

#include "misc.h"
#include "softcris.h"
#include "hsktab.h"

#define FRMSTART_ID	0x2ace		/* must match 
					 * Util/Downlink/cris_text_extr.c */
#define ALT_FRMSTART_ID	0xce2a		/* bytes swapped */

static void extract_hsk(u_char h1, u_char h2, u_char h3, int seq);
static void clear_hsk(void);

/* buffer for storing normal evt which extends over more than 1 minor frm */
#define NORM_BUF_SIZ 1000
static u_char Norm_buf[NORM_BUF_SIZ];
static u_char *Norm_bufp;

/* DIAGMAX = maximum size of a diagnostic evt. Value was arrived as
 * follows: there is a software limit of SC_SOFT_MAXSIZE (as of Feb, 1997,
 * the value was 6000) bytes for a SOFT image in CRIS. So a diagnostic could be
 *
 *	6000 bytes (SOFT) 	(SC_SOFT_MAXSIZE)
 *     +  72 bytes (raw stack)	(SC_STACKSIZE)
 *     + 200 bytes (cris normal approx. maximum)
 *	----------
 *	6272 bytes total
 */
#define DIAGMAX (SC_SOFT_MAXSIZE + SC_STACKSIZE + 200)
static u_char Diagbuf[DIAGMAX];	/* buffer for diagnostic evt */
#define LOSTDIAGAMT (SC_CRIS_TELEMBYTES - 3)

static void do_save_minor(struct sc_minorFrame *src, struct sc_minorFrame *dest);

/* Save_last_minor = 1 if we should be saving each minor frame in
 * sc_tlm_next_minor(). Set and reset by sc_tlm_save_minor(). */
static int Save_last_minor = 0;

/* Just_did_sync = 1 if we have just finished doing a sync_major. Set in
 * sc_tlm_sync_major(). Reset in sc_tlm_next_minor(). */
static int Just_did_sync = 0;

/* Doing_sync = 1 if we are in the process of trying to find a sync. Set
 * and reset by sc_tlm_sync_major(). This variable is here so that we can
 * clear the hsk table right after finding a sync, but before extracting
 * the hsk bits from that first event. */
static int Doing_sync = 0;

/* Curminor - most recently acquired sc_minorFrame */
static struct sc_minorFrame Curminor;

/* Cur - most recently acquired sc_crisSubset */
static sc_crisSubset Cur;

/* routines that deal with the minor frame date */
static int ConvertToSeconds1970( int date, int time);
static void ConvertFromSeconds1970( int seconds, int *date, int *time);

/* sc_tlm_next_evt - get the next event in the stream of minor frames,
 * whether it is a normal or diagnostic event. The return code is the same
 * as for sc_tlm_next_minor() plus -7 if the diagnostic event is too 
 * large and (-8 + ret) if sc_diag_unpack() returns an error code. 
 * The norm parameter will be 1 if the next evt is a normal evt, else 0. 
 * For normal evts, the c pointer will have valid data; for diagnostic, 
 * the d pointer. */
int
sc_tlm_next_evt(c, d, fp, norm)
    struct sc_crisEvt *c;
    struct sc_diagEvt *d;
    FILE *fp;
    int *norm;
{
    static int curevt = 0;
    static struct sc_minorFrame curminor;
    u_char *diagp;
    static int norm_evt_avail = 0;
    int ret;

    if (norm_evt_avail) {
	/* there is still another event left in the current minor frm */
	*c = curminor.e[curevt];
	curevt++;
	if (curevt >= curminor.nevt) {
	    norm_evt_avail = 0;
	}
	*norm = 1;
	return 1;
    }

    /* no evts available. Need to read more minor frames. */
    while (1) {
	ret = sc_tlm_next_minor(&curminor, fp);
	if (ret != 1 && ret != -6) {
	    return ret;
	}
	if (curminor.nevt > 0) {
	    /* we have one or more normal evts */
	    *c = curminor.e[0];
	    if (curminor.nevt > 1) {
		curevt = 1;
		norm_evt_avail = 1;
	    }
	    *norm = 1;
	    return 1;
	}
	if (curminor.diag) {
	    /* got a diagnostic evt */
	    diagp = Diagbuf;
	    memset(Diagbuf, '\0', DIAGMAX);
	    while (curminor.diag > 0) {
		if (curminor.lost_prev_frame) {
		    if ((diagp + LOSTDIAGAMT) - Diagbuf > DIAGMAX) {
			return -7;
		    }
		    memset(diagp, 0xee, LOSTDIAGAMT);
		    diagp += LOSTDIAGAMT;
		}
		if ((diagp + curminor.diag) - Diagbuf > DIAGMAX) {
		    /* diagnostic data is too big */
		    return -7;
		}
		memcpy(diagp, curminor.diagdata, curminor.diag);
		diagp += curminor.diag;
		ret = sc_tlm_next_minor(&curminor, fp);
		if (ret != 1 && ret != -6) {
		    return ret;
		}
	    }

	    if (curminor.nevt > 0) {
		/* save the normal evts for next time */
		curevt = 0;
		norm_evt_avail = 1;
	    }

#ifdef NOTDEF
/* the camera guessing stuff is not reliable enough so just accept the
 * default of SC_CAM2. If the programmer wants, he can explicitly use
 * sc_cris_set_cam(SC_CAM1) to change it. */
	    ret = sc_tlm_cam_guess();
	    if (ret != -1) {
		sc_cris_set_cam(ret);
	    }
#endif /* NOTDEF */
	    ret = sc_diag_unpack_cam(Diagbuf, d);
	    if (ret) {
		return (-8 + ret);
	    }

	    *norm = 0;
	    return 1;
	}
    }
}

/* sc_tlm_next_raw_diag - get the next diagnostic event in the 
 * stream of minor frames and fill in the pointer to the RAW DATA for 
 * this event, plus the number of bytes in it. The return code is the same 
 * as for sc_tlm_next_minor() plus -7 if the diagnostic event 
 * is too large. The norm parameter will be 1 if the next evt 
 * is a normal evt, else 0. */
int
sc_tlm_next_raw_diag(ptr, n, fp)
    u_char *ptr;	/* pointer to raw data */
    int *n;		/* no. bytes of raw data */
    FILE *fp;
{
    struct sc_minorFrame curminor;
    u_char *diagp;
    int ret;

    while (1) {
	ret = sc_tlm_next_minor(&curminor, fp);
	if (ret != 1 && ret != -6) {
	    return ret;
	}
	if (curminor.diag) {
	    /* got a diagnostic evt */
	    diagp = Diagbuf;
	    memset(Diagbuf, '\0', DIAGMAX);
	    while (curminor.diag > 0) {
		if (curminor.lost_prev_frame) {
		    if ((diagp + LOSTDIAGAMT) - Diagbuf > DIAGMAX) {
			return -7;
		    }
		    memset(diagp, 0xee, LOSTDIAGAMT);
		    diagp += LOSTDIAGAMT;
		}
		if ((diagp + curminor.diag) - Diagbuf > DIAGMAX) {
		    /* diagnostic data is too big */
		    return -7;
		}
		memcpy(diagp, curminor.diagdata, curminor.diag);
		diagp += curminor.diag;
		ret = sc_tlm_next_minor(&curminor, fp);
		if (ret != 1 && ret != -6) {
		    return ret;
		}
	    }

	    *n = diagp - Diagbuf;
	    memcpy(ptr, Diagbuf, *n);
	    return 1;
	}
    }
}



/* sc_tlm_next_cris_evt - return a pointer to the next cris normal event in
 * the sequence of minor frames. Returns the same values as
 * sc_tlm_next_minor() */
int
sc_tlm_next_cris_evt(c, fp)
    struct sc_crisEvt *c;
    FILE *fp;
{
    struct sc_diagEvt d;
    int norm;
    int ret;

    while (1) {
	ret = sc_tlm_next_evt(c, &d, fp, &norm);
	if (ret == 0) {
	    return 0;
	}
	if (!norm || ret <= -7) {
	    /* ignore diagnostic and diagnostic-related error codes */
	    continue;
	}
	if (norm || ret != 1) {
	    return ret;
	}
    }
}

/* sc_tlm_next_diag - return a point to the next diagnostic event in the
 * sequence of minor frames. Returns the same values as sc_tlm_next_minor()
 * plus -7 if the diagnostic event is too large and (-8 + ret) if 
 * sc_diag_unpack() returns an error code. */
int
sc_tlm_next_diag(d, fp)
    struct sc_diagEvt *d;
    FILE *fp;
{
    struct sc_crisEvt c;
    int norm;
    int ret;

    while (1) {
	ret = sc_tlm_next_evt(&c, d, fp, &norm);
	if (ret == 0) {
	    return 0;
	}
	if (norm || ret == -2 || ret == -4 || ret == -5 || ret == -6) {
	    /* -2 = cmd response too long
	     * -4 = bad normal event
	     * -5 = bad event ID
	     * -6 = bad search for next normal evt
	     * any of which shouldn't bother us if we are looking 
	     * for a diagnostic event */
	    continue;
	}
	if (!norm || ret != 1) {
	    /* either a diagnostic event or a different error */
	    return ret;
	}
    }
}


/* sc_tlm_sync_major - read data from fp until we come to the start of
 * a new major frame, which is indicated by major_frame_start element being
 * equal to 1 in the sc_minorFrame struct. Returns the number of frames we
 * had to read when we detect the start of the major frame; 0 on EOF; a 
 * negative return value indicates an error from sc_tlm_next_minor(). */
int
sc_tlm_sync_major(fp)
    FILE *fp;
{
    int cnt = 0;
    struct sc_minorFrame f;
    int old;
    int ret;

    Just_did_sync = 0;
    Doing_sync = 1;

    while (1) {
	cnt++;
	ret = sc_tlm_next_minor(&f, fp);
	if (ret == 0) {
	    break;
	} else if (ret == -1) {
	    break;
	} else if (ret < 0) {
	    continue;	/* probably some garbage in the data */
	} else if (f.major_frame_start) {
	    ret = cnt;
	    break;
	}
    }

    /* store the minor frame for sc_tlm_next_minor() */
    old = sc_tlm_save_minor(1);
    do_save_minor(&f, &Curminor);
    (void)sc_tlm_save_minor(old);

    Doing_sync = 0;
    Just_did_sync = 1;
    return ret;
}


/* Reads the next minor frame data from the open file pointer fp and fills 
 * in the values pointed to by the sc_minorFrame pointer f. Returns 1 if
 * all went well; 0 for EOF; -1 on read error; -2 if the command reponse
 * length is too long; -3 if a normal event length is too long; -4 on error
 * from sc_cris_unpack_cam(); -5 if the event doesn't start with the
 * expected data ID 0x2ace (inserted by cris_test_extr). */
int
sc_tlm_next_minor(f, fp)
    struct sc_minorFrame *f;
    FILE *fp;
{
    int amt;
    int amtleft;	/* no. bytes left in minor frame */
    int cmdlen;		/* length of command response (bytes) */
#define EXTRASZ 7	/* max no. extra bytes to leave over from prior frm  */
    /* The "+ EXTRASZ" below is to provide room for any extra leftover from 
     * the prior minor frame that we have to check for cris evt start. */
    static u_char Cur_cris_data[SC_CRIS_TELEMBYTES + EXTRASZ];
    u_char cursync;
    u_int diff;
    static u_char Extra[EXTRASZ];	/* extra bytes from prior frm */
    static int Firsttime = 1;
    int i;
    unsigned short id[2];	/* data id */
    static int In_normal = 0;	/* 1 if we are processing a long normal evt */
    int j;
    u_char k;
    static u_char Lastsync;
    static int Len;		/* length of normal evt left to read (bytes) */
    int lost_minor_frame;
    int n;
    int needswap;		/* 1 if we need to swap bytes in data */
    int nlost;
    static int Nextra = 0;	/* no. extra bytes left over from prior frm */
    int nbytes_cur_frame;	/* no. bytes in current minor frame */
    sc_crisSubset next;
    static u_int Prevcnt = 0;	/* previous PacketSecondaryHeader */
    int ret;
    sc_crisSubset revsubset;
    static int Seq = -1;	/* sequence no. in major frame */
    char *startbuf;
    long secs1996;

    /* If Just_did_sync == 1, then we have just synced to a major frame.
     * Send back the most recently acquired minor frame in that case. */
    if (Just_did_sync) {
	Just_did_sync = 0;
	sc_most_recent_minor(f);
	return 1;
    }

    /* We have to have read in two minor frames, the current one and the
     * "next" one. This annoying detail is necessary because the first 4
     * bytes of a minor frame is actually data belonging to the previous
     * minor frame. I think this even applies to the first minor frame in a
     * major frame.*/

    startbuf = (char *)id;
    needswap = 0;
    if (Firsttime) {
	/* read in the first "current" frame */
	Firsttime = 0;
	ret = fread(startbuf, 1, 8, fp);
	if (ret != 8) {
	    if (feof(fp)) {
		return 0;
	    } else if (ferror(fp)) {
		return -1;
	    }
	}

	if (id[1] == FRMSTART_ID) {
	    ;	/* OK */
	} else if (id[1] == ALT_FRMSTART_ID) {
	    needswap = 1;
	} else {
	    return -5;
	}

	if (needswap) {
	    ret = fread((char *)&revsubset, 1, sizeof(sc_crisSubset), fp);
	    sc_reorder_cris_subset(&revsubset, &Cur);
	} else {
	    ret = fread((char *)&Cur, 1, sizeof(sc_crisSubset), fp);
	}

	if (ret != sizeof(sc_crisSubset)) {
	    if (feof(fp)) {
		return 0;
	    } else if (ferror(fp)) {
		return -1;
	    }
	}
    }

    /* read in the next frame */
    ret = fread(startbuf, 1, 8, fp);
    if (ret != 8) {
	if (feof(fp)) {
	    return 0;
	} else if (ferror(fp)) {
	    return -1;
	}
    }

    if (id[1] == FRMSTART_ID) {
	;	/* OK */
    } else if (id[1] == ALT_FRMSTART_ID) {
	needswap = 1;
    } else {
	return -5;
    }

    if (needswap) {
	ret = fread((char *)&revsubset, 1, sizeof(sc_crisSubset), fp);
	sc_reorder_cris_subset(&revsubset, &next);
    } else {
	ret = fread((char *)&next, 1, sizeof(sc_crisSubset), fp);
    }

    if (ret != sizeof(sc_crisSubset)) {
	if (feof(fp)) {
	    return 0;
	} else if (ferror(fp)) {
	    return -1;
	}
    }

    f->s = Cur;
    f->hsk[0] = Cur.CrisData[4];
    f->hsk[1] = Cur.CrisData[5];
    f->hsk[2] = Cur.CrisData[6];

    if (Cur.PacketSecondaryHeader == 0) {
	/* assume this is some kind of garbage or filler */
	Cur = next;
	return -7;
    }

    /* PacketSecondaryHeader seems to be a running minor frame count. It
     * is a lot simpler to use it for our lost frame detection rather than
     * the spacecraft major and minor frame numbers which roll over too
     * quickly. */
    if (Prevcnt) {
	diff = Cur.PacketSecondaryHeader - Prevcnt;	/* normal case */
    } else {
	diff = 1;	/* first time through; fake it */
    }
    Prevcnt = Cur.PacketSecondaryHeader;

    if (diff == 1) {
	lost_minor_frame = 0;
	f->lost_prev_frame = 0;
    } else {
	nlost = diff - 1;
	lost_minor_frame = nlost;
	f->lost_prev_frame = nlost;
	Seq = (Seq + nlost) % 256;
    }

    /* A transition of the SYNC bit from 1 to 0 indicates the start of a
     * new major frame. */
    cursync = f->hsk[0] & SC_TLM_SYNC;
    if (!cursync && Lastsync) {
	if (Doing_sync) {
	    clear_hsk();
	}

	if (!Doing_sync && Seq < 255) {
	    /* make sure we are at or more than seq no. 255. If not, 
	     * assume we have a bogus lack of sync bit and this should 
	     * not be a major frame transition. The goto seems 
	     * understandable here. */
	    goto not_major_frame_start;
	} else {
	    f->major_frame_start = 1;
	    if (!lost_minor_frame) {
		Seq = -1; 	/* Seq will be incremented below */
	    }
	}
    } else {
not_major_frame_start:
	f->major_frame_start = 0;
    }

    Seq++;
    f->seq = Seq;
    Lastsync = cursync;

    cmdlen = 0;
    i = 0;
    if (f->hsk[0] & SC_TLM_CMD) {
	/* command response present */
	cmdlen = (int)Cur.CrisData[7];
	if (cmdlen > SC_CRIS_TELEMBYTES - 4) {
	    /* assume this is some kind of garbage */
	    /* make "next" frame into "current" for next time through loop */
	    Cur = next;
	    return -2;
	}
	for (j=8,i=0; j<cmdlen+8; i++,j++) {
	    k = Cur.CrisData[j];
	    if (k == '\n') {
		f->cmd[i] = '\\';
		i++;
		f->cmd[i] = 'n';
	    } else if (k == '\r') {
		f->cmd[i] = '\\';
		i++;
		f->cmd[i] = 'r';
	    } else {
		f->cmd[i] = k;
	    }
	}
    }
    f->cmd[i] = '\0';

    i = (cmdlen == 0) ? 7 : (8 + cmdlen);
    amt = SC_CRIS_TELEMBYTES - i;
    

    if (f->hsk[0] & SC_TLM_DIAG) {
	/* diagnostic data */
	memcpy(Cur_cris_data, Cur.CrisData+i, amt);
	memcpy(Cur_cris_data + amt, next.CrisData, 4);
	nbytes_cur_frame = amt + 4;
	/* make "next" frame into "current" frame for next time through loop */
	Cur = next;
	memcpy(f->diagdata, Cur_cris_data, nbytes_cur_frame);
	f->nevt = 0;	/* no normal evts */
	f->diag = nbytes_cur_frame;
	do_save_minor(f, &Curminor);
	ret = 1;
	goto handle_housekeeping;
    } else {
	/* normal evts */
	memcpy(Cur_cris_data, Extra, Nextra);
	memcpy(Cur_cris_data + Nextra, Cur.CrisData+i, amt);
	memcpy(Cur_cris_data + Nextra + amt, next.CrisData, 4);
	nbytes_cur_frame = Nextra + amt + 4;

	/* make "next" frame into "current" frame for next time through loop */
	Cur = next;

	if (f->major_frame_start) {
	    /* a new major frame terminates any pending normal evt */
	    In_normal = 0;
	}
	f->diag = 0;
	f->nevt = 0;
	do_save_minor(f, &Curminor);
	i = n = 0;
	amtleft = nbytes_cur_frame;

	if (lost_minor_frame) {
	    In_normal = 0;
	}

	if (In_normal) {
	    /* working on normal evt continued from previous minor frame */
	    if (Len > amtleft) {
		/* evt continues into next minor frame */
		memcpy(Norm_bufp, Cur_cris_data, amtleft);
		Norm_bufp += amtleft;
		Len -= amtleft;
		ret = 1;
		goto handle_housekeeping;
	    } else {
		/* evt ends in this minor frame */
		memcpy(Norm_bufp, Cur_cris_data, Len);
#ifdef NOTDEF
/* the camera guessing stuff is not reliable enough so just accept the
 * default of SC_CAM2. If the programmer wants, he can explicitly use
 * sc_cris_set_cam(SC_CAM1) to change it. */
		ret = sc_tlm_cam_guess();
		if (ret != -1) {
		    sc_cris_set_cam(ret);
		}
#endif /* NOTDEF */
		secs1996 = sc_tlm_sec1996(&f->s);
		if (secs1996 == (long)-1) {
		    sc_cris_set_secs1996(0);
		} else {
		    sc_cris_set_secs1996(secs1996);
		}
		ret = sc_cris_unpack_cam(Norm_buf, &(f->e[0]));
		if (ret) {
		    In_normal = 0;
		    ret = -4;
		    goto handle_housekeeping;
		}
		/* memset(Norm_buf, '\0', NORM_BUF_SIZ); */
		Norm_bufp = Norm_buf;
		i += Len;
		n++;
		f->nevt++;
		do_save_minor(f, &Curminor);
		In_normal = 0;
		amtleft -= Len;
		/* continue below to see if more evts in this minor frame */
	    }
	}

	while (amtleft) {
	    /* shouldn't be any 0 bytes at this point */
	    while (Cur_cris_data[i] == '\0' && amtleft > 0) {
		i++;
		amtleft--;
	    }

	    if (amtleft <= EXTRASZ) {
		memcpy(Extra, Cur_cris_data  + i, amtleft);
		Nextra = amtleft;
		ret = 1;
		goto handle_housekeeping;
	    }

	    j = sc_search_norm_start(Cur_cris_data + i, amtleft);
	    if (j == -1) {
		int offset;
		offset = i + (amtleft - EXTRASZ);
		memcpy(Extra, Cur_cris_data + offset, EXTRASZ);
		Nextra = EXTRASZ;
		ret = -6;
		goto handle_housekeeping;
	    }
	    i += j;
	    amtleft -= j;
	    Nextra = 0;

	    Len = (int) Cur_cris_data[i];
	    if (Len <= 0) {
		break;
	    }
	    if (Len > amtleft) {
		/* evt continues into next minor frame */
		In_normal = 1;
		memcpy(Norm_buf, Cur_cris_data+i, amtleft);
		Len -= amtleft;
		Norm_bufp = Norm_buf + amtleft;
		break;
	    }
#ifdef NOTDEF
	    ret = sc_tlm_cam_guess();
	    if (ret != -1) {
		sc_cris_set_cam(ret);
	    }
#endif /* NOTDEF */
	    secs1996 = sc_tlm_sec1996(&f->s);
	    if (secs1996 == (long)-1) {
		sc_cris_set_secs1996(0);
	    } else {
		sc_cris_set_secs1996(secs1996);
	    }
	    ret = sc_cris_unpack_cam(Cur_cris_data + i, &(f->e[n]));
	    if (ret) {
		ret = -4;
		goto handle_housekeeping;
	    }
	    n++;
	    f->nevt++;
	    do_save_minor(f, &Curminor);
	    i += Len;
	    if (i >= nbytes_cur_frame) {
		break;
	    }
	    amtleft -= Len;
	}
	ret = 1;
	goto handle_housekeeping;
    }

handle_housekeeping:
    if (ret >= 0) {
	/* only extract the housekeeping if we didn't have any problems
	 * with this event */
	extract_hsk(f->hsk[0], f->hsk[1], f->hsk[2], Seq);
    }
    return ret;
}

/* clear_hsk - set all the housekeeping values to -1 */
static void
clear_hsk(void)
{
    int i;
    for (i=0; i<Hsktabsize; i++) {
	Hsktab[i].rawval = -1;
    }
}

/* extract_hsk - fill in the Hsktab with the value from this minor frame */
static void
extract_hsk(h1, h2, h3, seq)
    u_char h1;
    u_char h2;
    u_char h3;
    int seq;
{
    u_int h;
    struct tlmHsk *hp;
    int i;
    int len;
    u_int mask;
    int shiftamt;
    int val;

    if (seq >= Hsktabsize) {
	/* bad minor frame number */
	return;
    }

    /* stuff the 3 hsk bytes into an u_int */
    h = 0;
    h = ((u_int)h1 << 24) | ((u_int)h2 << 16) | ((u_int)h3 << 8);

    /* search the Hsktab for items to be put in this minor frame */
    hp = &Hsktab[0];
    for (i=0; i<Hsktabsize; i++) {
	if (hp->minor == seq) {
	    len = (hp->stopbit - hp->startbit) + 1;
	    /* assumption that there are 32 bits in a u_int */
	    shiftamt = (32 - hp->stopbit) - 1;
	    mask = (1 << len) - 1;
	    hp->rawval = (h >> shiftamt) & mask;
	} else if (hp->minor > seq) {
	    break;
	}
	hp++;
    }
}

int
sc_tlm_hsk(idx)
    int idx;
{
    if (idx < Hsktabsize) {
	return(Hsktab[idx].rawval);
    } else {
	return -2;
    }
}

int sc_tlm_hsk_trg0rat(void)
{
    return(Hsktab[203].rawval);
}

int sc_tlm_hsk_trg1rat(void)
{
    return(Hsktab[204].rawval);
}

int sc_tlm_hsk_trg0and1(void)
{
    return(Hsktab[205].rawval);
}

void
sc_tlm_hsk_trig(trig0, trig1, trig0and1)
    int *trig0;
    int *trig1;
    int *trig0and1;
{
    *trig0 = sc_tlm_hsk_trg0rat();
    *trig1 = sc_tlm_hsk_trg1rat();
    *trig0and1 = sc_tlm_hsk_trg0and1();
}

int
sc_tlm_hsk_mcp1(void)
{
    return(Hsktab[42].rawval);
}

int
sc_tlm_hsk_mcp2(void)
{
    return(Hsktab[43].rawval);
}

/* sc_tlm_cam_guess - try to figure out which camera is on based on the mcp
 * monitor values. The mcp for the camera that is on should be lower. If
 * you want to convert to a value in kilovolts, the formula is:
 *	kv = (chan * -0.00970047) + 4.9659
 *				[courtesy of Mark Wiedenbeck]
 */
int
sc_tlm_cam_guess(void)
{
    int c1, c2;
    c1 = sc_tlm_hsk_mcp1();
    c2 = sc_tlm_hsk_mcp2();
    if (c1 < c2) {
	return SC_CAM1;
    } else if (c2 < c1) {
	return SC_CAM2;
    } else {
	return -1;
    }
}

void
sc_most_recent_cris_subset(s)
    sc_crisSubset *s;
{
    *s = Cur;
}

void 
sc_most_recent_minor(f)
    struct sc_minorFrame *f;
{
    *f = Curminor;
}

int
sc_tlm_save_minor(v)
    int v;
{
    int old;
    old = Save_last_minor;
    Save_last_minor = v;
    return old;
}

/* do_save_minor - store a minor frame for future reference */
static void
do_save_minor(src, dest)
    struct sc_minorFrame *src, *dest;
{
    if (Save_last_minor) {
	*dest = *src;
    }
}

/* sc_tlm_sec1996 - return the date of a cris subset as number of seconds
 * since 1996. This value can then be manipulated by the various time.h
 * routines such as gmtime(3). Returns -1 if it was passed a NULL pointer. */
long
sc_tlm_sec1996(sc_crisSubset *c)
{
    struct tm t;
    time_t sec1970;
    time_t sec1996;
    time_t sec1996_1970;

    sec1970 = sc_tlm_sec1970(c);
    if (sec1970 == (long)-1) {
	return (long)-1;
    }

    t.tm_year = 96;
    t.tm_mon = 0;
    t.tm_mday = 1;
    t.tm_hour = 0;
    t.tm_min = 0;
    t.tm_sec = 0;
    t.tm_isdst = 0;
    sec1996_1970 = mktime(&t);
    if (sec1996 == (time_t)-1) {
	return (long)-1;
    }
    sec1996 = sec1970 - sec1996-1970;

    /* we ignore leap seconds */
    return ((long)sec1996);
}

/* sc_tlm_sec1970 - return the date of a cris subset as number of seconds
 * since 1970. This value can then be manipulated by the various time.h
 * routines such as gmtime(3). Returns -1 if it was passed a NULL pointer. */
long
sc_tlm_sec1970(c)
    sc_crisSubset *c;
{
    long sec1970;
    if (c == NULL) {
	return ((long)-1);
    }
    sec1970 = ConvertToSeconds1970((c->year % 100)*10000 + 101, 0)
	   + (c->DayOfYear - 1)*86400 + (c->TimeOfDay)/100;
    return sec1970;
}

#ifdef NOTDEF
/* The following routine is here as an example of how to do the date/time
 * conversion ourselves. Based on Bob Radocynski and earlier versions of
 * the rdminor program. */
/* sc_tlm_setdate - set the date elements from the pointed-to cris subset */
void
sc_tlm_setdate(c)
    sc_crisSubset *c;
{
    if (c == NULL) {
	return;
    }
    td_sec1970 = ConvertToSeconds1970((c->year % 100)*10000 + 101, 0)
	   + (c->DayOfYear - 1)*86400 + (c->TimeOfDay)/100;
    ConvertFromSeconds1970(td_sec1970, &td_date, &td_time);
    td_year = td_date/10000;
    td_month = (td_date - 10000*td_year)/100;
    td_day = td_date - td_year*10000 - td_month*100;
    td_hour = td_time/10000;
    td_minute = (td_time - td_hour*10000)/100;
    td_second = td_time - td_hour*10000 - td_minute*100;
}
#endif /* NOTDEF */

/* ConvertFromSeconds1970 - mostly intact from Bob Radocynski */
static void 
ConvertFromSeconds1970(seconds, date, time)
    int seconds;
    int *date;
    int *time;
{
   static int table[61][12];
   int i;
   int j;
   static int first = 1;
   int year;
   int month;
   int day;
   int days;
   int hour;
   int minute;
   int second;

   if (first)
   {
      table[0][ 0] = 0;
      table[0][ 1] = table[0][ 0] + 31;
      table[0][ 2] = table[0][ 1] + 28;
      table[0][ 3] = table[0][ 2] + 31;
      table[0][ 4] = table[0][ 3] + 30;
      table[0][ 5] = table[0][ 4] + 31;
      table[0][ 6] = table[0][ 5] + 30;
      table[0][ 7] = table[0][ 6] + 31;
      table[0][ 8] = table[0][ 7] + 31;
      table[0][ 9] = table[0][ 8] + 30;
      table[0][10] = table[0][ 9] + 31;
      table[0][11] = table[0][10] + 30;
      year = 1970;
      for (i=1; i<=60; i++)
      {
         year++;
         for (j=0; j<12; j++) table[i][j] = table[i-1][j] + 365;
         if (!(year % 4))
         {
            for (j=0; j<12; j++) table[i][j] += 1;
         }
      }
      year = 1970;
      for (i=1; i<=60; i++)
      {
         year++;
         if (!(year % 4))
         {
            table[i][0] -= 1;
            table[i][1] -= 1;
         }
      }
      first = 0;
   }
   days = seconds/86400;
   year = days/365;
   if (days < table[year][11] + 31)
   {
      for (i=11; 0 <= i; i--)
      {
         if (days >= table[year][i])
         {
            day = (days - table[year][i]) + 1;
            month = i+1;
            year += 70;
            if (year >= 100) year -= 100;
            break;
         }
      }
   }
   else
   {
      day = (days - table[year+1][0]) + 1;
      month = 1;
      year += 71;
      if (year >= 100) year -= 100;
   }
   
   hour = (seconds/3600) % 24;
   minute = (seconds/60) % 60;
   second = seconds % 60;

   *date = year*10000 + month*100 + day;
   *time = hour*10000 + minute*100 + second;
}


/* ConvertToSeconds1970 - also from Bob R. */
static int 
ConvertToSeconds1970(date, time)
    int date;
    int time;
{
   static int table[61][12];
   int i;
   int j;
   static int first = 1;
   int year;
   int month;
   int day;
   int days;
   int hour;
   int minute;
   int second;
   int seconds;

   if (first)
   {
      table[0][ 0] = 0;
      table[0][ 1] = table[0][ 0] + 31;
      table[0][ 2] = table[0][ 1] + 28;
      table[0][ 3] = table[0][ 2] + 31;
      table[0][ 4] = table[0][ 3] + 30;
      table[0][ 5] = table[0][ 4] + 31;
      table[0][ 6] = table[0][ 5] + 30;
      table[0][ 7] = table[0][ 6] + 31;
      table[0][ 8] = table[0][ 7] + 31;
      table[0][ 9] = table[0][ 8] + 30;
      table[0][10] = table[0][ 9] + 31;
      table[0][11] = table[0][10] + 30;
      year = 1970;
      for (i=1; i<=60; i++)
      {
         year++;
         for (j=0; j<12; j++) table[i][j] = table[i-1][j] + 365;
         if (!(year % 4))
         {
             for (j=0; j<12; j++) table[i][j] += 1;
         }
      }
      year = 1970;
      for (i=1; i<=60; i++)
      {
         year++;
         if (!(year %4))
         {
             table[i][0] -= 1;
             table[i][1] -= 1;
         }
      }
      first = 0;
   }
   year = date/10000;
   month = (date/100) % 100;
   day = date % 100;
   if (year < 70) year += 100;
   days = table[year-70][month-1] + day - 1;
   hour = time/10000;
   minute = (time/100) % 100;
   second = time % 100;
   seconds = hour*3600 + minute*60 + second;
   return (days*86400 + seconds);
}

/********************************
		Format of CRIS major and minor frames

Byte	Item	Count	Bits	Sum	Description


	A. Major frame = 256 minor frames

0	a	4	8	32	4 byte offset
4	b	128	58x8	xxx	minor frames w/SYNC=0
xxx	c	128	58x8	yyy	minor frames w/SYNC=1



	B. Minor frame = 58 bytes


		Housekeeping bytes

	(Note that these bytes really start at an offset of 4 bytes from
	the start of the data. Those 4 bytes contain the last 4 bytes of
	the previous minor frame.)

0	0	1	1	1	SYNC = 0 for 1st 128 minor frames,
					     = 1 for 2nd 128 minor frames
	1	1	1	2	CMD  = 1 if we have a cmd response
	2	1	1	3	DIAG = 1 for diagnostic mode
	3	1	5	8	housekeeping bits
1	4	1	8	16	more housekeeping bits
2	5	1	8	24	more housekeeping bits


		Command response (if item #1 [CMD bit] is true)

3	6	1	8	8	length L of response (bytes)
4	7	L	Lx8	8+(Lx8)	command echo (ascii chars)


		CRIS normal events (if item #2 [DIAG bit] is false)

	(May be continued from previous minor frame unless this is the
	start of a new major frame. Note that the event may have a
	length of 0, in which case the rest of the current minor frame
	should be skipped.)

3 or
4+L	8	N	X	Y	CRIS Normal events 


		DIAGNOSTIC mode data (if item #2 [DIAG bit] is true)

	(Can extend across minor and major frames. Can interrupt a CRIS
	normal event, which will then resume after Diag mode is done.)

3 or 
4+L	9	N	X	Y	Diagnostic mode data (Normal evt +
					stack data + SOFT image)
********************************/

