/* realpos.c - routines for converting from fiber positions to real
 * 	space positions (in inches)
 * 
 *	Marty Olevitch, Mar 1996
 *
 *	(Joe Klarmann, July 1997: added some comments, new interp[xy]
 *	routines, and error exit from interp[xy])
 */

static char cvsid[] = "$Id: realpos.c,v 1.8 1999/05/27 17:47:09 marty Exp $";

#include <stdio.h>
#include <ctype.h>	/* isdigit */
#include <math.h>	/* atof */
#include <string.h>	/* strtok, strerror */
#include <stdlib.h>	/* malloc */
#include <errno.h>
#include <stdarg.h>

#include "misc.h"
#include "msu.h"
#include "softcris.h"

#define REFPTSFILE	"refpts.soft"

/* H2X_offset and H2Y_offset are corrections by Mike Lijowski, June 10,
 * 1998. These values are used in the sc_fiber_to_real() routine. The
 * values were derived using flight data, which has only camera 2. */
static float H2X_offset = -0.01808;
static float H2Y_offset =  0.02948;

#define NLYR (SC_NLAYERS/2)
#define NROW 22
#define NCOL 23

/* ref_elem - basic structure element for (x,y) of reference points */
static struct ref_elem {
    float x;
    float y;
} pos[NCAM][NLYR][NROW][NCOL];

static int fiber_to_real_init(void);
static int read_ref_points(void);

/* default interpolation method. Can be changed by running the
 * sc_interp_method() routine. Current methods are:
 *
 *	0 = Marty's original simple-minded method
 *	1 = Joe Klarmann's method
 *
 * Be sure to change sc_interp_method() below if another method is added.
 */
static int Interp_method = 1;

static float interpx0(int cam, int layer, int row, int col, float xf);
static float interpy0(int cam, int layer, int row, int col, float yf);

static float interpx1(int cam, int layer, int row, int col, float xf, float yf);
static float interpy1(int cam, int layer, int row, int col, float xf, float yf);

static int Init = 0;		/* nonzero if we've been initialized */
static int Use_inches = 0;	/* nonzero to get inch values, default is mm */

static int
fiber_to_real_init(void)
{
    if (read_ref_points()) {
	return -1;
    }
    Init = 1;
    return 0;
}

/* sc_interp_method - use a different interpolation routine in
 * sc_fiber_to_real(). Returns the previous method id. */
int
sc_interp_method(int m)
{
    int ret = Interp_method;
    if (m != 0 && m != 1) {
	return -1;
    }
    Interp_method = m;
    return (ret);
}

void
sc_fiber_inches(void)
{
    Use_inches = 1;
}

void
sc_fiber_mm(void)
{
    Use_inches = 0;
}

/* sc_fiber_to_real - computes the real coordinates (xi,yi) (in mm, 
 * unless sc_fiber_inches() has been called) given cam, layer and the 
 * fiber coordinates (xf,yf). */
int
sc_fiber_to_real(int c, int l, float xf, float yf, float *xi, float *yi)
{
    int cam;
    int col;
    int layer;
    int row;
    int ret;

    if (!Init) {
	if (fiber_to_real_init()) {
	    fprintf(stderr,
		"sc_fiber_to_real: bad return from  fiber_to_real_init()\n");
	    return -1;
	}
    }

    if (c == 1 || c == 2) {
	cam = c - 1;
    } else {
	fprintf(stderr, "sc_fiber_to_real: cam must be 1 or 2, not %d\n", c);
	return -1;
    }

    if (l == 1 || l == 2 || l == 3) {
	layer = l - 1;
    } else {
	fprintf(stderr, "sc_fiber_to_real: layer must 1, 2, or 3, not %d\n", l);
	return -1;
    }

    /* Now find the right row in the reference point matrix for the 
     * given yf coordinate.  This assumes a perfectly recangualr grid 
     * of reference points which is not true:  The row found in column 
     * zero might be too high or too low when we come to the right 
     * column because of variations in the y coordinates of the grid.
     * In that case, however the yf coordinate is very close to the y
     * cordinate of the ref. point and the interpolation below will 
     * give a good answer anyway. */ 

    for (row=0; row < NROW; row++) {
	if (yf < pos[cam][layer][row][0].y) {
	    break;
	}
    }
    if (row == 0 || row >= NROW) {
#ifdef SOFT_VERBOSE
	fprintf(stderr,"sc_fiber_to_real: (%.2f, %.2f) out of bounds in Y\n", 
	    xf, yf);
#endif
	return -1;
    }

    /* In the selected row search for the proper column to match xf.
     * Same comments as above apply. */

    for (col=0; col < NCOL; col++) {
	if (xf < pos[cam][layer][row][col].x) {
	    break;
	}
    }
    if (col == 0 || col >= NCOL) {
#ifdef SOFT_VERBOSE
	fprintf(stderr, "sc_fiber_to_real: (%.2f, %.2f) out of bound in X\n", 
	    xf, yf);
#endif
	return -1;
    }
    col--;

    /* Interpolate between the reference points, using the chosen
     * interpolation method. */
    switch(Interp_method) {
	case 0:
	    /* Marty's original simple-minded method */
	    *xi = interpx0(cam, layer, row, col, xf);
	    *yi = interpy0(cam, layer, row, col, yf);
	    ret = 0;
	    break;
	case 1:
	    /* Joe's method */
	    *xi = interpx1(cam, layer, row, col, xf, yf);
	    *yi = interpy1(cam, layer, row, col, xf, yf);
	    if (*xi == -1000.0 || *yi == -1000.0) {
		ret = -1;
	    } else {
		ret = 0;
	    }
	    break;
	default:
	    fprintf(stderr, "sc_fiber_to_real: invalid interp method (%d)\n", 
		Interp_method);
	    ret = -1;
	    break;
    }

    if (l == 2) {
	/* H2 layer mean correction */
	*xi += H2X_offset;
	*yi += H2Y_offset;
    }

    return (ret);
}

/* interpx0 - Marty's simple-minded interpolation routine */
static float 
interpx0(int cam, int layer, int row, int col, float xf)
{
    float x0, x1, xdiff, val;
    x0 = pos[cam][layer][row][col].x;
    if (xf == x0) {
	return(xf);
    }
    x1 = pos[cam][layer][row][col+1].x;
    if (x1 == x0) {
	fprintf(stderr, "interpx: x0 = x1 = %.2f (should not happen)\n", x1);
	return(xf);
    }
    xdiff = (xf - x0) / (x1 - x0);
    if (Use_inches) {
	val = (col + xdiff) / 2.0;	/* div by 2 to get inch units */
    } else {
	val = (col + xdiff) * 12.7;	/* mul by (25.4 / 2) to get mm */
    }
    return(val);
}

/* interpx0 - Marty's simple-minded interpolation routine */
static float 
interpy0(int cam, int layer, int row, int col, float yf)
{
    float y0, y1, ydiff, val;
    y0 = pos[cam][layer][row-1][col].y;
    if (y0 == yf) {
	return(yf);
    }
    y1 = pos[cam][layer][row][col].y;
    if (y1 == y0) {
	fprintf(stderr, "interpy: y0 = y1 = %.2f (should not happen)\n", y1);
	return(yf);
    }
    ydiff = (yf - y0) / (y1 - y0);
    if (Use_inches) {
	val = ((row-1) + ydiff) / 2.0;	/* div by 2 to get inch units */
    } else {
	val = ((row-1) + ydiff) * 12.7;	/* mul by (25.4 / 2) to get mm */
    }
    return(val);
}

/* interpx1 - interpolates between the nearest 4 reference points to 
 * find the proper x coordinate. Returns the bogus value -1000.0 for
 * certain errors. (jkl) */
static float 
interpx1(int cam, int layer, int row, int col, float xf, float yf)
{
    float val;
    float x00, x01, x10, x11; 
    float xdiff, xdiff0, xdiff1; 
    float y0, y1;

    /* Find the fractional distance that xf is between the ref points
     * along the row which is either below, or very close to yf.
     * Because of the selection of the row and column, the fractions 
     * resulting may be slightly negative or slightly larger than 1!! */

    x00 = pos[cam][layer][row-1][col].x;
    x01 = pos[cam][layer][row-1][col+1].x;
    if (x01 == x00) {
	fprintf(stderr,"interpx1: x00 = x01 = %.2f (should not happen)\n", x01);
	return(-1000.0);
    }
    xdiff0 = (xf - x00) / (x01 - x00);

    /* Find the fractional distance along the row above or 
     * very close to yf */

    x10 = pos[cam][layer][row][col].x;
    x11 = pos[cam][layer][row][col+1].x;
    if (x11 == x10) {
	fprintf(stderr, "interpx: x10 = x11 = %.2f (should not happen)\n", x10);
	return(-1000.0);
    }
    xdiff1 = (xf - x10) / (x11 - x10);

    /* The final fractional distance is computed from the lower and upper
     * results by weighting them according to their y distance from the
     * ref. points along the left hand column (arbitrary). */

    y0 = pos[cam][layer][row-1][col].y;
    y1 = pos[cam][layer][row][col].y;
    if (y1 == y0) {
	fprintf(stderr, "interpx: y0 = y1 = %.2f (should not happen)\n", y1);
	return(-1000.0);
    }
    xdiff = (yf-y0)/(y1-y0) * xdiff1 + (y1-yf)/(y1-y0) * xdiff0;
    if (Use_inches) {
	val = (col + xdiff) / 2.0;	/* div by 2 to get inch units */
    } else {
	val = (col + xdiff) * 12.7;	/* mul by (25.4 / 2) to get mm */
    }
    return(val);
}


/* interpy - interpolates between the nearest 4 reference points to 
 * find the proper y coordinate. Returns the bogus value -1000.0 for
 * certain errors. (jkl) */
static float 
interpy1(int cam, int layer, int row, int col, float xf, float yf)
{
    float val;
    float y00, y01, y10, y11;
    float ydiff, ydiff0, ydiff1;
    float x0, x1;

    /* Find the fractional distance that yf is between the ref points
     * along the column which is either left of, or very close to xf.
     * Because of the selection of the row and column, the fractions
     * resulting may be slightly negative or slightly larger than 1!! */

    y00 = pos[cam][layer][row-1][col].y;
    y01 = pos[cam][layer][row][col].y;
    if (y01 == y00) {
	fprintf(stderr, "interpy: y00 = y01 = %.2f (should not happen)\n", y01);
	return(-1000.0);
    }
    ydiff0 = (yf - y00) / (y01 - y00);

    /* Find the fractional distance along the column right of or 
     * very close to xf */

    y10 = pos[cam][layer][row-1][col+1].y;
    y11 = pos[cam][layer][row][col+1].y;
    if (y11 == y10) {
	fprintf(stderr, "interpy: y10 = y11 = %.2f (should not happen)\n", y11);
	return(-1000.0);
    }
    ydiff1 = (yf - y10) / (y11 - y10);

    /* The final fractional distance is computed from the left and right
     * results by weighting them according to their x distance from the
     * ref. points along the upper row (arbitrary). */

    x0 = pos[cam][layer][row][col].x;
    x1 = pos[cam][layer][row][col+1].x;
    if (x1 == x0) {
	fprintf(stderr, "interpy: x0 = x1 = %.2f (should not happen)\n", x1);
	return(-1000.0);
    }
    ydiff = (xf-x0)/(x1-x0) * ydiff1 + (x1-xf)/(x1-x0) * ydiff0;
    if (Use_inches) {
	val = (row-1 + ydiff) / 2.0;	/* div by 2 to get inch units */
    } else {
	val = (row-1 + ydiff) * 12.7;	/* mul by (25.4 / 2) to get mm */
    }
    return(val);
}

static int
read_ref_points(void)
{
    int cam;
    int col;
    char *cp;
    int expecting_data;
    FILE *fp;
    int layer;
    int refnamelen;
    int lineno;
    char *refname;
    int ret;
    int row;
#define TLEN 80
    char tmpstr[TLEN];
    char xstr[TLEN];
    char ystr[TLEN];

    expecting_data = 0;
    lineno = 0;

    refnamelen = strlen(MAPDIR) + strlen(REFPTSFILE) + 1;
    refname = malloc(refnamelen);
    if (refname == NULL) {
	fprintf(stderr,
	    "read_ref_points: can't malloc %d bytes for file name!",
	    refnamelen);
	return -1;
    }
#ifndef _WINDOWS
    sprintf(refname, "%s/%s", MAPDIR, REFPTSFILE);
#else
    sprintf(refname, "%s\\%s", MAPDIR, REFPTSFILE);
#endif /* _WINDOWS */

    fp = fopen(refname, "r");
    if (fp == (FILE *)NULL) {
	fprintf(stderr, "read_ref_points: can't open '%s' (%s)\n",
	    refname, strerror(errno));
	return -1;
    }

    while (fgets(tmpstr, TLEN, fp) != NULL) {
	lineno++;

	/* trim leading white space */
	for (cp=tmpstr; *cp == ' ' || *cp == '\t'; cp++) {
	    ;
	}

	if (*cp == '#') {
	    continue;	/* comment */
	}

	ret = split(cp, " \t", 2, xstr, ystr);
	if (SOFT_STREQ(xstr, "cam")) {
	    if (expecting_data) {
		fprintf(stderr, "%s: line #%d: got 'cam', expecting data\n",
		    refname, lineno);
		return -1;
	    }
	    ret = atoi(ystr);
	    if (ret == 1 || ret == 2) {
		cam = ret - 1;
	    } else {
		fprintf(stderr, "%s: line #%d: cam must be 1 or 2, not %d\n", 
		    refname, lineno, ret);
		return -1;
	    }
	} else if (SOFT_STREQ(xstr, "layer")) {
	    if (expecting_data) {
		fprintf(stderr, "%s: line #%d: got 'layer', expecting data\n",
		    refname, lineno);
		return -1;
	    }
	    ret = atoi(ystr);
	    if (ret == 1 || ret == 2 || ret == 3) {
		layer = ret - 1;
	    } else {
		fprintf(stderr,
		    "%s: line #%d: layer must be 1, 2, or 3, not %d\n\n",
		    refname, lineno, ret);
		return -1;
	    }
	    expecting_data = 1;
	    row = col = 0;
	} else {
	    if (!expecting_data) {
		fprintf(stderr,
		    "%s: line #%d: expecting 'cam' or 'layer', got '%s'\n",
		    refname, lineno, cp);
		return -1;
	    }
	    if (!isdigit(xstr[0]) && xstr[0] != '-') {
		fprintf(stderr, "%s: line #%d: bogus data '%s'\n",
		    refname, lineno, xstr);
		return -1;
	    }
	    pos[cam][layer][row][col].x = atof(xstr);
	    if (!isdigit(ystr[0]) && ystr[0] != '-') {
		fprintf(stderr, "%s: line #%d: bogus data '%s'\n",
		    refname, lineno, ystr);
		return -1;
	    }
	    pos[cam][layer][row][col].y = atof(ystr);
	    col++;
	    if (col >= NCOL) {
		col = 0;
		row++;
		if (row >= NROW) {
		    row = 0;
		    expecting_data = 0;
		}
	    }
	}
    }
    return 0;
}

