/* softcris - a library of routines for dealing with soft and cris data
 *
 *  Marty Olevitch, Jan 1994, Nov 1996
 */

static char cvsid[] = "$Id: softcris.c,v 1.1.1.1 1997/03/14 15:53:04 marty Exp $";

#include <stdio.h>

#include "softcris.h"
#include "msu.h"

#ifdef MAXDATA
#undef MAXDATA
#endif
#define MAXDATA 200000

static FILE *Infp;	/* input file */
static FILE *Errfp;	/* error messages file */

static int Verbose;	/* say much if nonzero */
static u_char *Db;	/* data buffer */
static int Init;	/* non-zero if we have successfully initialized */
static u_char *Curvid;
static u_char *Curvid2;
static char Progname[100];

static int chkinit(char *s);
static u_short calcxy(u_short id, u_short xy1, u_short xy2);
static u_long get_long(u_char *cp);

/* sc_msu_init - initialize the library: set Infp and Errfp to the open file
 * pointers passed in as infile and errfile. Allocate space for data buffer. */
int
sc_msu_init(FILE *infile, FILE *errfile, int verb, char *progname)
    /* infile: ptr to opened input file */
    /* errfile: ptr to opened error file */
    /* verb: produce more voluminous error or debug messages */
    /* progname: string to identify the caller of these routines */
{
    Verbose = (verb > 0) ? 1 : 0;
    if (progname == NULL) {
	Progname[0] = '\0';
    } else {
	strncpy(Progname, progname, 100);
    }

    /* infile and errfile must be FILE pointers to already opened files, or
     * NULL if we are to use stdin or stderr */
    Infp = (infile != NULL) ? infile : stdin;
    Errfp = (errfile != NULL) ? errfile : stderr;
    Db = (u_char *)malloc((unsigned)MAXDATA);
    if (Db == NULL) {
	fprintf(Errfp, "%s: sc_msu_init: can't allocate data buffer\n",
		Progname);
	Init = 0;
	return -1;
    }
    Init = 1;
    return 0;
}

/* sc_msu_finito - free memory, reset Init flag, Infp and Errfp. After this, a
 * new file can be opened for input using sc_msu_init() again. */
int
sc_msu_finito(void)
{
    if (Init != 1) {
	fprintf(stderr,"sc_msu_finito: we haven't been initialized!\n");
	return -1;
    }
    Init = 0;
    free(Db);
    Infp = stdin;
    Errfp = stderr;
    return 0;
}

/* soft_next_event - try to find the start of the next event. Return 0 if
 * all went well, else -1. Fills in the pointed-to sc_msuEvt struct.
 */
int
sc_msu_next_evt(struct sc_msuEvt *e)
{
    u_char *cp;
    int discsiz;
    int i;
    int phasiz;
    int replysiz;
    int ret;
    u_short stmp;

    e->ptr = NULL;
    e->valid = SC_MSU_INVALID;
    
    if (chkinit("sc_msu_next_evt"))
	return -1;

    /* Eat byte until we get a HDR_HDR */
    while (1) {
	if ((ret = fread(Db, 1, 1, Infp)) != 1)  {
	    if (ferror(Infp)) {
		fprintf(Errfp, "%s: sc_msu_next_evt: bad read (1)\n",
			Progname);
	    }
	    return -1;
	}
	if (Db[0] == ALIGN_FILL)
	    continue;
	if ((ret = fread(Db+1, 1, 1, Infp)) != 1)  {
	    if (ferror(Infp)) {
		fprintf(Errfp, "%s: sc_msu_next_evt: bad read (2)\n",
			Progname);
	    }
	    return -1;	/* EOF or error */
	}
	cp = Db;
	stmp = *(u_short *)cp;
	if (stmp != HDR_HDR)
	    continue;
	break;
    }
    if ((ret = fread(Db+2, 1, 2, Infp)) != 2)  {
	if (ferror(Infp)) {
	    fprintf(Errfp, "%s: sc_msu_next_evt: bad read (3)\n",
		    Progname);
	}
	return -1;		/* EOF or error */
    }
    cp = GET_EVT_HDR(Db);
    stmp = *(u_short *)cp;
    if (stmp == SPILL_HDR) {
	if ((ret = fread(Db+4, 1, 4, Infp)) != 4)  {
	    if (ferror(Infp)) {
		fprintf(Errfp, "%s: sc_msu_next_evt: bad read (4)\n",
			Progname);
	    }
	    return -1;	/* EOF or error */
	}
	cp = GET_spill(Db);
	e->spill = *(u_long *)cp;
	e->valid = SC_MSU_SPILL;
    } else if (stmp == EVT_HDR) {
	/* read in the STD stuff */
	ret = fread(Db+4, 1, STDSIZ - 4, Infp);
	if (ret != STDSIZ - 4) {
	    if (ferror(Infp)) {
		fprintf(Errfp, "%s: sc_msu_next_evt: bad read (5)\n",
			Progname);
	    }
	    return -1;	/* EOF or error */
	}
	cp = GET_type(Db);
	e->type = *(u_short *)cp;
	cp = GET_exp_id(Db);
	e->exp_id = *(u_short *)cp;
	cp = GET_size(Db);
	e->size = *(u_long *)cp;
	cp = GET_evtno(Db);
	e->evtno = *(u_long *)cp;
	cp = GET_clock(Db);
	e->clock = *(u_long *)cp;
	if (e->exp_id == 6) {
	    cp = GET_corr_abs(Db);
	    e->abs = *(u_long *)cp;
	    cp = GET_absevtno(Db);
	    e->abs_orig = *(u_long *)cp;
	} else {
	    cp = GET_absevtno(Db);
	    e->abs = *(u_long *)cp;
	}
	cp = GET_vidsize(Db);
	e->vidsize = *(u_long *)cp;
	if (e->exp_id >= 5) {
	    phasiz = PHASIZ_5;
	    discsiz = DISCSIZ_5;
	    replysiz = REPLYSIZ_5;
	    cp = GET_vidsize2(Db);
	    e->vidsize2 = *(u_long *)cp;
	} else {
	    phasiz = PHASIZ;
	    discsiz = DISCSIZ;
	    replysiz = REPLYSIZ;
	    e->vidsize2 = 0L;
	}

	/* read in the rest of the evt */
	ret = fread(Db+STDSIZ, 1, e->size-STDSIZ, Infp);
	if (ret != e->size - STDSIZ) {
	    if (ferror(Infp)) {
		fprintf(Errfp, "%s: sc_msu_next_evt: bad read (5)\n",
			Progname);
	    }
	    return -1;	/* EOF or error */
	}

	switch (e->type) {
	case T_PHA_ONLY:
	    /* don't do anything, but this is a valid event type. Maybe in
	     * the future it will be possible to look at all th pha data. */
	     break;
	case T_REAL:
	    cp = GET_PHA_HDR(Db);
	    if (*(u_short *)cp != HDR_HDR || *(u_short *)(cp+2) != PHA_HDR) {
		fprintf(Errfp,
			"%s: sc_msu_next_evt: bad HDR_HDR PHA_HDR (abs: %ld)\n",
			Progname, e->abs);
		return -1;
	    }
	    cp = GET_phas(Db);
	    if (e->exp_id == 1 || e->exp_id == 2 || e->exp_id == 3) {
		/* 1 = MSU Dec 93; 2 = post MSU Dec 93 merged frame;
		 * 3 = pre MSU Fall 94 */
		e->phas.s1_raw = *(u_short *)cp;
		e->phas.x1_raw = *(u_short *)(cp+2);
		e->phas.x2_raw = *(u_short *)(cp+4);
		e->phas.y1_raw = *(u_short *)(cp+6);
		e->phas.y2_raw = *(u_short *)(cp+8);
		e->phas.pd1_raw = *(u_short *)(cp+10);
		e->phas.pd2_raw = *(u_short *)(cp+12);
		e->phas.par = *(u_short *)(cp+14);
	    } else if (e->exp_id >= 5) {
		/* id 5 = MSU Nov 95 */
		e->phas.s1_raw = *(u_short *)cp;
		e->phas.x1_raw = *(u_short *)(cp+2);
		e->phas.x2_raw = *(u_short *)(cp+4);
		e->phas.y1_raw = *(u_short *)(cp+6);
		e->phas.y2_raw = *(u_short *)(cp+8);
		e->phas.s2_raw = *(u_short *)(cp+10);
		e->phas.x1b_raw = *(u_short *)(cp+12);
		e->phas.x2b_raw = *(u_short *)(cp+14);
		e->phas.y1b_raw = *(u_short *)(cp+16);
		e->phas.y2b_raw = *(u_short *)(cp+18);
		e->phas.par = *(u_short *)(cp+20);
	    } else if (e->exp_id == 0) {
		/* MSU Jun 93 */
		e->phas.s1_raw = *(u_short *)cp;
		e->phas.pd1_raw = *(u_short *)(cp+2);
		e->phas.pd2_raw = *(u_short *)(cp+4);
		e->phas.x1_raw = *(u_short *)(cp+6);
		e->phas.x2_raw = *(u_short *)(cp+8);
		e->phas.y1_raw = *(u_short *)(cp+10);
		e->phas.y2_raw = *(u_short *)(cp+12);
		e->phas.par = *(u_short *)(cp+14);
	    } else {
		fprintf(Errfp, "%s: sc_msu_next_evt: bad exp id (%d)\n",
			Progname, e->exp_id);
		e->phas.s1_raw = e->phas.x1_raw = 
		  e->phas.x2_raw = e->phas.y1_raw = 
		  e->phas.y2_raw = e->phas.pd1_raw = 
		  e->phas.pd2_raw = e->phas.par = 0;
	    }
	    e->phas.s1 = e->phas.s1_raw & 0xfff;
	    e->phas.x1 = e->phas.x1_raw & 0xfff;
	    e->phas.x2 = e->phas.x2_raw & 0xfff;
	    e->phas.y1 = e->phas.y1_raw & 0xfff;
	    e->phas.y2 = e->phas.y2_raw & 0xfff;
	    e->phas.x = calcxy(e->exp_id, e->phas.x1, e->phas.x2);
	    e->phas.y = calcxy(e->exp_id, e->phas.y1, e->phas.y2);

	    if (e->exp_id < 5) {
		e->phas.pd1 = e->phas.pd1_raw & 0xfff;
		e->phas.pd2 = e->phas.pd2_raw & 0xfff;
	    } else {
		e->phas.s2 = e->phas.s2_raw & 0xfff;
		e->phas.x1b = e->phas.x1b_raw & 0xfff;
		e->phas.x2b = e->phas.x2b_raw & 0xfff;
		e->phas.y1b = e->phas.y1b_raw & 0xfff;
		e->phas.y2b = e->phas.y2b_raw & 0xfff;
		e->phas.xb = calcxy(e->exp_id, e->phas.x1b, e->phas.x2b);
		e->phas.yb = calcxy(e->exp_id, e->phas.y1b, e->phas.y2b);
	    }

	    cp = GET_DISC_HDR(Db, phasiz);
	    if (*(u_short *)cp != HDR_HDR || *(u_short *)(cp+2) != DISC_HDR) {
		fprintf(Errfp,
			"%s: sc_msu_next_evt: bad HDR_HDR DISC_HDR (abs: %ld)\n",
			Progname, e->abs);
		return -1;
	    }
	    cp = GET_disc(Db, phasiz);
	    if (e->exp_id < 5) {
		cp++;
		e->disc.camdisc = *cp;
		e->disc.trig1 = *(cp + 2);
		e->disc.trig2 = *(cp + 4);
		e->disc.iigain = *(cp + 6);
		e->disc.coinen = *(cp + 8);
		e->disc.cointhr1 = *(cp + 10);
		e->disc.cointhr2 = *(cp + 11);
		e->disc.cointhr3 = *(cp + 12);
	    } else {
		e->disc.camdisc = *cp;
		e->disc.trig1 = *(cp + 1);
		e->disc.trig2 = *(cp + 2);
		e->disc.iigain = *(cp + 3);
		e->disc.camdiscb = *(cp + 4);
		e->disc.trig1b = *(cp + 5);
		e->disc.trig2b = *(cp + 6);
		e->disc.iigainb = *(cp + 7);
		e->disc.coinen = *(cp + 8);
		e->disc.cointhr1 = *(cp + 9);
		e->disc.cointhr2 = *(cp + 10);
		e->disc.cointhr3 = *(cp + 11);
		e->disc.scalars[0] = *(u_short *)(cp + 12);
		e->disc.scalars[1] = *(u_short *)(cp + 14);
		e->disc.scalars[2] = *(u_short *)(cp + 16);
		e->disc.scalars[3] = *(u_short *)(cp + 18);
		e->disc.scalars[4] = *(u_short *)(cp + 20);
		e->disc.scalars[5] = *(u_short *)(cp + 22);
	    }

	    cp = GET_VID_HDR(Db, phasiz, discsiz);
	    if (*(u_short *)cp != HDR_HDR || *(u_short *)(cp+2) != VID_HDR) {
		fprintf(Errfp,
			"%s: sc_msu_next_evt: bad HDR_HDR VID_HDR (abs: %ld)\n",
			Progname, e->abs);
		return -1;
	    }
	    cp = GET_vid(Db, phasiz, discsiz);
	    e->vidptr = cp;
	    Curvid = cp;
	    if (*(u_short *)Curvid != SC_START_HDR) {
		fprintf(Errfp,"%s:sc_msu_next_evt:no SC_START_HDR (abs: %ld)\n",
			Progname, e->abs);
		Curvid = NULL;
	    }
	    
	    cp += e->vidsize;


	    if (e->exp_id >= 5) {
		if (*(u_short *)cp != HDR_HDR ||
			*(u_short *)(cp+2) != VID_HDR2) {
		    fprintf(Errfp,
			    "%s: sc_msu_next_evt: bad HDR_HDR VID_HDR2 (abs: %ld)\n",
			    Progname, e->abs);
		    return -1;
		}
		cp += sizeof(u_short);	/* skip HDR_HDR */
		cp += sizeof(u_short);	/* skip VID_HDR2 */
		e->vidptr2 = cp;
		Curvid2 = cp;
		if (*(u_short *)Curvid2 != SC_START_HDR) {
		    fprintf(Errfp, "%s: sc_msu_next_evt: no START_HDR (abs: %ld)\n",
			    Progname, e->abs);
		    Curvid2 = NULL;
		}
    		cp += e->vidsize2;
	    }

	    if (e->exp_id < 5) {
		/* experiments before id 5 must have blob headers */
		if (*(u_short *)cp != HDR_HDR || *(u_short *)(cp+2) != BLOB_HDR) {
		    fprintf(Errfp,
			    "%s: sc_msu_next_evt: bad HDR_HDR BLOB_HDR (abs: %ld)\n",
			    Progname, e->abs);
		    return -1;
		}
		cp += 4;
		e->nblobs = *(u_short *)cp;
		cp += 2;
		for (i=0; i < e->nblobs; i++) {
		    e->blobx[i] = *(u_short *)cp;
		    cp += sizeof(u_short);
		    e->bloby[i] = *(u_short *)cp;
		    cp += sizeof(u_short);
		    if (e->exp_id == 3) {
			/* pre-msu-fall-94 has blob intensities in a long */
			e->blobI[i] = get_long(cp);
			cp += sizeof(u_long);
		    }
#ifdef NOTDEF
		    /* This is such an amazing bug, I can't bear to remove
		     * these lines. They actually produced correct output,
		     * but only worked by luck (or maybe hasgacha pratis)! */
		    e->blobx[i] = *(u_short *)(cp + (i * 2));
		    e->bloby[i] = *(u_short *)(cp + ((i * 2) + 2));
		    cp += 2;
#endif
		}
	    } else {
		/* exp id >= 5: blob headers may or may not be present, but
		 * if vid sys 1 has them, so does vid sys 2 */

		if (*(u_short *)cp == HDR_HDR &&
			*(u_short *)(cp+2) == BLOB_HDR) {
		    cp += 4;
		    e->nblobs = *(u_short *)cp;
		    cp += 2;
		    for (i=0; i < e->nblobs; i++) {
			e->blobx[i] = *(u_short *)cp;
			cp += sizeof(u_short);
			e->bloby[i] = *(u_short *)cp;
			cp += sizeof(u_short);
			e->blobI[i] = get_long(cp);
			cp += sizeof(u_long);
		    }
		    /* now the blobs for vid sys 2 */
		    if (*(u_short *)cp != HDR_HDR ||
			    *(u_short *)(cp+2) != BLOB_HDR) {
			fprintf(Errfp,
				"%s: sc_msu_next_evt: bad HDR_HDR BLOB_HDR vid sys 2 (abs: %ld)\n",
				Progname, e->abs);
			return -1;
		    }
		    cp += 4;
		    e->nblobsb = *(u_short *)cp;
		    cp += 2;
		    for (i=0; i < e->nblobsb; i++) {
			e->blobxb[i] = *(u_short *)cp;
			cp += sizeof(u_short);
			e->blobyb[i] = *(u_short *)cp;
			cp += sizeof(u_short);
			e->blobIb[i] = get_long(cp);
			cp += sizeof(u_long);
		    }
		}
	    }
	    break;
	case T_REPLY:
	    cp = GET_REPLY_HDR(Db);
	    if (*(u_short *)cp != HDR_HDR || *(u_short *)(cp+2) != REPLY_HDR) {
		fprintf(Errfp,
			"%s: sc_msu_next_evt: bad HDR_HDR REPLY_HDR (abs: %ld)\n",
			Progname, e->abs);
		return -1;
	    }
	    cp = GET_reply(Db);
	    e->reply = (char *)cp;
	    break;
	default:
	    fprintf(Errfp, "%s: sc_msu_next_evt: unknown event type (%u)\n",
		    Progname, e->type);
	    return -1;
	    break;
	}
	e->valid = SC_MSU_EVT;
    }
    e->ptr = Db;
    return 0;
}

static int
chkinit(char *s)
{
    if (Init == 0) {
	fprintf(stderr, "%s: %s: not initialized (run sc_msu_init)\n",
		Progname, s);
	return -1;
    }
    return 0;
}

int
sc_msu_next_pix(u_short *xp, u_short *yp, u_short *valp, int vidsys)
{
    u_char *curvidp;
    u_short stmp;
    static u_short curx, cury, d;
    int gotdata = 0;
    
    if (chkinit("sc_msu_next_pix")) {
	return -1;
    }

    if (vidsys == 1) {
	curvidp = Curvid;
    } else if (vidsys == 2) {
	curvidp = Curvid2;
    } else {
	fprintf(Errfp, "sc_msu_next_pix: bad vidsys = %d (must be 1 or 2)\n",
		vidsys);
	return -1;
    }

    if (curvidp == NULL) {
	return -1;
    }
    while (!gotdata) {
	stmp = *(u_short *)curvidp;
	if (stmp == SC_END_HDR) {
	    return 0;
	}
	if (SC_IS_VIDEO_HDR(stmp)) {
	    if (SC_IS_Y_HDR(stmp)) {
		cury = SC_Y_VAL(stmp);
	    } else if (SC_IS_X_HDR(stmp)) {
		curx = SC_X_VAL(stmp);
	    }
	} else {
	    d = SC_VID_DATA(stmp);
	    gotdata = 1;
	}
	curvidp += sizeof(u_short);
    }
    *xp = curx;
    curx++;
    *yp = cury;
    *valp = d;
    if (vidsys == 1) {
	Curvid = curvidp;
    } else {
	Curvid2 = curvidp;
    }
    return 1;
}

static u_short
calcxy(u_short id, u_short xy1, u_short xy2)
{
    float denom;
    float xyfactor;
 
   if (id == 1 || id == 2 || id == 3 || id == 5) {
	/* (1)msu-dec-93; (2)post-msu-dec-93-merged-frm; (3)pre-msu-fall-94 */
	xyfactor = 165.0;
    } else if (id == 0) {
	/* (0)msu jun 93 */
	xyfactor = 400.0;
    } else {
	return 0;
    }
    
    denom = xy1 + xy2;
    if (denom == 0.0) {
	return 0;
    }
    return (u_short)(xyfactor * (xy1 / denom));
}

/* sc_soft_img_convert - take an array of u_shorts representing the values
 * in a "raw" video image and convert it to SOFT format. It is assumed that
 * rawp points to an array of SC_SOFT_HEIGHTxSC_SOFT_WIDTH u_shorts and
 * that vidp points to enough memory to hold the converted image. The size
 * of the converted video data (including START_HDR and END_HDR) is returned */
long
sc_soft_img_convert(u_short *rawp, u_short *vidp)
{
    u_short *yp;    /* ptr to start of y row */
    u_short *p;	    /* ptr into pixel data */
    u_short *op;    /* ptr into vidp */
    int inseg;	    /* nonzero if we are in a pixel segment */
    u_short x;
    u_short y;
    long size;

    inseg = 0;
    op = vidp;
    *op++ = SC_START_HDR;
    for (y=0; y<SC_SOFT_HEIGHT; y++) {
	yp = rawp + (y * SC_SOFT_WIDTH);
	for (x=0; x<SC_SOFT_WIDTH; x++){
	    p = yp + x;
	    if (*p > 0) {
		if (inseg == 0) {
		    /* start of new pixel segment - emit Y and X_HDRs */
		    *op++ = SC_Y_HDR | (y & 0x0ff);
		    *op++ = SC_X_HDR | (x & 0x3ff);
		    inseg = 1;
		}
		*op++ = *p & 0x0fff;
	    } else {
		/* possible end of pixel segment */
		inseg = 0;
	    }
	}
    }
    *op++ = SC_END_HDR;
    size = (op - vidp) * sizeof(u_short);
    return (size);
}

/* get_long - return the long value found at the pointer cp. This is needed
 * because of data alignment problems. The long may not be aligned on 4-byte(?)
 * boundaries. */
static u_long
get_long(u_char *cp)
{
    u_long ltmp;
    u_char *cp2;
    int i;
    
    cp2 = (u_char *)&ltmp;
    for (i=0; i<sizeof(u_long); i++) {
	*cp2++ = *cp++;
    }
    return (ltmp);
}

/* sc_soft_next_pix - return pointers to (x,y,I) of the next pixel in the
 * soft image. If imgp == NULL, it will continue with the current image.
 * Otherwise, imgp should point to the START_HDR of a new soft image.
 * Returns 1 if it found a valid pixel, 0 for no more pixels in the image,
 * and -1 if there is some kind of problem. */
int
sc_soft_next_pix(xp, yp, Ip, imgp)
    u_short *xp;
    u_short *yp;
    u_short *Ip;
    u_short *imgp;
{
    static u_short curx, cury, d;
    static u_short *curvidp = NULL;
    int gotdata = 0;
    static int maxsize;
    u_short stmp;

    if (imgp != NULL) {
	curvidp = imgp;
	maxsize = SC_SOFT_MAXSIZE;
    }

    if (curvidp == NULL) {
	return -1;
    }

    while (!gotdata) {
	stmp = *curvidp;
	if (stmp == SC_END_HDR) {
	    return 0;
	}
	if (SC_IS_VIDEO_HDR(stmp)) {
	    if (SC_IS_Y_HDR(stmp)) {
		cury = SC_Y_VAL(stmp);
	    } else if (SC_IS_X_HDR(stmp)) {
		curx = SC_X_VAL(stmp);
	    }
	} else {
	    d = SC_VID_DATA(stmp);
	    gotdata = 1;
	}
	curvidp++;
	if (maxsize-- <= 0) {
	    /* evt too big */
	    return -1;
	}
    }

    *xp = curx;
    curx++;
    *yp = cury;
    *Ip = d;
    return 1;
}
