
/* traj.c - routines dealing with calculating trajectories, finding
 * coordinates, and sorting blobs by intensity.
 *
 *	Marty Olevitch, Oct 1996
 */

static char cvsid[] = "$Id: traj.c,v 1.52 1999/05/26 19:42:06 marty Exp $";

/* _TRAJ_C_QSORT_QUIRK - kludge to avoid #include'ing <L1.h> inside
 * "softcris.h" to avoid #include'ing <stdlib.h> which causes inappropriate
 * compiler warnings regarding qsort() */
#define _TRAJ_C_QSORT_QUIRK	

#include <math.h>
#include "softcris.h"
#include "traj.h"

#ifdef SQR
#undef SQR
#endif
#define SQR(x) 	((x)*(x))

/* ONE_HOP causes trajectories with more than one hop to be rejected. */
/* #define ONE_HOP */

/* MAXTRACKS limits the number of three blob combinations to yield a track
 * in the x or y direction in multitraj. tracks differ from trajectories in
 * that they are one dimensional. */
#define MAXTRACKS	75
static int Maxtracks = MAXTRACKS;

/* MAXTRAJ limits the number of trajectories in multitraj that may be
 * combined from different tracks. Trajectories are of course two
 * dimensional. */
#define MAXTRAJ		20
static int Maxtraj = MAXTRAJ;

/* These values of SC_Z were corrected by Mike Lijowski, June 10, 1998.
 * The Y2 and X2 values are different. */
static float SC_Z[] = {
    0.102,      /* Y3 */
    0.305,      /* X3 */  
    32.950,     /* Y2 */
    33.150,     /* X2 */
    72.110,     /* Y1 */
    72.310,     /* X1 */
}; 

#ifdef NOTDEF
/* These are the values in mm using bottom of layer Y3 as the Z=0 origin. */
static float SC_Z[] = {
    0.102,	/* Y3 */
    0.305,	/* X3 */
    33.12,	/* Y2 */
    33.33,	/* X2 */
    72.11,	/* Y1 */
    72.31,	/* X1 */
};
#endif

#ifdef NOTDEF
/* Don't use this version: we want the Z=0 to be the bottom of layer Y3. */

/* The SC_Z array gives the Z-coordinate in mm for each of the fiber layers
 * using the top of the E1 detector as the Z=0 origin. */
static float SC_Z[] = {
    1.682,      /* Y3 */
    1.885,      /* X3 */
    35.11,      /* Y2 */
    35.31,      /* X2 */
    74.503,     /* Y1 */
    74.707,     /* X1 */
};
#endif /* NOTDEF */

/* Soft_stack_sep is the distance in mm from the top of the stack detectors to
 * the bottom of hodo layer Y3 (Z=0 origin). */
static float Soft_stack_sep = 15.9;

/* Stack_sep - the separation in mm between each layer of the stack From
 * Brian Dougherty's paper of 10 Jan 1997. */
static float Stack_sep[] = {
        0,	/* E1 */
     -4.1,	/* E2 */
    -10.5,	/* E3_1 */
    -14.0,	/* E3_2 */
    -10.5,	/* E3 (same as E3_1) */
    -18.8,	/* E4_1 */
    -22.2,	/* E4_2 */
    -18.8,	/* E4 (same as E4_1) */
    -29.8,	/* E5_1 */
    -33.3,	/* E5_2 */
    -29.8,	/* E5 (same as E5_1) */
    -38.1,	/* E6_1 */
    -41.5,	/* E6_2 */
    -38.1,	/* E6 (same as E6_1) */
    -48.5,	/* E7_1 */
    -52.0,	/* E7_2 */
    -48.5,	/* E7 (same as E7_1) */
    -56.8,	/* E8_1 */
    -60.2,	/* E8_2 */
    -56.8,	/* E8 (same as E8_1) */
    -64.4,	/* E9 */
};
#define NSTACKLAYER	(sizeof(Stack_sep)/sizeof(Stack_sep[0]))

static float SC_STACK_Z[NSTACKLAYER];
static void set_stack_z(void);

/* blobhodo structures are used in the msu & raw blob sorting routines */
struct blhodo {
    short x;
    short y;
    unsigned long I;
};
static struct blhodo *blobhodo = NULL;

/* raw_evt - this struct is used to pass an array of x, y, and intensity to
 * multitraj from sc_raw_traj(). */
struct raw_evt {
    int *x;
    int *y;
    long *I;
    int nblobs;
};

/* These macros are used for the 'evttype' parameter to multitraj. */
#define ISCRIS	1
#define ISMSU	2
#define ISRAW	3

/* [Comment by Joe Klarmann, 1 Sep 1998]
 * Explanatory note on the following methods:
 *
 * Method zero returns the trajectory based on the brightest blob in each
 * layer.
 *
 * Method 1-4 keep running through all the x-layer blobs until a rough
 * alignment is found for the three layers. The same is being done for the
 * y-layers. Methods 1&2 check for the alignment using a linear fit to the
 * real coordinate vs fiber_no plot. Methods 3&4 use, a much more precise
 * but slower, quadratic fit. Methods 1&3 do not allow for electron
 * hopping, 2&4 do.
 *
 * Methods 5-8 save all the tracks formed by the blobs in the x-layers that
 * roughly align and do the same for the y-direction. Trajectories are then
 * formed by combining each x_track with each y_track. Trajectories are
 * sorted in order of a decreasing "quality factor". The routine returns
 * the number of trajectories being returned, This is limited by the number
 * desired by the user. In case of error, a negative number is returned.
 * The subdivision among the multitraj methods is the same as for the fiber
 * space methods.
 */

/* default trajectory selection method. Can be changed by running the
 * sc_traj_select_method() routine. If you make changes, be sure to also
 * change code in traj_init. Current methods are: */
static char *Method_desc[] = {
    "brightest blob in each layer",		/* 0 */
    "linear fit in fiber space, no hop",	/* 1 */
    "linear fit in fiber space, hop",		/* 2 */
    "quadratic fit in fiber space, no hop",	/* 3 */
    "quadratic fit in fiber space, hop",	/* 4 */
    "multiple trajectories, linfit, no hop",	/* 5 */
    "multiple trajectories, linfit, hop",	/* 6 */
    "multiple trajectories, quad fit, no hop",	/* 7 */
    "multiple trajectories, quad fit, hop",	/* 8 */
};
#define NMETHODS 	(sizeof(Method_desc)/sizeof(Method_desc[0]))
static int Traj_select_method = 8;

/* quadratic fiber-to-real fit factors */
#ifdef NOTDEF
#define QUADX_FIT_FAC   0.544
#define QUADY_FIT_FAC   0.544
#else
#define QUADX_FIT_FAC  ((SC_Z[SC_X2] - SC_Z[SC_X1])/(SC_Z[SC_X3]-SC_Z[SC_X1]))
#define QUADY_FIT_FAC  ((SC_Z[SC_Y2] - SC_Z[SC_Y1])/(SC_Z[SC_Y3]-SC_Z[SC_Y1]))
    /* in other words (z2-z1)/(z3-z1) */
#endif
#define QUADX_FIT_LIM_L	-1
#define QUADX_FIT_LIM_H	 1.1
#define QUADY_FIT_LIM_L	-1
#define QUADY_FIT_LIM_H	 1

/* linear fiber-to-real fit factors */
#ifdef NOTDEF
#define LINX_FIT_FAC	0.544
#define LINY_FIT_FAC	0.544
#else
#define LINX_FIT_FAC	((SC_Z[SC_X2] - SC_Z[SC_X1])/(SC_Z[SC_X3]-SC_Z[SC_X1]))
#define LINY_FIT_FAC	((SC_Z[SC_Y2] - SC_Z[SC_Y1])/(SC_Z[SC_Y3]-SC_Z[SC_Y1]))
    /* in other words (z2-z1)/(z3-z1) */
#endif
#define LINX_FIT_LIM_L	-5
#define LINX_FIT_LIM_H	3
#define LINY_FIT_LIM_L	-3
#define LINY_FIT_LIM_H	4

static float Fit_x_factor;
static float Fit_x_lim_lo;
static float Fit_x_lim_hi;

static float Fit_y_factor;
static float Fit_y_lim_lo;
static float Fit_y_lim_hi;

/* limits on real position in multitraj() */
#define DX_LIM_L	-1.0
#define DX_LIM_H	 1.0
#define DY_LIM_L	-1.0
#define DY_LIM_H	 1.0

/* macros and variables for the brightest_each_layer routine */
static u_char Layervec;	/* which layers are present in an event */
static u_char Xmask;	/* bits for the x layers */
static u_char Ymask;	/* bits for the y layers */

#define INIT_LAYERS	(Layervec = (u_char)0)
#define GOT_ALL_LAYERS	(Layervec == 0x3f)
#define SET_LAYER(r)	(Layervec |= (1 << (r)))
#define GOT_LAYER(r)	(Layervec & (1 << (r)))

#define CLR_XMASK	(Xmask = (u_char)0)
#define SET_XMASK(x)    (Xmask |= (1 << (x)))
#define CLR_YMASK	(Ymask = (u_char)0)
#define SET_YMASK(y)    (Ymask |= (1 << (y)))

#define GOT_ALL_X	((Layervec & Xmask) == Xmask)
#define GOT_ALL_Y	((Layervec & Ymask) == Ymask)

static float Real_x_lim_lo = DX_LIM_L;
static float Real_x_lim_hi = DX_LIM_H;
static float Real_y_lim_lo = DY_LIM_L;
static float Real_y_lim_hi = DY_LIM_H;

static struct sc_evt_params Evtprm;	/* params for most recent event */

static int brightest_each_layer(int *x, int *y, long *I, int nblobs, int cam,
    struct sc_softCoords *kp);
static int fiber_fit(int *xp, int *yp, long *Ip, int nblobs, int cam,
    struct sc_softCoords *kp);

/* return codes for blob selection routines */
#define RET_NO_BLOB_SELECT_DONE	-31	/* check for correct traj method */
#define RET_NOT_ALL_LAYERS	-4
#define RET_BAD_FIB_REAL_1	-5	/* bad fiber-to-real layer 1 */
#define RET_BAD_FIB_REAL_2	-6	/* bad fiber-to-real layer 2 */
#define RET_BAD_FIB_REAL_3	-7	/* bad fiber-to-real layer 3 */
#define RET_UNMAPPED_CENTROID	-8
#define RET_NOT_ENOUGH_BLOBS	-9
#define RET_NO_X_FIT		-10
#define RET_NO_Y_FIT		-11
#define RET_NO_FIT		-12
#define RET_NO_E1_PHA		-13	/* zero pha in top detector */
#define RET_X_DEV		-14	/* x deviation too big */
#define RET_Y_DEV		-15	/* y deviation too big */
#define RET_BUNDLE_HIT		-16	/* hit fiber bundle */
#define RET_NOT_ALL_LAYERS_1	-17	/* missing lyr1 but not bundle_hit*/
#define RET_NOT_ALL_LAYERS_23	-18	/* missing layer 2 or 3 */

/* blobdata is used in the multiple trajectories routine */
struct blobdata {
    int	x;		/* xpix */
    int	y;		/* ypix */
    long I;		/* intensity */
    float fiber;	/* fiber no. */
    float ypos;		/* actual value - already divided by 128 */
    float ratio;	/* intensity of this blob / total I in its layer */
    int	hop;		/* hop code */
    u_char id;		/* sequential number of original blob */
};
static struct blobdata *blob[SC_NLAYERS];

static u_char Blobid[SC_NLAYERS];

/* Layer_max_intensity - maximum intensity in each layer. Available to
 * application programs via the function sc_layer_maxI(). Values set and
 * reset in multitraj(). Note that this maximum intensity may have been
 * affected by hopping. */
static long Layer_maxI[SC_NLAYERS];

/* trackidx - indices into the blob array. Used in multiple trajectories
 * routine and quality(). */
struct trackidx {
    int idx[SC_NLAYERS];
    short offset;
};

static void accum_hop_param(int hop, int n);
static void assign_blob(int n, int layer, int oldlayer, int xpix, int ypix, 
    int I, struct sc_fmapElem *f, int hopdir, int id);
static int blob_cmp(struct blhodo *i, struct blhodo *j);
static void clear_blobid(void);
static int cris_blob_cmp(struct sc_crisHodoData *i, struct sc_crisHodoData *j);
static void coords_init(void);
static int fib_bundle_chk(float	xb2, float xb3, float yb2, float yb3);
static double fitw( double *x, double *y, double *sigy, int n, 
    double *a, double *b); 
static int get_hop_param(void);
static void init_evtparams(void);
static void init_hop_param(void);
static void init_traj_mb(void);
static int sc_multitraj(void *scrp, struct sc_Coords_Traj *kp, 
    int cam, int ntraj, int iscris);
static float quad_fit(int layer, float fiber);
static float lin_fit(int layer, float fiber);
static int qual_compare(struct sc_Coords_Traj *i, struct sc_Coords_Traj *j);
static float quality(struct trackidx *trk,  int hop, float fit, int tel);
static void set_traj_mb(double mx, double bx, double my, double by);

/* Fitfunc - can point to either lin_fit() or quad_fit() depending on
 * value of Traj_select_method. */
static float (*Fitfunc)(int layer, float fiber);

static int Traj_init = 0;
static int Traj_maxblobs;

static long *I = NULL;
static int *X = NULL;
static int *Y = NULL;

/* Mx, Bx, My, By: value of slope (M) and intercept (B) for X and Y as
 * calculated for each trajectory in the sc_trajectory() routine.  */
static double *Mx = NULL;
static double *Bx = NULL;
static double *My = NULL;
static double *By = NULL;
static int Nmb = 0;	/* number of slopes or intercepts */

/* Do_hop - 1 if we should apply the hopping correction. Depends on the
 * value of Traj_select_method. */
static int Do_hop = 0;

#define HOPDELTA	5	/* y pixel increment or decrement for hopping */
#define HOPTHRESH	5000	/* only hop blobs with intensity below this */
static int Hopthresh = HOPTHRESH;

/* Chi-squared stuff. Chi-squared is normally checked in sc_trajectory. The
 * function check_chisq() can be used to disable either or both of the X
 * and Y checks. If the checked are enabled, the calculated chisq is
 * compared to the Max_chi_x or Max_chi_y value. These values may be
 * changed by using the set_max_traj_chisq() function. */
static int Check_chi_x = 1;		/* if nonzero, don't check */
static int Check_chi_y = 1;		/* if nonzero, don't check */
static float Max_chi_x = 1000.0;
static float Max_chi_y = 1000.0;

/* Check_E1 - check that the stack E1 value is nonzero before calculating
 * trajectories. May be unset using the sc_check_e1() function. */
static int Check_E1 = 1;	

/* Check_deviation - check the x2 and y2 real space deviation. May be unset
 * using the sc_check_dev() function. */
static int Check_deviation = 1;	

/* Check_fiber_bundle_hit - check if the trajectory would hit the fiber
 * bundle when projected back. May be uset using the sc_check_bundle()
 * function. */
static int Check_fiber_bundle_hit = 1;

/* Check_telescope - make sure the trajectory comes reasonably close to the
 * telescope the stack reported it hit. May be unset using the
 * sc_check_telescope() function. */
static int Check_telescope = 1;	

/* traj_reset - force initialization next time a routine here is called.
 * The sc_set_maxblobs() routine uses this. */
void
traj_reset(void)
{
    Traj_init = 0;
}

/* traj_init() sets the fit_functions and hop flags according to the
 * trajectory method desired and allocates space for the structures and
 * variable arrays needed. */
static int
traj_init(void)
{
    int i;

    if (Traj_init) {
	return 0;
    }

    set_stack_z();

    Traj_maxblobs = sc_get_maxblobs();

    switch (Traj_select_method) {
	case 0:
	    /* brightest each layer */
	    Fitfunc = NULL;
	    Do_hop = 0;
	    break;
	case 1:
	    /* linear fit in fiber space, no hop */
	    Fitfunc = lin_fit;
	    Do_hop = 0;
	    break;
	case 2:
	    /* linear fit in fiber space, hop */
	    Fitfunc = lin_fit;
	    Do_hop = 1;
	    break;
	case 3:
	    /* quadratic fit in fiber space, no hop */
	    Fitfunc = quad_fit;
	    Do_hop = 0;
	    break;
	case 4:
	    /* quadratic fit in fiber space, hop */
	    Fitfunc = quad_fit;
	    Do_hop = 1;
	    break;
	case 5:
	    /* multitraj, lin fit, no hop */
	    Fitfunc = lin_fit;
	    Do_hop = 0;
	    break;
	case 6:
	    /* multitraj, lin fit, hop */
	    Fitfunc = lin_fit;
	    Do_hop = 1;
	    break;
	case 7:
	    /* multitraj, quad fit, no hop */
	    Fitfunc = quad_fit;
	    Do_hop = 0;
	    break;
	case 8:
	    /* multitraj, quad fit, hop */
	    Fitfunc = quad_fit;
	    Do_hop = 1;
	    break;
	default:
	    return -1;	/* should not happen */
	    break;
    }

    if (Fitfunc == lin_fit) {
	Fit_x_factor = LINX_FIT_FAC;
	Fit_x_lim_lo = LINX_FIT_LIM_L;
	Fit_x_lim_hi = LINX_FIT_LIM_H;
	Fit_y_factor = LINY_FIT_FAC;
	Fit_y_lim_lo = LINY_FIT_LIM_L;
	Fit_y_lim_hi = LINY_FIT_LIM_H;
    } else if (Fitfunc == quad_fit || Fitfunc == NULL) {
	/* These values are needed by brightest_each_layer() for when it
	 * checks the dx and dy, even though b.e.l() does not use the
	 * Fitfunc per se. */
	Fit_x_factor = QUADX_FIT_FAC;
	Fit_x_lim_lo = QUADX_FIT_LIM_L;
	Fit_x_lim_hi = QUADX_FIT_LIM_H;
	Fit_y_factor = QUADY_FIT_FAC;
	Fit_y_lim_lo = QUADY_FIT_LIM_L;
	Fit_y_lim_hi = QUADY_FIT_LIM_H;
    } else {
	set_err_code(0, "traj_init: bad fitfunc");
	return -1;
    }

    if (blobhodo != NULL) {
	free(blobhodo);
    }
    blobhodo = (struct blhodo *)calloc(Traj_maxblobs, sizeof(struct blhodo));
    if (blobhodo == (struct blhodo *)NULL) {
	set_err_code(1, "traj_init: can't calloc 'blobhodo'");
	return -1;
    }

    if (I != NULL) {
	free(I);
    }
    I = (long *)calloc(Traj_maxblobs, sizeof(long));
    if (I == (long *)NULL) {
	set_err_code(2, "traj_init: can't calloc 'I'");
	return -1;
    }

    if (X != NULL) {
	free(X);
    }
    X = (int *)calloc(Traj_maxblobs, sizeof(long));
    if (X == (int *)NULL) {
	set_err_code(3, "traj_init: can't calloc 'X'");
	return -1;
    }

    if (Y != NULL) {
	free(Y);
    }
    Y = (int *)calloc(Traj_maxblobs, sizeof(long));
    if (Y == (int *)NULL) {
	set_err_code(4, "traj_init: can't calloc 'Y'");
	return -1;
    }

    /* fiber_fit and multitraj uses these structures */
    for (i=0; i<SC_NLAYERS; i++) {
	if (blob[i] != NULL) {
	    free(blob[i]);
	}
	blob[i] = (struct blobdata *)calloc(Traj_maxblobs,
	    sizeof(struct blobdata));
	if (blob[i] == (struct blobdata *)NULL) {
	    set_err_code(8+i, "traj_init: can't calloc 'blob[%d]'", i);
	    return -1;
	}
    }

    Nmb = 0;

    if (Mx != NULL) {
	free(Mx);
    }
    Mx = (double *)calloc(Maxtraj, sizeof(double));
    if (Mx == NULL) {
	set_err_code(14, "traj_init: Can't calloc 'Mx'");
	return -1;
    }

    if (Bx != NULL) {
	free(Bx);
    }
    Bx = (double *)calloc(Maxtraj, sizeof(double));
    if (Bx == NULL) {
	set_err_code(15, "traj_init: Can't calloc 'Bx'");
	return -1;
    }

    if (My != NULL) {
	free(My);
    }
    My = (double *)calloc(Maxtraj, sizeof(double));
    if (My == NULL) {
	set_err_code(16, "traj_init: Can't calloc 'My'");
	return -1;
    }

    if (By != NULL) {
	free(By);
    }
    By = (double *)calloc(Maxtraj, sizeof(double));
    if (By == NULL) {
	set_err_code(17, "traj_init: Can't calloc 'By'");
	return -1;
    }

    Traj_init = 1;

    return 0;
}

/* sc_msu_blob_sort - sort the blob information in a sc_msuEvt struct by
 * intensity. The brightest blobs will appear first.
 * 
 * srcp and destp are pointers to sc_msuEvt structs.  If both srcp and destp
 * are non-NULL, that event is processed.  The src data remains unchanged,
 * but the dest data contains the reordered blobs. */
void
sc_msu_blob_sort(srcp, destp)
    struct sc_msuEvt *srcp;
    struct sc_msuEvt *destp;
{
    struct blhodo *bp;
    int i;

    if (srcp == NULL || destp == NULL) {
	return;
    }

    if (!Traj_init) {
	traj_init();
    }

    *destp = *srcp;
    bp = &blobhodo[0];

    /* first do vid system 1 */
    for (i=0; i < srcp->nblobs; i++) {
	bp->x = (short)srcp->blobx[i];
	bp->y = (short)srcp->bloby[i];
	bp->I = srcp->blobI[i];
	bp++;
    }

    qsort((char *)blobhodo, (int)srcp->nblobs, sizeof(struct blhodo),
	blob_cmp);

    bp = &blobhodo[0];
    for (i=0; i < srcp->nblobs; i++) {
	destp->blobx[i] = (u_short)bp->x;
	destp->bloby[i] = (u_short)bp->y;
	destp->blobI[i] = bp->I;
	bp++;
    }

    /* now do vid system 2 */
    bp = &blobhodo[0];
    for (i=0; i < srcp->nblobsb; i++) {
	bp->x = (short)srcp->blobxb[i];
	bp->y = (short)srcp->blobyb[i];
	bp->I = srcp->blobIb[i];
	bp++;
    }

    qsort((char *)blobhodo, (int)srcp->nblobsb, sizeof(struct blhodo),
	blob_cmp);

    bp = &blobhodo[0];
    for (i=0; i < srcp->nblobsb; i++) {
	destp->blobxb[i] = (u_short)bp->x;
	destp->blobyb[i] = (u_short)bp->y;
	destp->blobIb[i] = bp->I;
	bp++;
    }
}

/* sc_cris_blob_sort - sort the blob information in a sc_crisEvt struct by
 * intensity. The brightest blobs will appear first.
 * 
 * srcp and destp are pointers to sc_crisEvt structs.  If both srcp and destp
 * are non-NULL, that event is processed.  The src data remains unchanged,
 * but the dest data contains the reordered blobs. */
void
sc_cris_blob_sort(srcp, destp)
    struct sc_crisEvt *srcp;
    struct sc_crisEvt *destp;
{
    struct blhodo *bp;
    int i;

    if (srcp == NULL || destp == NULL) {
	return;
    }

    if (!Traj_init) {
	traj_init();
    }

    *destp = *srcp;
    qsort((char *)(destp->hodo), (int)destp->nblobs,
    	sizeof(struct sc_crisHodoData), cris_blob_cmp);
}

/* sc_raw_blob_sort - sort the blob information in the x, y, and I arrays
 * by intensity. The brightest blobs will appear first.
 * 
 * xsrcp, ysrcp, and Isrcp are pointers to the input arrays. xdestp, ydestp, and
 * Idestp are the output arrays. If any of the arrays are NULL, we don't do
 * anything. The src data remains unchanged, but the dest data contains the 
 * reordered blobs. */
void
sc_raw_blob_sort(xsrcp, ysrcp, Isrcp, xdestp, ydestp, Idestp, nblobs)
   int *xsrcp;
   int *ysrcp;
   long *Isrcp;
   int *xdestp;
   int *ydestp;
   long *Idestp;
   int nblobs;
{
    struct blhodo *bp;
    int i;

    if (xsrcp == NULL || ysrcp == NULL || Isrcp == NULL ||
	    xdestp == NULL || ydestp == NULL || Idestp == NULL) {
	return;
    }

    if (!Traj_init) {
	traj_init();
    }

    if (nblobs > Traj_maxblobs) {
	nblobs = Traj_maxblobs;
    }

    bp = &blobhodo[0];

    for (i=0; i < nblobs; i++) {
	bp->x = (short)xsrcp[i];
	bp->y = (short)ysrcp[i];
	bp->I = Isrcp[i];
	bp++;
    }

    qsort((char *)blobhodo, (int)nblobs, sizeof(struct blhodo),
	blob_cmp);

    bp = &blobhodo[0];
    for (i=0; i < nblobs; i++) {
	xdestp[i] = (int)bp->x;
	ydestp[i] = (int)bp->y;
	Idestp[i] = bp->I;
	bp++;
    }
}

int
sc_cris_coords_cam(srcp, kp)
    struct sc_crisEvt *srcp;
    struct sc_softCoords *kp;
{
    int cam;
    cam = sc_cris_get_cam();
    return(sc_cris_coords(srcp, kp, cam));
}

/* sc_cris_coords - using the data in a sc_crisEvt struct, fill in the
 * pointer to the sc_softCoords struct for that event. Return 0 if all goes
 * well, else return some negative value. Used in trajectory methods 0-4
 * where only one trajectory is computed for each event. */
int
sc_cris_coords(srcp, kp, cam)
    struct sc_crisEvt *srcp;
    struct sc_softCoords *kp;
    int cam;
{
    struct sc_crisEvt dest;
    int i;
    int ret;

    if (srcp == NULL || kp == NULL) {
	set_err_code(200, "sc_cris_coords: recv'd NULL param");
	return -1;
    }

    if (srcp->nblobs > SC_CRIS_MAXBLOBS) {
	set_err_code(202, "sc_cris_coords: nblobs (%d) > max (%d)",
	    srcp->nblobs, SC_CRIS_MAXBLOBS);
	return -1;
    }

    if (!Traj_init) {
	traj_init();
    }

#ifdef NOTDEF
    /* Cris data are delivered sorted by intensity so sorting is not
     * needed. But since we refer to the dest structure, from now on we
     * better copy the original using memcpy. [jkl] */
    sc_cris_blob_sort(srcp, &dest);
#else
    memcpy(&dest,srcp,sizeof(struct sc_crisEvt));
#endif

    kp->cam = cam;

    for (i=0; i < dest.nblobs; i++) {
	X[i] = dest.hodo[i].x;
	Y[i] = dest.hodo[i].y;
	I[i] = dest.hodo[i].I;
    }

    ret = RET_NO_BLOB_SELECT_DONE;

    switch (Traj_select_method) {
	case 0:
	    /* brightest blob in each layer */
	    ret = brightest_each_layer(X, Y, I, dest.nblobs, cam, kp);
	    break;
	case 1:
	case 2:
	case 3:
	case 4:
	    /* fiber fit */
	    ret = fiber_fit(X, Y, I, dest.nblobs, cam, kp);
	    break;
	default:
	    set_err_code(201, "sc_cris_coords: bad Traj_select_method = %d", 
		Traj_select_method);
	    break;
    }

    return(ret);
}

/* sc_msu_coords - using the data in a sc_msuEvt struct, fill in the pointer
 * to two sc_softCoords structs for that event, one for camera 1 and one for
 * camera 2. Return 0 if all goes well, else return some other value (see
 * below). If a sc_softCoords pointer is NULL, we won't do that camera. 
 *
 * Return code is a bitmap. The lower 8 bits give the error code for camera
 * 1 and the next 8 bits give the error code for camera 2. Bit 16 is set
 * for a cam 1 error, and bit 17 for a cam 2 error. See the code for the
 * exact values for each error. */
int
sc_msu_coords(srcp, kp1, kp2)
    struct sc_msuEvt *srcp;
    struct sc_softCoords *kp1;
    struct sc_softCoords *kp2;
{
    int cam;
    struct sc_msuEvt dest;
    int i;
    int ret;
    int retcode = 0;

    if (srcp == NULL) {
	return -1;
    }

    ret = 0;
    if (srcp->nblobs > SC_MSU_MAXBLOBS) {
	ret |= 0x01;
    }
    if (srcp->nblobsb > SC_MSU_MAXBLOBS) {
	ret |= 0x02;
    }
    if (ret) {
	if (ret & 0x03) {
	    set_err_code(210, 
	    	"sc_msu_coords: nblobs (%d) and nblobsb (%d) > max (%d)",
		srcp->nblobs, srcp->nblobsb, SC_MSU_MAXBLOBS);
	} else if (ret & 0x02) {
	    set_err_code(211,
	    	"sc_msu_coords: nblobsb (%d) > max (%d)",
		srcp->nblobsb, SC_MSU_MAXBLOBS);
	} else {
	    set_err_code(212,
	    	"sc_msu_coords: nblobs (%d) > max (%d)",
		srcp->nblobs, SC_MSU_MAXBLOBS);
	}
	return -1;
    }

    if (!Traj_init) {
	traj_init();
    }

    if (kp1 == NULL && kp2 == NULL) {
	return 0;
    }

    sc_msu_blob_sort(srcp, &dest);

    /* first do vid system 1 */
    if (kp1 != NULL) {
	for (i=0; i < dest.nblobs; i++) {
	    X[i] = dest.blobx[i];
	    Y[i] = dest.bloby[i];
	    I[i] = dest.blobI[i];
	}

	cam = 1;
	kp1->cam = SC_CAM1;
	ret = RET_NO_BLOB_SELECT_DONE;
	switch (Traj_select_method) {
	    case 0:
		/* 6 brightest blobs */
		ret = brightest_each_layer(X, Y, I, dest.nblobs, cam, kp1);
		break;
	    case 1:
	    case 2:
	    case 3:
	    case 4:
		ret = fiber_fit(X, Y, I, dest.nblobs, cam, kp1);
		break;
	}

	if (ret == 0) {
	    /* OK */
	    ;
	} else if (ret == RET_UNMAPPED_CENTROID) {
	    retcode = 0x10001;
	} else if (ret == RET_NOT_ALL_LAYERS) {
	    retcode = 0x10002;
	} else if (ret == RET_BAD_FIB_REAL_1) {
	    retcode = 0x10004;
	} else if (ret == RET_BAD_FIB_REAL_2) {
	    retcode = 0x10008;
	} else if (ret == RET_BAD_FIB_REAL_3) {
	    retcode = 0x10010;
	} else if (ret == RET_NOT_ENOUGH_BLOBS) {
	    retcode = 0x10020;
	} else if (ret == RET_NO_X_FIT) {
	    retcode = 0x10040;
	} else if (ret == RET_NO_Y_FIT) {
	    retcode = 0x10080;
	} else {
	    /* unknown error */
	    retcode = 0x100ff;
	}
    }

    /* next do vid system 2 */
    if (kp2 != NULL) {
	for (i=0; i < dest.nblobsb; i++) {
	    X[i] = dest.blobxb[i];
	    Y[i] = dest.blobyb[i];
	    I[i] = dest.blobIb[i];
	}

	cam = 2;
	kp2->cam = SC_CAM2;
	ret = RET_NO_BLOB_SELECT_DONE;
	switch (Traj_select_method) {
	    case 0:
		/* 6 brightest blobs */
		ret = brightest_each_layer(X, Y, I, dest.nblobsb, cam, kp2);
		break;
	    case 1:
	    case 2:
	    case 3:
	    case 4:
		ret = fiber_fit(X, Y, I, dest.nblobsb, cam, kp2);
		break;
	}

	if (ret == 0) {
	    /* OK */
	    ; /* don't disturb retcode */
	} else if (ret == RET_UNMAPPED_CENTROID) {
	    retcode |= 0x20100;
	} else if (ret == RET_NOT_ALL_LAYERS) {
	    retcode |= 0x20200;
	} else if (ret == RET_BAD_FIB_REAL_1) {
	    retcode |= 0x20400;
	} else if (ret == RET_BAD_FIB_REAL_2) {
	    retcode |= 0x20800;
	} else if (ret == RET_BAD_FIB_REAL_3) {
	    retcode |= 0x21000;
	} else if (ret == RET_NOT_ENOUGH_BLOBS) {
	    retcode |= 0x21100;
	} else if (ret == RET_NO_X_FIT) {
	    retcode |= 0x21200;
	} else if (ret == RET_NO_Y_FIT) {
	    retcode |= 0x21400;
	} else {
	    /* unknown error */
	    retcode |= 0x2ff00;
	}
    }

    return(retcode);
}

int
sc_raw_coords_cam(xsrcp, ysrcp, Isrcp, nblobs, kp)
    int *xsrcp;
    int *ysrcp;
    long *Isrcp;
    int nblobs;
    struct sc_softCoords *kp;
{
    int cam;
    cam = sc_cris_get_cam();
    return(sc_raw_coords(xsrcp, ysrcp, Isrcp, nblobs, cam, kp));
}

/* sc_raw_coords - using the data in the x, y, and I arrays , fill in the
 * pointer to the sc_softCoords struct for that event. Return 0 if all goes
 * well, else return some negative value.  */
int
sc_raw_coords(xsrcp, ysrcp, Isrcp, nblobs, cam, kp)
    int *xsrcp;
    int *ysrcp;
    long *Isrcp;
    int nblobs;
    int cam;
    struct sc_softCoords *kp;
{
    int i;
    int ret;

    if (xsrcp == NULL || ysrcp == NULL || Isrcp == NULL || kp == NULL) {
	return -1;
    }

    if (!Traj_init) {
	traj_init();
    }

    if (nblobs > Traj_maxblobs) {
	nblobs = Traj_maxblobs;
    }

    sc_raw_blob_sort(xsrcp, ysrcp, Isrcp, X, Y, I, nblobs);
    kp->cam = cam;

    ret = RET_NO_BLOB_SELECT_DONE;

    switch (Traj_select_method) {
	case 0:
	    /* 6 brightest blobs */
	    ret = brightest_each_layer(X, Y, I, nblobs, cam, kp);
	    break;
	case 1:
	case 2:
	case 3:
	case 4:
	    ret = fiber_fit(X, Y, I, nblobs, cam, kp);
	    break;
    }

    return(ret);
}

/* cris_blob_cmp - qsort compare function for an array of sc_crisHodoData
 * structures. */
static int 
cris_blob_cmp(i, j)
    struct sc_crisHodoData *i;
    struct sc_crisHodoData *j;
{
    return (j->I - i->I);
}

/* blob_cmp - qsort compare function for an array of blobhodo structures.
 * Used for msu and raw data. */
int 
blob_cmp(struct blhodo *i, struct blhodo *j)
{
    return(j->I - i->I);
}

/* sc_trajectory - calculates the trajectory (angles theta and phi) for the
 * coordinates pointed to by the coordsPtr parameter. Returns the details
 * in the pointed-to sc_softTraj structure. Returns 0 if all went well, else a
 * negative value. */
int
sc_trajectory(coordsPtr, trajPtr)
    struct sc_softCoords *coordsPtr;
    struct sc_softTraj *trajPtr;
{
    double bx;
    double by;
    double chisq;
    double costheta;
    int i;
    double mx;
    double my;
    double tanphi;
    double x[3], y[3], z[3];
    double x1, y1, z1, x0, y0, z0;
    double xval, yval, zval;
    double sig[3];

    if (!Traj_init) {
	traj_init();
    }

    sig[0] = sig[1] = sig[2] = .130;	/* 130 microns in mm */

    x[0] = coordsPtr->x1;
    x[1] = coordsPtr->x2;
    x[2] = coordsPtr->x3;
    z[0] = SC_Z[SC_X1];
    z[1] = SC_Z[SC_X2];
    z[2] = SC_Z[SC_X3];
    chisq = fitw(z, x, sig, 3, &bx, &mx);
    if (Check_chi_x) {
	if (chisq > Max_chi_x) {
	    set_err_code(310, "sc_trajectory: X chisq = %.2f (> %.2f)", 
	    	chisq, Max_chi_x);
	    return(-1);
	}
    }
    trajPtr->chisqx = chisq;

    y[0] = coordsPtr->y1;
    y[1] = coordsPtr->y2;
    y[2] = coordsPtr->y3;
    z[0] = SC_Z[SC_Y1];
    z[1] = SC_Z[SC_Y2];
    z[2] = SC_Z[SC_Y3];
    chisq = fitw(z, y, sig, 3, &by, &my);
    if (Check_chi_y) {
	if (chisq > Max_chi_y) {
	    set_err_code(311, "sc_trajectory: Y chisq = %.2f (> %.2f)", 
	    	chisq, Max_chi_y);
	    return(-2);
	}
    }
    trajPtr->chisqy = chisq;
    
    /* calculate phi, theta, and entrypt - see notebook 
     * ACE/SOFIE #5, p. 136 for an explanation of this. */
    z1 = 100.0; 
    x1 = (z1 * mx) + bx;
    y1 = (z1 * my) + by;
    z0 = 0.0;
    x0 = bx;
    y0 = by;

    xval = x1 - x0;
    yval = y1 - y0;
    zval = z1 - z0;

    /* tanphi = yval / xval; */
#ifndef DEGRAD
#define DEGRAD	57.29578	/* degrees per radian */
#endif
    costheta = zval /(sqrt(SQR(xval)+SQR(yval)+SQR(zval)));

    trajPtr->layer1x = x[0];
    trajPtr->layer1y = y[0];
    trajPtr->layer2x = x[1];
    trajPtr->layer2y = y[1];
    trajPtr->layer3x = x[2];
    trajPtr->layer3y = y[2];
    for (i=0; i<SC_NSTACK; i++) {
	trajPtr->stackx[i] = (SC_STACK_Z[i] * mx) + bx;
	trajPtr->stacky[i] = (SC_STACK_Z[i] * my) + by;
    }
    trajPtr->theta = acos(costheta) * DEGRAD;
    trajPtr->phi = atan2(yval, xval) * DEGRAD;
    
    set_traj_mb(mx, bx, my, by);
    return(0);
}

/* sc_short_traj computes a trajectory based on x2,y2,x3,y3 coordinates
 * only. */
int
sc_short_traj(struct sc_softCoords *coordsPtr, struct sc_softTraj *trajPtr)
{
    double bx;
    double by;
    double costheta;
    int i;
    int cam = SC_CAM2;
    double mx;
    double my;
    double tanphi;
    double x[3], y[3], z[3];
    double x1, y1, z1, x0, y0, z0;
    double xval, yval, zval;

    if(!GOT_LAYER(SC_X2) || !GOT_LAYER(SC_X3) ||
    	!GOT_LAYER(SC_Y2) || !GOT_LAYER(SC_Y3) ) {
	set_err_code(280,"sc_short_traj: not enough layers");
	return RET_NOT_ALL_LAYERS_23;
    }

    if (sc_fiber_to_real(cam, 2, coordsPtr->fiber[SC_X2], 
        coordsPtr->fiber[SC_Y2],&(coordsPtr->x2), &(coordsPtr->y2))) {
	    set_err_code(283, "sc_short_traj: bad fiber_to_real 2");
	    return RET_BAD_FIB_REAL_2;
    }

    if (sc_fiber_to_real(cam, 3, coordsPtr->fiber[SC_X3], 
        coordsPtr->fiber[SC_Y3], &(coordsPtr->x3), &(coordsPtr->y3))) {
	    set_err_code(284, "sc_short_traj: bad fiber_to_real 3");
	    return RET_BAD_FIB_REAL_3;
    }


    x[0] = coordsPtr->x2;
    x[1] = coordsPtr->x3;
    z[0] = SC_Z[SC_X2];
    z[1] = SC_Z[SC_X3];
    mx = (x[1]-x[0])/(z[1]-z[0]);
    bx = x[1] - mx*z[1];

    y[0] = coordsPtr->y2;
    y[1] = coordsPtr->y3;
    z[0] = SC_Z[SC_Y2];
    z[1] = SC_Z[SC_Y3];
    my = (y[1]-y[0])/(z[1]-z[0]);
    by = y[1] - my*z[1];
    
    /* calculate phi, theta, and entrypt - see notebook 
     * ACE/SOFIE #5, p. 136 for an explanation of this. */
    z1 = 100.0; 
    x1 = (z1 * mx) + bx;
    y1 = (z1 * my) + by;
    z0 = 0.0;
    x0 = bx;
    y0 = by;

    xval = x1 - x0;
    yval = y1 - y0;
    zval = z1 - z0;

    /* tanphi = yval / xval; */
#ifndef DEGRAD
#define DEGRAD	57.29578	/* degrees per radian */
#endif
    costheta = zval /(sqrt(SQR(xval)+SQR(yval)+SQR(zval)));

    trajPtr->layer1x = mx*SC_Z[SC_X1] + bx;
    trajPtr->layer1y = my*SC_Z[SC_Y1] + by;
    trajPtr->layer2x = x[0];
    trajPtr->layer2y = y[0];
    trajPtr->layer3x = x[1];
    trajPtr->layer3y = y[1];
    for (i=0; i<SC_NSTACK; i++) {
	trajPtr->stackx[i] = (SC_STACK_Z[i] * mx) + bx;
	trajPtr->stacky[i] = (SC_STACK_Z[i] * my) + by;
    }
    trajPtr->theta = acos(costheta) * DEGRAD;
    trajPtr->phi = atan2(yval, xval) * DEGRAD;
    
    set_traj_mb(mx, bx, my, by);
    return(0);
}

/* fitw - weighted least-squares linear fit - Bevington pp. 104-5 */
static double
fitw(x, y, sigy, n, a, b)
    double *x;
    double *y; 
    double *sigy; 
    int n; 
    double *a; 
    double *b; 
{
    double chisq;
    double delta;
    int i;
    double sum, sumx, sumy, sumx2, sumy2, sumxy;
    double wt;

    sum = sumx = sumy = sumx2 = sumy2 = sumxy = 0.0;

    for (i=0; i<n; i++) {
	wt = 1.0 / SQR(sigy[i]);
	sum += wt;
	sumx += wt * x[i];
	sumy += wt * y[i];
	sumx2 += wt * SQR(x[i]);
	sumxy += wt * x[i] * y[i];
	sumy2 += wt * SQR(y[i]);
    }

    delta = (sum * sumx2) - SQR(sumx);
    *a = (sumx2*sumy - sumx*sumxy) / delta;
    *b = (sumxy*sum - sumx*sumy) / delta;
    chisq = sumy2 - (2 * (*b) * sumxy) - (2 * (*a) * sumy) + 
    	(SQR(*b) * sumx2) + (2 * (*b) * (*a) * sumx) + (SQR(*a) * sum); 
#ifdef NOTDEF
    /* if you want to use these extra sigmas, add 'double *siga' and
     * 'double *sigb' to the function prototype above. And of course,
     * change the calling routine. */
    *siga = sqrt(chisq*sumx2/delta);
    *sigb = sqrt(chisq*sum/delta);
#endif
    return (chisq);
}

double
sc_layer_z(layer)
    int layer;
{
    return (SC_Z[layer]);
}

/* brightest_each_layer - fills the structure sc_softCoords *kp using the
 * brightest blobs in each layer. Errors are returned if:
 *	1) Not one blob in each layer
 *	2) The trajectory based on the bottom layers points to the fiber
 *	   bundle
 *	3) A fiber-to-real conversion fails.
 *	4) The delta_x or delta_y of the trajectory exceed limits. */
static int 
brightest_each_layer(x, y, I, nblobs, cam, kp)
    int *x;
    int *y;
    long *I;
    int nblobs;
    int cam;
    struct sc_softCoords *kp;
{
    struct sc_fmapElem *fmapp;
    int i;
    int lyr;
    int ret;
    float ypos;

#ifdef NOTDEF
    /* Deleted [by jkl] in order to check for out of geometry events */
    if( nblobs < SC_NLAYERS )  {
	set_err_code(270, "brightest: not enough blobs");
	return RET_NOT_ENOUGH_BLOBS;
    }
#endif

    coords_init();

    for (i=0, INIT_LAYERS; i < nblobs && !GOT_ALL_LAYERS; i++) {
	if(x[i]>=SC_FMAPX) {
	    continue;
	}
	fmapp = sc_centr_to_fiber(cam, x[i], y[i]);
	if (fmapp == NULL) {
	    continue;
	}
	lyr = fmapp->layer;
	if (lyr == SC_NOLAYER) {
	    continue;
	}
	ypos = (float)fmapp->ypos / 128.0;

	if (!GOT_LAYER(lyr)) {
	    SET_LAYER(lyr);
	    kp->xpix[lyr] = x[i];
	    kp->ypix[lyr] = y[i];
	    kp->I[lyr] = I[i];
	    kp->fiber[lyr] = fmapp->fiber;
	    kp->ypos[lyr] = ypos;
	    Blobid[lyr] = (u_char)i;
	}
    }

    if(!GOT_LAYER(SC_X2) || !GOT_LAYER(SC_Y2) || !GOT_LAYER(SC_X3) || 
    	!GOT_LAYER(SC_Y3)){
	set_err_code(271, "brightest: not all layers (2 or 3)");
	return RET_NOT_ALL_LAYERS_23;
    }


    if (sc_fiber_to_real(cam, 2, kp->fiber[SC_X2], kp->fiber[SC_Y2],
	    &(kp->x2), &(kp->y2))) {
	set_err_code(273, "brightest: bad fiber_to_real 2");
	return RET_BAD_FIB_REAL_2;
    }

    if (sc_fiber_to_real(cam, 3, kp->fiber[SC_X3], kp->fiber[SC_Y3],
	    &(kp->x3), &(kp->y3))) {
	set_err_code(274, "brightest: bad fiber_to_real 3");
	return RET_BAD_FIB_REAL_3;
    }

    if (Check_fiber_bundle_hit) {
	if(fib_bundle_chk(kp->x2, kp->x3, kp->y2, kp->y3)) {
	    set_err_code(277,"brightest: fiber bundle hit");
	    return RET_BUNDLE_HIT;
	}
    }

    if(!GOT_LAYER(SC_X1) || !GOT_LAYER(SC_Y1)){
	set_err_code(278, "brightest: not all layers (1)");
	return RET_NOT_ALL_LAYERS_1;
    }

    if (sc_fiber_to_real(cam, 1, kp->fiber[SC_X1], kp->fiber[SC_Y1],
	    &(kp->x1), &(kp->y1))) {
	set_err_code(272, "brightest: bad fiber_to_real 1");
	return RET_BAD_FIB_REAL_1;
    }


    if (Check_deviation) {
	/* keep only if real space coordinates align properly */
	if (!sc_x_dev_ok(kp->x1, kp->x2, kp->x3, NULL)) {
	    set_err_code(275, "brightest: bad real x offset");
	    return RET_X_DEV;
	}

	if (!sc_y_dev_ok(kp->y1, kp->y2, kp->y3, NULL)) {
	    set_err_code(276, "brightest: bad real y offset");
	    return RET_Y_DEV;
	}
    }

    init_traj_mb();
    return 0;
}

/* coords_init - initializes items for brightest_each_layer()
*/
static void
coords_init()
{
    CLR_XMASK;
    SET_XMASK(SC_X1);
    SET_XMASK(SC_X2);
    SET_XMASK(SC_X3);

    CLR_YMASK;
    SET_YMASK(SC_Y1);
    SET_YMASK(SC_Y2);
    SET_YMASK(SC_Y3);
}

/* The following is the fiber-fit code by Joe Klarmann. It fits one blob
 * from each layer to a straight line. If the fit is good enough, we use
 * those blobs. We start with the brightest blob in each layer.
 * Calculates the offset in x2 from a straight line passed through
 * x1 and x3. The same is done for y coordinates. The coordinates are 
 * calculated using a linear fit y = mx + b to the x vs fiber plot on 
 * the msu hole run. The coefficients are listed as bxi and mxi. The 
 * layer separations are taken from the SC_Z table, to define LINX_FIT_FAC 
 * and LINY_FIT_FAC. */

static int
fiber_fit(xp, yp, Ip, nblobs, cam, kp)
    int *xp;
    int *yp;
    long *Ip;
    int nblobs;
    int cam;
    struct sc_softCoords *kp;
{
    int dy;		/* diff in y for hopping */
    struct sc_fmapElem felem;
    struct sc_fmapElem *fmapp;
    int hop;
    int hopdir;
    int	i;
    int index;
    int lblobs[SC_NLAYERS];	/* no. of blobs in each layer */
    int lblobs1,lblobs2,lblobs3;
    int	lyr;
    int n;
    float oldfiber;
    int oldlayer;
    int ret;
    float x1_raw, x2_raw, x3_raw, x2_offset;
    float xoffset;
    float y1_raw, y2_raw, y3_raw, y2_offset;
    float yoffset;
    float xb2, xb3, yb2, yb3;	/* coords for fib_bundle_chk */

#ifdef NOTDEF
    /* deleted [by jkl] to check for out of geometry events */
    if (nblobs < SC_NLAYERS )  {
	set_err_code(250, "fiber_fit: not enough blobs");
	return RET_NOT_ENOUGH_BLOBS;
    }
#endif

    if (nblobs > Traj_maxblobs) {
	nblobs = Traj_maxblobs;
    }

    for (i=0;i<SC_NLAYERS;i++) {
	lblobs[i] = 0;
    }

    for (i=0; i < nblobs ; i++) {
	if(xp[i]>=SC_FMAPX) {
	    continue;
	}
	fmapp = sc_centr_to_fiber(cam, xp[i], yp[i]);
	if (fmapp == NULL) {
	    continue;
	}
	lyr = fmapp->layer;
	if (lyr != SC_NOLAYER) {
	    n = lblobs[lyr];
	    hopdir = 0;
	    assign_blob(n, lyr, lyr, xp[i], yp[i], Ip[i], fmapp, hopdir, i);
	    oldfiber = fmapp->fiber;
	    if (++lblobs[lyr] >= Traj_maxblobs) {
		goto Got_blobs;
	    }
	}
    }

    if (Do_hop) {
	/* if intensity is low add hopped blobs to structures */
        for (i=0; i < nblobs ; i++) {
	    if(Ip[i] < Hopthresh) {
		if(xp[i]>=SC_FMAPX) {
		    continue;
		}
		fmapp = sc_centr_to_fiber(cam, xp[i], yp[i]);
	        if (fmapp == NULL) {
	            continue;
	        }
		if(fmapp->layer == SC_NOLAYER) {
		    oldfiber = 0;
		    oldlayer = -1;
		} else {
		    oldfiber = fmapp->fiber;
		    oldlayer = fmapp->layer;
		}
		fmapp = &felem;
		for(hopdir = 1;hopdir > -2;hopdir -= 2) {	/* hop up then down*/
		    ret = sc_hop(hopdir, cam, xp[i], yp[i], 
		        oldfiber,  &dy, fmapp);
		    if (ret == 0) {
		        lyr = fmapp->layer;
		        n = lblobs[lyr];
		        assign_blob(n, lyr, oldlayer, xp[i], yp[i]+dy, Ip[i], 
		    	    fmapp, hopdir, i);
		        if(++lblobs[lyr] >= Traj_maxblobs) {
			    goto Got_blobs;
		        }
		    }
		}
	    }
	}
    }

Got_blobs:
    if (!lblobs[SC_X2] || !lblobs[SC_Y2] || !lblobs[SC_X3] || !lblobs[SC_Y3]) {
	/* we do not have a blob for each layer */
	set_err_code(253, "fiber_fit: not all layers (2 or 3)");
	return (RET_NOT_ALL_LAYERS_23);
    }
#ifdef NOTDEF
    /* deleted [by jkl] to check for out of geometry events */
    for (i=0; i<SC_NLAYERS; i++) {
	if (lblobs[i] == 0) {
	    /* we do not have a blob for each layer */
	    set_err_code(253, "fiber_fit: not all layers");
	    return (RET_NOT_ALL_LAYERS);
	}
    }
#endif

    /* First check whether this event was probably triggered by hitting the
     * fiber bundle.  Use the brightest blobs in layers 2 and 3 to
     * construct a trajectory.  If the extrapolation into layer 1 goes into
     * the bundle area a bundle hit is the most likely origin for this
     * event and further analysis would just yield noise!! (Joe Klarmann) */
    if (!sc_fiber_to_real(cam, 2, blob[SC_X2][0].fiber, blob[SC_Y2][0].fiber,
		&xb2,&yb2) &&
	   !sc_fiber_to_real(cam, 3, blob[SC_X3][0].fiber, blob[SC_Y3][0].fiber,
		&xb3,&yb3)) {
	if (Check_fiber_bundle_hit) {
	    if(fib_bundle_chk(xb2,xb3,yb2,yb3)) {
		set_err_code(262, "fiber_fit: fiber bundle hit");
		return RET_BUNDLE_HIT;
	    }
	}
    }

    if (!lblobs[SC_X1] || !lblobs[SC_Y1]) {
	/* we do not have a blob for each layer */
	set_err_code(254, "fiber_fit: not all layers (1)");
	return (RET_NOT_ALL_LAYERS_1);
    }

    init_hop_param();

    for (lblobs1=0; lblobs1 < lblobs[SC_X1]; lblobs1++)  {
	x1_raw = Fitfunc(SC_X1, blob[SC_X1][lblobs1].fiber); 
	for (lblobs2=0; lblobs2 < lblobs[SC_X2];lblobs2++) {
	    x2_raw = Fitfunc(SC_X2, blob[SC_X2][lblobs2].fiber);
	    for (lblobs3=0; lblobs3 < lblobs[SC_X3]; lblobs3++) {
		if(blob[SC_X1][lblobs1].hop && blob[SC_X2][lblobs2].hop &&
		       blob[SC_X3][lblobs3].hop) {
		    /* skip if all three have hopped */
		   continue;
		}
		x3_raw = Fitfunc(SC_X3, blob[SC_X3][lblobs3].fiber);
		x2_offset = x2_raw - x1_raw - (Fit_x_factor * (x3_raw - x1_raw));
		if ((Fit_x_lim_lo < x2_offset) && (x2_offset < Fit_x_lim_hi)) {
		    kp->xpix[SC_X1] = blob[SC_X1][lblobs1].x;
		    kp->ypix[SC_X1] = blob[SC_X1][lblobs1].y;
		    kp->I[SC_X1] = blob[SC_X1][lblobs1].I;
		    kp->fiber[SC_X1] = blob[SC_X1][lblobs1].fiber;
		    kp->ypos[SC_X1] = blob[SC_X1][lblobs1].ypos;
		    Blobid[SC_X1] = blob[SC_X1][lblobs1].id;
		    
		    kp->xpix[SC_X2] = blob[SC_X2][lblobs2].x;
		    kp->ypix[SC_X2] = blob[SC_X2][lblobs2].y;
		    kp->I[SC_X2] = blob[SC_X2][lblobs2].I;
		    kp->fiber[SC_X2] = blob[SC_X2][lblobs2].fiber;
		    kp->ypos[SC_X2] = blob[SC_X2][lblobs2].ypos;
		    Blobid[SC_X2] = blob[SC_X2][lblobs2].id;
		   
		    kp->xpix[SC_X3] = blob[SC_X3][lblobs3].x;
		    kp->ypix[SC_X3] = blob[SC_X3][lblobs3].y;
		    kp->I[SC_X3] = blob[SC_X3][lblobs3].I;
		    kp->fiber[SC_X3] = blob[SC_X3][lblobs3].fiber;
		    kp->ypos[SC_X3] = blob[SC_X3][lblobs3].ypos;
		    Blobid[SC_X3] = blob[SC_X3][lblobs3].id;
		    goto Got_x_fit;
		}
	    }
	}
    }
    set_err_code(255, "fiber_fit: no X fit");
    return (RET_NO_X_FIT);	/* didn't get no X fit */
Got_x_fit:
    accum_hop_param(blob[SC_X1][lblobs1].hop, SC_X1);
    accum_hop_param(blob[SC_X2][lblobs2].hop, SC_X2);
    accum_hop_param(blob[SC_X3][lblobs3].hop, SC_X3);

    for (lblobs1 =0; lblobs1 < lblobs[SC_Y1]; lblobs1++)  {
	y1_raw = Fitfunc(SC_Y1, blob[SC_Y1][lblobs1].fiber);
	for (lblobs2 = 0; lblobs2 < lblobs[SC_Y2];lblobs2++) {
	    y2_raw = Fitfunc(SC_Y2, blob[SC_Y2][lblobs2].fiber);
	    for (lblobs3 = 0; lblobs3 < lblobs[SC_Y3]; lblobs3++) {
		if(blob[SC_Y1][lblobs1].hop && blob[SC_Y2][lblobs2].hop &&
		       blob[SC_Y3][lblobs3].hop) {
		    /* skip if all three have hopped */
		   continue;
		}
		y3_raw = Fitfunc(SC_Y3, blob[SC_Y3][lblobs3].fiber);
		y2_offset = y2_raw - y1_raw - (Fit_y_factor * (y3_raw - y1_raw));
		if ((Fit_y_lim_lo < y2_offset) && (y2_offset < Fit_y_lim_hi))  {
		    kp->xpix[SC_Y1] = blob[SC_Y1][lblobs1].x;
		    kp->ypix[SC_Y1] = blob[SC_Y1][lblobs1].y;
		    kp->I[SC_Y1] = blob[SC_Y1][lblobs1].I;
		    kp->fiber[SC_Y1] = blob[SC_Y1][lblobs1].fiber;
		    kp->ypos[SC_Y1] = blob[SC_Y1][lblobs1].ypos;
		    Blobid[SC_Y1] = blob[SC_Y1][lblobs1].id;
		    
		    kp->xpix[SC_Y2] = blob[SC_Y2][lblobs2].x;
		    kp->ypix[SC_Y2] = blob[SC_Y2][lblobs2].y;
		    kp->I[SC_Y2] = blob[SC_Y2][lblobs2].I;
		    kp->fiber[SC_Y2] = blob[SC_Y2][lblobs2].fiber;
		    kp->ypos[SC_Y2] = blob[SC_Y2][lblobs2].ypos;
		    Blobid[SC_Y2] = blob[SC_Y2][lblobs2].id;
		    
		    kp->xpix[SC_Y3] = blob[SC_Y3][lblobs3].x;
		    kp->ypix[SC_Y3] = blob[SC_Y3][lblobs3].y;
		    kp->I[SC_Y3] = blob[SC_Y3][lblobs3].I;
		    kp->fiber[SC_Y3] = blob[SC_Y3][lblobs3].fiber;
		    kp->ypos[SC_Y3] = blob[SC_Y3][lblobs3].ypos;
		    Blobid[SC_Y3] = blob[SC_Y3][lblobs3].id;
		    goto Got_y_fit;
		}
	    }
	}
    }
    set_err_code(256, "fiber_fit: no Y fit");
    return (RET_NO_Y_FIT);	/* didn't get no y fit */
Got_y_fit:
    accum_hop_param(blob[SC_Y1][lblobs1].hop, SC_Y1);
    accum_hop_param(blob[SC_Y2][lblobs2].hop, SC_Y2);
    accum_hop_param(blob[SC_Y3][lblobs3].hop, SC_Y3);
    
    if (sc_fiber_to_real(cam, 1, kp->fiber[SC_X1], kp->fiber[SC_Y1],
	    &(kp->x1), &(kp->y1))) {
	set_err_code(257, "fiber_fit: bad fiber_to_real 1");
	return (RET_BAD_FIB_REAL_1);
    }

    if (sc_fiber_to_real(cam, 2, kp->fiber[SC_X2], kp->fiber[SC_Y2],
	    &(kp->x2), &(kp->y2))) {
	set_err_code(258, "fiber_fit: bad fiber_to_real 2");
	return (RET_BAD_FIB_REAL_2);
    }

    if (sc_fiber_to_real(cam, 3, kp->fiber[SC_X3], kp->fiber[SC_Y3],
	    &(kp->x3), &(kp->y3))) {
	set_err_code(259, "fiber_fit: bad fiber_to_real 3");
	return (RET_BAD_FIB_REAL_3);
    }

    if (Check_deviation) {
	/* keep only if real space coordinates align properly */
	if (!sc_x_dev_ok(kp->x1, kp->x2, kp->x3, &xoffset)) {
	    set_err_code(260, "fiber_fit: bad real x offset");
	    return RET_X_DEV;
	}

	if (!sc_y_dev_ok(kp->y1, kp->y2, kp->y3, &yoffset)) {
	    set_err_code(261, "fiber_fit: bad real y offset");
	    return RET_Y_DEV;
	}
    }

    init_traj_mb();
    return 0;
}

/* sc_traj_select_method - use a different blob selection routine in
 * the coords routines. Returns the previous method id or -1 if an invalid
 * method id was selected. */
int
sc_traj_select_method(int m)
{
    int ret;
    if (m >= NMETHODS || m < 0) {
	return -1;
    }
    ret = Traj_select_method;
    Traj_select_method = m;
    traj_reset();	/* need to reinitialize. See traj_init(). */
    return (ret);
}

/* sc_describe_all_traj_methods() - return a pointer to a string describing the
 * methods of determining a trajectory. */
char *
sc_describe_all_traj_methods(void)
{
    char *cp;
    int m;
#define SSIZE (80 * sizeof(Method_desc))
    static char s[SSIZE];

    memset(s, '\0', SSIZE);
    for (m=0, cp=s; m<NMETHODS; m++) {
	sprintf(cp, "%d. %s\n", m, Method_desc[m]);
	cp += strlen(cp);
    }
    return s;
}

char *
sc_describe_traj_method(m)
    int m;
{
    static char s[80];
    if (m < 0 || m >= NMETHODS) {
	return NULL;
    }
    memset(s, '\0', SSIZE);
    sprintf(s, "%s", Method_desc[m]);
    return s;
}

static void
set_traj_mb(double mx, double bx, double my, double by)
{
    if (Nmb >= Maxtraj) {
	return;
    }
    Mx[Nmb] = mx;
    Bx[Nmb] = bx;
    My[Nmb] = my;
    By[Nmb] = by;
    Nmb++;
}

static void
init_traj_mb(void)
{
    Nmb = 0;
}

void 
sc_get_traj_mb(mx, bx, my, by)
    double *mx;
    double *bx;
    double *my;
    double *by;
{
    int i;

    for(i=0; i<Nmb; i++) {
	mx[i] = Mx[i];
	bx[i] = Bx[i];
	my[i] = My[i];
	by[i] = By[i];
    }
}

/* The following is the multiple trajectories code by Joe Klarmann. [Words
 * inside the square brackets added by MAO.] Blob structures are created
 * for each given centroid which yields a meaningful fiber number.  For low
 * intensity centroids [less than Hopthresh] blobs are also created for
 * centroids hopped up and down one tab of fibers. The blob structures are
 * labeled by layer and index of decreasing intensity. Straight lines,
 * using quadratic fits to the "coordinate vs fiber" distribution, are
 * fitted to all triads of coresponding blobs (x1,x2,x3 or y1,y2,y3). If
 * the offset of the center layer is within limits, the indices of the
 * three blob structures are saved [...in trackidx structures...] Real
 * space coordinates are computed for all combinations of these triads of x
 * and y blobs.  Those whose offsets (in x and y) in the center layer are
 * within limits yield structures sc_Coords_Traj. Each structure includes
 * [... various measures ...] of quality of fit. The structures are then
 * sorted according to quality of fit and the desired number returned to
 * the calling program. */

/* quality() - alter the quality factor assigned to each trajectory.  It is
 * given the blobdata for each centroid participating in the trajectory as
 * well as the composite hop_index, the fit parameter and the result of the
 * telescope_hit routine.  */
static float 
quality(trk, hop, fit, tel)
    struct trackidx *trk;
    int hop;
    float fit;
    int tel;
{
    int i;
    float I[SC_NLAYERS];
    float Iqual=0;
    float Irat[SC_NLAYERS];
    int have_x = 0;	/* number of x layers hopped */
    int have_y = 0;	/* number of y layers hopped */
    int hop_no = 0;
    int n;
    float qual;

    /* collect the ratio of intensity to the maximum intensity in the layer
     * for each blob */
    for (i=0; i<SC_NLAYERS; i++) {
	n = trk->idx[i];
	Irat[i] = blob[i][n].ratio;
    }

    /* quality factor = average of these ratios */
    for (i=0; i<SC_NLAYERS; i++) {
    	Iqual += Irat[i];
    }
    qual = Iqual/(float)SC_NLAYERS;

    /* compute the number of hopped layers */
    for (i=0; i<SC_NLAYERS; i++) {
        if(((hop >> (3*i)) & 0x7 ) != 0) {
	    hop_no++;
	    if(IS_X_LAYER(i)) {
	        have_x++;
	    } else {
	        have_y++;
	    }
	}
    }
    if (have_x > 1 || have_y > 1) {
        hop_no++;
    }
    
    /* multiply the quality factor by (1 - hop_no/10)  */
    qual *= (1.0 - ((float)hop_no)/10.0);
    /*  qual *= (fit < 0.35) ? 1 : (1.35 - fit);  */

    /* Take the straight line fit into account. 
    *  drop very little until 0.35 is reached then fast */
    qual *= (fit < 0.35 ? (1 - fit/10.) : (1.315 - fit));  

    /* Last not least,  if the trajectory misses the entrance window of the
     * prime telescope reduce the quality factor by 50%
     */
    if (tel < 0) {
        qual *= 0.5;
    }

#ifdef NOTDEF
    printf("\n");
    for(i=0;i<6;i++) printf("%5.3f ",Irat[i]);
    printf("\n%.3f\n",qual);
#endif
    return (qual);

}


#ifdef NOTDEF
    /* this is an old version */
static float 
quality(trk,  hop, fit, tel)
    struct trackidx *trk;
    int hop;
    float fit;
    int tel;
{
    int i;
    float I[SC_NLAYERS];
    float Iqual=0;
    float Irat[SC_NLAYERS];
    int hop_no = 0;
    int n;
    float qual;

    /* collect the ratio of intensity to the maximum intensity in the layer
     * for each blob */
    for (i=0; i<SC_NLAYERS; i++) {
	n = trk->idx[i];
	Irat[i] = blob[i][n].ratio;
    }

    /* quality factor = average of these ratios */
    for (i=0; i<SC_NLAYERS; i++) {
    	Iqual += Irat[i];
    }
    qual = Iqual/(float)SC_NLAYERS;

    /* compute the number of hopped layers */
    for (i=0; i<SC_NLAYERS; i++) {
        if(((hop >> (3*i)) & 0x7 ) != 0) {
	    hop_no++;
	}
    }
    /* multiply the quality factor by (1 - hop_no/10)  */
    qual *= (1.0 - ((float)hop_no)/10.0);
    /*  qual *= (fit < 0.35) ? 1 : (1.35 - fit);  */

    /* Take the straight line fit into account. 
    * currently droping 10% for every 200 micron offset */
    qual *= (1.0 - (float)(int)(fit/.2)/10.);  

    /* Last not least,  if the trajectory misses the entrance window of the
     * prime telescope reduce the quality factor by 50%
     */
    if (tel < 0) {
        qual *= 0.5;
    }

#ifdef NOTDEF
    printf("\n");
    for(i=0;i<6;i++) printf("%5.3f ",Irat[i]);
    printf("\n%.3f\n",qual);
#endif
    return (qual);

}
#endif /* NOTDEF at top of function */

static int
qual_compare(i,j)
    struct sc_Coords_Traj *i;
    struct sc_Coords_Traj *j;
{
    if(i->Q.qual_factor > j->Q.qual_factor) {
	return (-1);
    }
    if(i->Q.qual_factor < j->Q.qual_factor) {
	return (1);
    }
    return(0);
}

/* lin_fit yields the linear fit of coordinate to fiber. */
static float
lin_fit(layer, fiber)
    int layer;
    float fiber;
{
    float m[] = {0.204352, 0.198983, 0.201164, 0.21499, 0.22644, 0.21072};
    float b[] = {4.71893, -246.309, -521.274, -835.004, -1150.57, -1315.16};
    return(b[layer] + (m[layer] * fiber));
}

/* quad_fit yields the quadratic fit of coordinate to fiber. */
static float
quad_fit(layer, fiber)
    int layer;
    float fiber;
{
    float a[] = {4.202, -266.491, -464.196, -591.977, -1572.467, -1049.912};
    float b[] = {0.206488, 0.219827, 0.165696, 0.107317, 0.375145, 0.133717};
    float c[] = {-1.60906e-6, -5.15057e-6, 5.44165e-6, 1.186425e-5, 
    	-1.305987e-5, 5.57449e-6};
    return(a[layer] + b[layer]*fiber + c[layer]*SQR(fiber));
#ifdef NOTDEF
    /* [commented out by jkl] */
    return((c[layer]*fiber + b[layer])*fiber + a[layer]);
#endif
}

static int
multitraj(srcp, kp, cam, ntraj, evttype)
    void *srcp;		/* points to either an crisEvt or msuEvt */
    struct sc_Coords_Traj *kp;
    int cam;
    int ntraj;		/* number of coord structures wanted */  
    int evttype;	/* ISCRIS, ISMSU, or ISRAW */
{
#ifndef _WINDOWS
    struct sc_Coords_Traj ct[Maxtraj];	/* struct for each trajectory found */
#else
    struct sc_Coords_Traj ct[MAXTRAJ];	/* struct for each trajectory found */
#endif /* _WINDOWS */
    struct sc_crisEvt destc;
    struct sc_msuEvt destm;
    float df;		/* difference in old vs new fiber no. */
    int dy;		/* diff in y for hopping */
    struct sc_fmapElem felem;
    float fit_qual;	/* some measure of fit quality */
    struct sc_fmapElem *fmapp;
    int hopdir;		/* hop direction (-1, 0, 1) see assign_blob() */
    long I;
    int i;
    int lblobs1;	/* loop counter for no. blobs layer 1 */
    int lblobs2;	/* loop counter for no. blobs layer 2 */
    int lblobs3;	/* loop counter for no. blobs layer 3 */
    int	lblobs[SC_NLAYERS];	/* number of blobs in each layer */
    int ltraj=0;
    int ltrkx=0;	/* number of X tracks found */
    int ltrky=0;	/* number of Y tracks found */
    int	lyr;
#ifdef NOTDEF
    long minI;		/* minimum intensity - unused */
#endif
#ifdef ONE_HOP
    int nhops;
    unsigned hop_parm;
#endif
    int n;
    int nblobs;
    float oldfiber;	/* original fiber - used in hopping */
    int oldlayer;	/* original layer - used to see if we've crossed a
			 * layer boundary in hopping */
    struct trackidx qualtrack;	/* track to check quality of */
    int ret;
    struct sc_crisEvt *srcc;
    struct sc_msuEvt *srcm;
    struct raw_evt *srcr;
    int sum;
#ifdef NOTDEF
    long sumI;		/* intensity sum - unused */
#endif
    int tel_no;		/* telescope that was hit */
#ifndef _WINDOWS
    /* trackx - index into blob structs for each X track for each X layer */
    struct trackidx trackx[Maxtracks];
    /* tracky - index into blob structs for each Y track for each Y layer */
    struct trackidx tracky[Maxtracks];
#else
    /* trackx - index into blob structs for each X track for each X layer */
    struct trackidx trackx[MAXTRACKS];
    /* tracky - index into blob structs for each Y track for each Y layer */
    struct trackidx tracky[MAXTRACKS];
#endif /* _WINDOWS */
    int trx;		/* counter over no. X tracks found */
    int try;		/* counter over no. Y tracks found */
    int	x;
    float x1;		/* real space x1 */
    float x1_raw;	/* fitted value of x1 fiber */
    float x2;		/* real space x2 */
    float x2_offset;	/* delta-x in fiber space? */
    float x2_raw;	/* fitted value of x2 fiber */
    float x3;		/* real space x3 */
    float x3_raw;	/* fitted value of x3 fiber */
    float xoffset;	/* delta-x in real space? */
    int y;
    float y1;		/* real space y1 */
    float y1_raw;	/* fitted value of y1 fiber */
    float y2;		/* real space y2 */
    float y2_offset;	/* delta-y in fiber space? */
    float y2_raw;	/* fitted value of y2 fiber */
    float y3;		/* real space y3 */
    float y3_raw;	/* fitted value of y3 fiber */
    float yoffset;	/* delta-y in real space? */
    float xb1,yb1,xb2,yb2,xb3,yb3;   /* for use in bundle_hit search */
    float mx,bx,my,by;	/* ditto */

    int *xtemp;		/* used in sorting of raw evt blobs */
    int *ytemp;
    long *Itemp;

    if (srcp == NULL || kp == NULL) {
	set_err_code(420, "multitraj: recv'd NULL ptr");
	return -1;
    }

    if (evttype == ISCRIS) {
	srcc = (struct sc_crisEvt *)srcp;
    } else if (evttype == ISMSU) {
	srcm = (struct sc_msuEvt *)srcp;
    } else if (evttype == ISRAW) {
	srcr = (struct raw_evt *)srcp;
    } else {
	set_err_code(431, "multitraj: bad evttype parameter");
	return -1;
    }

    if (evttype == ISCRIS) {
	/* cris event */
	nblobs = srcc->nblobs;
    } else if (evttype == ISMSU) {
	/* msu event */
	if (cam == SC_CAM2) {
	    nblobs = srcm->nblobsb;
	} else {
	    nblobs = srcm->nblobs;
	}
    } else {
	/* ISRAW */
	nblobs = srcr->nblobs;
    }

#ifdef NOTDEF
    /* deleted to first check for out of geometry events */
    if(nblobs < SC_NLAYERS) {
	/* not all layewrs present */
	set_err_code(421, "multitraj: not enough blobs (%d)", nblobs);
	return RET_NOT_ENOUGH_BLOBS;
    }
#endif
    if (!Traj_init) {
	traj_init();
    }

    memset(ct, '\0', Maxtraj * sizeof(struct sc_Coords_Traj));

    if (evttype == ISCRIS) {
#ifdef NOTDEF
	/*  Cris data are delivered sorted by intensity so sorting is not
	 *  needed. But since we refer to the dest structure from now on we
	 *  better copy the original [using memcpy]. [jkl] */
	sc_cris_blob_sort(srcc, &destc);
#else
        memcpy(&destc, srcc, sizeof(struct sc_crisEvt));
#endif
    } else if (evttype == ISMSU) {
	/* this actually sorts both the cam1 and cam2 events */
	sc_msu_blob_sort(srcm, &destm);
    } else {
	/* ISRAW */
	xtemp = (int *)calloc(srcr->nblobs, sizeof(int));
	ytemp = (int *)calloc(srcr->nblobs, sizeof(int));
	Itemp = (long *)calloc(srcr->nblobs, sizeof(long));
	if (xtemp == NULL || ytemp == NULL || Itemp == NULL) {
	    free(xtemp);
	    free(ytemp);
	    free(Itemp);
	    set_err_code(432, 
		"multitraj: can't calloc xtemp, ytemp, or Itemp");
	    return -1;
	}
	sc_raw_blob_sort(srcr->x, srcr->y, srcr->I, 
	    xtemp, ytemp, Itemp, srcr->nblobs);

	memcpy(srcr->x, xtemp, srcr->nblobs * sizeof(int));
	memcpy(srcr->y, ytemp, srcr->nblobs * sizeof(int));
	memcpy(srcr->I, Itemp, srcr->nblobs * sizeof(long));
	free(xtemp);
	free(ytemp);
	free(Itemp);
    }

    for (i=0; i<SC_NLAYERS; i++) {
	Layer_maxI[i] = 0L;
	lblobs[i] = 0;
    }

    for (i=0; i < nblobs ; i++) {
	if (evttype == ISCRIS) {
	    x = destc.hodo[i].x;
	    y = destc.hodo[i].y;
	    I = destc.hodo[i].I;
	} else if (evttype == ISMSU) {
	    if (cam == SC_CAM2) {
		x = destm.blobxb[i];
		y = destm.blobyb[i];
		I = destm.blobIb[i];
	    } else {
		x = destm.blobx[i];
		y = destm.bloby[i];
		I = destm.blobI[i];
	    }
	} else {
	    x = srcr->x[i];
	    y = srcr->y[i];
	    I = srcr->I[i];
	}

	if (x >= SC_FMAPX) {
	    continue;
	}
	fmapp = sc_centr_to_fiber(cam, (int)x, (int)y);
	if (fmapp == NULL) {
	    continue;
	}
	lyr = fmapp->layer;
	if (lyr != SC_NOLAYER) {
	    n = lblobs[lyr];
	    hopdir = 0;
	    assign_blob(n, lyr, lyr, x, y, I, fmapp, hopdir, i);
	    if (Layer_maxI[lyr] < I) {
		/* set maximum intensity for this layer */
		Layer_maxI[lyr] = I;
	    }
	    if(++lblobs[lyr] >= Traj_maxblobs) {
		goto Find_ratios;
	    }
	}
    }
    if (Do_hop) {
	/* if intensity is low add hopped blobs to structures */
        for (i=0; i < nblobs ; i++) {
	    if (evttype == ISCRIS) {
	        x = destc.hodo[i].x;
	        y = destc.hodo[i].y;
	        I = destc.hodo[i].I;
	    } else if (evttype == ISMSU) {
	        if (cam == SC_CAM2) {
	     	    x = destm.blobxb[i];
	     	    y = destm.blobyb[i];
	     	    I = destm.blobIb[i];
	        } else {
	     	    x = destm.blobx[i];
	     	    y = destm.bloby[i];
	     	    I = destm.blobI[i];
	        }
	    } else {
		/* ISRAW */
		x = srcr->x[i];
		y = srcr->y[i];
		I = srcr->I[i];
	    }
	    if(I < Hopthresh) {
		if(x>=SC_FMAPX) continue;
		fmapp = sc_centr_to_fiber(cam, (int)x, (int)y);
	        if (fmapp == NULL) {
	            continue;
	        }
		if(fmapp->layer == SC_NOLAYER) {
		    oldfiber = 0;
		    oldlayer = -1;
		} else {
		    oldfiber = fmapp->fiber;
		    oldlayer = fmapp->layer;
		}
		fmapp = &felem;
		for(hopdir = 1;hopdir>-2;hopdir -= 2) {	/* hop up then down*/
		    ret = sc_hop(hopdir, cam, (int)x, (int)y, 
		        oldfiber,  &dy, fmapp);
		    if (ret == 0) {
		        lyr = fmapp->layer;
		        n = lblobs[lyr];
		        assign_blob(n, lyr, oldlayer, x, y+dy, I, 
				fmapp, hopdir, i);
		        if(Layer_maxI[lyr] < I) {
			    Layer_maxI[lyr] = I;
		        }
		        if(++lblobs[lyr] >= Traj_maxblobs) {
			    goto Find_ratios;
		        }
		    }
		}
	    }
	}
    }

Find_ratios:
    /* If we do not have 4 centroids in the bottom layers - quit */
    if (!lblobs[SC_X2] || !lblobs[SC_Y2] || !lblobs[SC_X3] || !lblobs[SC_Y3]) {
	    set_err_code(424, "multitraj: not all layers (2 or 3)");
	    return (RET_NOT_ALL_LAYERS_23);
    }

    /* First check whether this event was probably triggered by hitting the
     * fiber bundle.  Use the brightest blobs in layers 2 and 3 to
     * construct a trajectory.  If the extrapolation into layer 1 goes into
     * the bundle area a bundle hit is the most likely origin for this
     * event and further analysis would just yield noise!! */

    if(!sc_fiber_to_real(cam,2,blob[SC_X2][0].fiber,blob[SC_Y2][0].fiber,
    		&xb2,&yb2) &&
	    !sc_fiber_to_real(cam,3,blob[SC_X3][0].fiber,blob[SC_Y3][0].fiber,
		    &xb3,&yb3))  {
	if (Check_fiber_bundle_hit) {
	    if(fib_bundle_chk(xb2, xb3, yb2, yb3)) {
		set_err_code(426,"multitraj: fiber bundle hit");
		return RET_BUNDLE_HIT;
	    }
	}
    }

    /* now check if we have blobs in the top layer */
    if (!lblobs[SC_X1] || !lblobs[SC_Y1]) {
	/* we do not have a blob for each layer */
	set_err_code(425, "multitraj: not all layers (1)");
	return (RET_NOT_ALL_LAYERS_1);
    }

    /* For each blob find the ratio of its intensity to the maximum
     * intensity in its layer.  */
    for(lyr=0; lyr<SC_NLAYERS; lyr++) {
	 for(n=0; n<lblobs[lyr]; n++) {
	    blob[lyr][n].ratio = (float)blob[lyr][n].I / (float)Layer_maxI[lyr];
	 }
    }

    /* for the entire array of blobs find those who yield alignment in 
     * the x direction. For the triads that meet the criterion save 
     * the trio of indices of the blob structures.  */
    ltrkx = 0;
    for(i=0;i<=lblobs[SC_X1]+lblobs[SC_X2]+lblobs[SC_X3]-3;i++) {
        for (lblobs1=0; lblobs1 <= i; lblobs1++) {
	    if(lblobs1 >= lblobs[SC_X1]) continue;
	    x1_raw = Fitfunc(SC_X1, blob[SC_X1][lblobs1].fiber);
	    for (lblobs2=0; lblobs2 <= i - lblobs1; lblobs2++) {
	        if(lblobs2 >= lblobs[SC_X2]) continue;
	        x2_raw = Fitfunc(SC_X2, blob[SC_X2][lblobs2].fiber);
	        lblobs3 = i - lblobs1 - lblobs2;
	        if (lblobs3 >= lblobs[SC_X3]) continue;
		if(blob[SC_X1][lblobs1].hop && blob[SC_X2][lblobs2].hop &&
		       blob[SC_X3][lblobs3].hop) {
		    /* skip if all three have hopped */
		   continue;
		}
	        x3_raw = Fitfunc(SC_X3, blob[SC_X3][lblobs3].fiber);
		x2_offset = x2_raw - x1_raw - Fit_x_factor * (x3_raw - x1_raw);
		if ((x2_offset > Fit_x_lim_lo) && (x2_offset < Fit_x_lim_hi)) {
		    trackx[ltrkx].idx[SC_X1] = lblobs1;
		    trackx[ltrkx].idx[SC_X2] = lblobs2;
		    trackx[ltrkx].idx[SC_X3] = lblobs3;
		    trackx[ltrkx].offset = (short) (x2_offset*1000);
		    if(++ltrkx >= Maxtracks) {
			goto y_tracks;
		    }
		}
	    }
        }
    }

    /* if no triad was found, quit */
    if(!ltrkx) {
	set_err_code(427, "multitraj: no X fit");
	return(RET_NO_X_FIT);
    }
    Evtprm.nxtrack = ltrkx;

y_tracks:
    ltrky = 0;
    for(i=0;i<=lblobs[SC_Y1]+lblobs[SC_Y2]+lblobs[SC_Y3]-3;i++) {
        for (lblobs1=0; lblobs1 <= i; lblobs1++) {
	    if(lblobs1 >= lblobs[SC_Y1]) continue;
	    y1_raw = Fitfunc(SC_Y1, blob[SC_Y1][lblobs1].fiber);
	    for (lblobs2=0; lblobs2 <= i - lblobs1; lblobs2++) {
	        if(lblobs2 >= lblobs[SC_Y2]) continue;
	        y2_raw = Fitfunc(SC_Y2, blob[SC_Y2][lblobs2].fiber);
	        lblobs3 = i - lblobs1 - lblobs2;
	        if (lblobs3 >= lblobs[SC_Y3]) continue;
		if(blob[SC_Y1][lblobs1].hop && blob[SC_Y2][lblobs2].hop &&
		       blob[SC_Y3][lblobs3].hop) {
		    /* skip if all three have hopped */
		   continue;
		}
		y3_raw = Fitfunc(SC_Y3, blob[SC_Y3][lblobs3].fiber);
		y2_offset = y2_raw - y1_raw - Fit_y_factor * (y3_raw - y1_raw);
		if ((y2_offset > Fit_y_lim_lo) && (y2_offset < Fit_y_lim_hi)) {
		    tracky[ltrky].idx[SC_Y1] = lblobs1;
		    tracky[ltrky].idx[SC_Y2] = lblobs2;
		    tracky[ltrky].idx[SC_Y3] = lblobs3;
		    tracky[ltrky].offset = (short) (y2_offset*1000);
		    if(++ltrky >= Maxtracks) {
			goto Real_traj;
		    }
		}
	    }
	}
    }

    if (!ltrky) {
	set_err_code(429, "multitraj: no Y fit");
	return (RET_NO_Y_FIT);
    }
    Evtprm.nytrack = ltrky;

Real_traj:
    init_traj_mb();

    /* find the real space coordinates for all combinations of triads */
    for(sum=0; sum <= ltrkx+ltrky -2; sum++) {
          for(trx=0; trx <= sum; trx++) {
	    if(trx >= ltrkx) {
		break;
	    }
	    try = sum - trx;
	    if(try >= ltrky) {
		continue;
	    }
	    if (sc_fiber_to_real(cam, 1,
		    blob[SC_X1][trackx[trx].idx[SC_X1]].fiber, 
		    blob[SC_Y1][tracky[try].idx[SC_Y1]].fiber, &x1, &y1)) {
	        continue;
	    }

	    if (sc_fiber_to_real(cam, 2,
		    blob[SC_X2][trackx[trx].idx[SC_X2]].fiber, 
		    blob[SC_Y2][tracky[try].idx[SC_Y2]].fiber, &x2, &y2)) {
	        continue;
	    }

	    if (sc_fiber_to_real(cam, 3,
		    blob[SC_X3][trackx[trx].idx[SC_X3]].fiber, 
		    blob[SC_Y3][tracky[try].idx[SC_Y3]].fiber, &x3, &y3)) {
	        continue;
	    }
	     
	    if (Check_deviation) {
		/* keep only if real space coordinates align properly */
		if (!sc_x_dev_ok(x1, x2, x3, &xoffset)) {
		    continue;
		}

		if (!sc_y_dev_ok(y1, y2, y3, &yoffset)) {
		    continue;
		}
	    }
	     
	    /* Fill crisCoords structures for each trajectory found
	     * including the quality of the fit and the hopping associated
	     * with each centroid */
	    fit_qual = (float)sqrt((double)SQR(xoffset) + (double)SQR(yoffset));
	    ct[ltraj].Q.hop_param = 0;  /* initialization */
	    ct[ltraj].C.cam = cam;
	    ct[ltraj].Q.fit_param = fit_qual;
	    ct[ltraj].C.x1 = x1;
	    ct[ltraj].C.x2 = x2;
	    ct[ltraj].C.x3 = x3;
	    ct[ltraj].C.y1 = y1;
	    ct[ltraj].C.y2 = y2;
	    ct[ltraj].C.y3 = y3;
#ifdef NOTDEF
	    minI = 1e9;
	    sumI = 0;
#endif
	    init_hop_param();
	    for (i=0; i<SC_NLAYERS; i++) {
		if (i == SC_X1 || i == SC_X2 || i == SC_X3) {
		    n = trackx[trx].idx[i];
		    ct[ltraj].Q.raw_dx_1000 = trackx[trx].offset;
		} else {
		    n = tracky[try].idx[i];
		    ct[ltraj].Q.raw_dy_1000 = tracky[try].offset;
		}
		ct[ltraj].C.fiber[i] = blob[i][n].fiber;
		ct[ltraj].C.ypos[i] = blob[i][n].ypos;
		ct[ltraj].C.xpix[i] = blob[i][n].x;
		ct[ltraj].C.ypix[i] = blob[i][n].y;
		ct[ltraj].C.I[i] = I = blob[i][n].I;
		ct[ltraj].Q.Iratio[i] = blob[i][n].ratio;
		ct[ltraj].blobid[i] = blob[i][n].id;
#ifdef NOTDEF
		sumI += I;
		if (I < minI) {
		    minI = I;
		}
#endif
		accum_hop_param(blob[i][n].hop, i);
		qualtrack.idx[i] = n;
	    }
	    ct[ltraj].Q.hop_param = get_hop_param();
#ifdef ONE_HOP
	    hop_parm = ct[ltraj].Q.hop_param;
	    if (hop_parm > 0) {
	        nhops = 0;
	        for(i=0; i<SC_NLAYERS; i++) {
	            if(hop_parm & 0x7) {
			nhops++;
		    }
		    hop_parm >>= 3;
	        }
	        if (nhops > 1) {
		    continue;
		}
	    }
#endif /* ONE_HOP */

	    ret = sc_trajectory(&ct[ltraj].C, &ct[ltraj].T);
	    tel_no = 0;
	    if (Check_telescope) {
		if (evttype == ISCRIS) {
		    tel_no = sc_telescope_hit(srcc, &ct[ltraj].T, SC_E1, 50);
		}
	    }
	    if(tel_no < 0) {
	       Nmb--;
	       continue;
	    } else {
	       Evtprm.tel = (char) tel_no;
	    }
	    ct[ltraj].Q.qual_factor = 
	    	quality(&qualtrack, ct[ltraj].Q.hop_param, fit_qual, tel_no);

	    if(++ltraj == Maxtraj) {
	        break;
	    }
	}
	if(ltraj == Maxtraj) {
	    break;
	}
    }
    if (ltraj == 0) {
	set_err_code(430, "multitraj: no fit");
	return(RET_NO_FIT);
    }

    Evtprm.ntraj = ltraj;

    /* sort trajectories by quality... */
    if(ltraj > 1) {
	qsort((char *)ct, ltraj, sizeof(struct sc_Coords_Traj), qual_compare); 
    }

    /* ...and copy the best Coords and Traj into the calling program.
     * ntraj were desired and will be delivered unless we found fewer
     * (ltraj is less). */
    if(ltraj > ntraj) {
	ltraj = ntraj;
        Evtprm.nextqual = ct[ntraj].Q.qual_factor;
    } else {
        Evtprm.nextqual = 0;
    }
    for(i=0; i < ltraj; i++) {
	kp[i] = ct[i];
    }
    return (ltraj);
}

/* sc_telescope_hit - determine if the trajectory falls within the radius r
 * of the "telescope" it was supposed to have hit. Returns the telescope
 * value (SC_TEL_A, SC_TEL_B, etc) on success, else -1. */
int
sc_telescope_hit(c, t, stack, r)
    struct sc_crisEvt *c;
    struct sc_softTraj *t;
    int stack;		/* which stack layer (SC_E1, SC_E2, SC_E3_1, etc) */
    int r;		/* radius of "hit area" in mm. */
{
    int	tel;
    double x;
    double y;
#ifdef NOTDEF
    double x_center[4] = { 75.70, 75.70, 196.30, 196.30};
    double y_center[4] = {196.30, 75.70, 196.30,  75.70};
#endif
    double x_center[4] = { 80.0, 80.0, 197.0, 197.0};
    double y_center[4] = {194.0, 78.0, 194.0,  78.0};

    if (r < 0) {
	set_err_code(1000, "sc_telescope_hit: negative radius (%d)", r);
	return -1;
    }

    if (stack < 0 || stack >= SC_NSTACK) {
	set_err_code(1001, "sc_telescope_hit: bad stack value (%d)", stack);
	return -1;
    }

    if (c->prim_tel_pair == SC_CRIS_AB) {
	if (c->prim_ctr_ph[0] > c->prim_ctr_ph[1]) {
	    tel = SC_TEL_A;
	} else {
	    tel = SC_TEL_B;
	}
    } else {
	if (c->prim_ctr_ph[0] > c->prim_ctr_ph[1]) {
	    tel = SC_TEL_C;
	} else {
	    tel = SC_TEL_D;
	}
    }

    x = t->stackx[stack] - x_center[tel];
    y = t->stacky[stack] - y_center[tel];
    if(SQR(x) + SQR(y) > SQR(r)) {
	set_err_code(1002, "sc_telescope_hit: no hit");
	return -1;
    } else {
        return (tel);
    }
}

/* sc_hop - do centroid hopping in y. Based on code by Joe Klarmann.
 * Returns 0 if we hopped successfully, else -1. The amount dy that we
 * hopped will be pointed to by dyp, and information about the new fiber
 * will be pointed to by fp. */
int
sc_hop(dir, cam, xpix,  ypix, fiberno, dyp, fp)
    int dir;			/* hop direction, either < 0 or > 0 */
    int cam;
    int xpix;
    int ypix;
    float fiberno;		/* original fiber number */
    int *dyp;			/* diff in y that we hopped */
    struct sc_fmapElem *fp;	/* new hopped fiber */
{
    int delta;			/* hop increment */
    float df;			/* difference in fiber number */
    int dy;
    struct sc_fmapElem *f;

    if (dir >= 0) {
	/* hop in positive y */
	delta = HOPDELTA;
    } else {
	/* hop in negative y */
	delta = -HOPDELTA;
    }

    /* We exit the loop below if df, the fiber difference, exceeds 2. Joe
     * Klarmann explains the reasoning behind this:
     *
     *  Changing ypix should not normally change the fiber number at all
     *  until one hits a new tab.  However I was afraid of a tilt somewhere
     *  in the fiber arrangement and so I allowed a small change without
     *  accepting it as a new tab.  I chose to use a minimum of 3 fibers to
     *  indicate a change in tab,  although of course the tab change will
     *  always be a much larger change in fiber number.  I did not test
     *  this though!!!  Indeed the whole routine should be replaced by a
     *  table lookup at some stage (but after the final mapping is in).
     */

    dy = 0;
    do {
	dy += delta;
	if (ypix + dy < 0) {
	    return -1;
	}
	if(xpix>=SC_FMAPX) continue;
	f = sc_centr_to_fiber(cam, xpix, ypix + dy);
	if (f == NULL || f->layer == SC_NOLAYER) {
	    return -1;
	}
	df = f->fiber - fiberno;
    } while (-2.0 <= df && df <= 2.0);	/* got new tab if |df| > 2 */

    *dyp = dy;
    *fp = *f;
    return 0;
}

static void
assign_blob(n, layer, oldlayer, xpix, ypix, I, f, hopdir, id)
    int n;
    int layer;
    int oldlayer;
    int xpix;
    int ypix;
    int I;
    struct sc_fmapElem *f;
    int hopdir;		/* hop direction (-1, 0, 1) */
    int id;		/* blob id */
{
    int hop;
    
    /* hop = 1 if up, 2 if down */
    /* the '4' bit is set if we switched layers while hopping */
    if (hopdir < 0) {
	hop = 2;
    } else if (hopdir > 0) {
	hop = 1;
    } else {
	hop = 0;
    }
    if (hop != 0 && layer != oldlayer) {
	hop |= 0x4;
    }

    blob[layer][n].x = xpix;
    blob[layer][n].y = ypix;
    blob[layer][n].I = I;
    blob[layer][n].fiber = f->fiber;
    blob[layer][n].ypos = f->ypos/128.0;
    blob[layer][n].hop = hop;
    blob[layer][n].id = (u_char)id;
}

int
sc_do_hopping(int doit)
{
    int tmp;

    tmp = Do_hop;
    if (doit) {
	Do_hop = 1;
    } else {
	Do_hop = 0;
    }
    return tmp;
}
 
static void
init_evtparams(void)
{
    /* Initialize the event parameters, update as needed */
    Evtprm.nextqual = 0.;
    Evtprm.nxtrack = 0;
    Evtprm.nytrack = 0;
    Evtprm.ntraj = 0;
    Evtprm.tel = -1;
}

/* [These comments are by Joe Klarmann, 1 Sep 1998]
 * sc_cris_traj(struct sc_crisEvt *e, struct sc_Coords_Traj *t,int *ntraj);
 *    ntraj points to the number of trajectories desired, upon return it
 *        will contain the number of trajectories found (up to the original
 *        request);
 *    e is the input event structure defined in crisevt.h;
 *    t points to the output structure(s) (ntraj of them)
 *
 * If all goes well, sc_cris_traj returns 0, and fills in the int pointed to 
 * by ntraj, the number of trajectories having been stored in t. 
 *
 * However, if an error occurs, it returns -1.
 *
 * If more detailed information on the nature of the error is desired, 
 * the routine sc_get_err_code(void) returns an integer code number, and 
 * the routine sc_get_err_mesg(void) returns a pointer to a string 
 * describing the failure.
 *
 * There are 9 different modes of operation for sc_cris_traj.  Unless the
 * default is acceptable a different method may be set using 
 * sc_traj_select_method(int m), where m is the number from the following
 * table:
 *
 *    0  "brightest each layer"
 *    1  "fiber_fit, no-hopping "linear_fit"
 *    2  "fiber_fit, hopping allowed "linear_fit
 *    3  "fiber_fit, no_hopping "quadratic_fit"
 *    4  "fiber_fit, hopping allowed "quadratic_fit"
 *    5  "multitraj, no-hopping "linear_fit"
 *    6  "multitraj, hopping allowed "linear_fit
 *    7  "multitraj, no_hopping "quadratic_fit"
 *    8  "multitraj, hopping allowed "quadratic_fit"
 *
 * The default is method 8.
 * 
 * brightest each layer, simply computes the trajectory based on the
 * brightest centroid in each layer. (This method was the initial one
 * used.)  
 * 
 * fiber_fit checks for coarse alignment of the x-centroids sequentially
 * proceeding to dimmer blobs until alignment is found.  This is followed
 * by a similar search through the y-centroids. The search stops once an
 * alignment is found.
 *
 * multitraj checks all reaonable combinations of blobs for trajectories,
 * retaining x_tracks for those which have coarse alignment in the
 * x-direction and y-tracks for those having same in the y-direction.
 * Every x-track is combined with every y-track and a quality factor is
 * assigned to the resulting trajectories based on the goodness of fit,
 * the brightness of the blobs involved and the nature of the hopping
 * involved.
 * 
 * linear vs quadratic fit refer to the coarse alignment calculated which
 * is based on the fiber number only and is independent of the conjugate
 * coordinate layer.  (The linear fit is quite obsolete).
 *
 * Errors:  (Apart from bad commands etc.)
 * Trajectories are rejected if:
 *	a) There are not enough centroids to form a trajectory,
 *	b) Trajectories based on the second and third hodo-layer alone
 *	point to the fiber bundle in the top layer.
 *     c) delta x or delta y for the center layer is greater than 0.5mm
 */

int
sc_cris_traj(e, t, ntraj)
    struct sc_crisEvt *e;	/* ptr to cris evt */
    struct sc_Coords_Traj *t;	/* ptr to array of traj structs */
    int *ntraj;			/* no. trajectories desired for multitraj */
{
#ifdef ONE_HOP
    int nhops;
    unsigned hop_parm;
#endif
    int cam;
    int i;
    int ret;
    float xoffset,yoffset;

    if (e == NULL || t == NULL || ntraj == NULL) {
	set_err_code(100, "sc_cris_traj: recv'd NULL parameter c or t");
	return -1;
    }

    cam = sc_cris_get_cam();

    init_evtparams();
    if (Check_E1) {
	if(e->prim_ctr_ph[0] == 0 && e->prim_ctr_ph[1] == 0) {
	    set_err_code(102, "sc_cris_traj: zero pha in top detector");
	    Evtprm.ret_code = RET_NO_E1_PHA;
	    return -1;
	}
    }

    clear_blobid();

    switch (Traj_select_method) {
	case 0:					/* brightest each layer */
	case 1: case 2: case 3: case 4:		/* fiber fit */
	    ret = sc_cris_coords(e, &t->C, cam);
	    if (ret == 0) {
		ret = sc_trajectory(&t->C, &t->T);
	    }
	    *ntraj = (ret == 0) ? 1 : 0;
	    for(i=0; i<SC_NLAYERS;i++) {
	        t->blobid[i] = Blobid[i];
	    }
	    xoffset = sc_x_dev(t->C.x1,t->C.x2,t->C.x3);
	    yoffset = sc_y_dev(t->C.y1,t->C.y2,t->C.y3);
	    t->Q.fit_param = (float)sqrt((double)SQR(xoffset) + 
	        (double)SQR(yoffset));
	    if(Traj_select_method < 1) break;
	    t->Q.hop_param = get_hop_param();
#ifdef ONE_HOP
	    hop_parm = t->Q.hop_param;
	    if(hop_parm > 0) {
	        nhops = 0;
	        for(i=0; i<SC_NLAYERS; i++) {
	            if(hop_parm & 0x7) {
			nhops++;
		    }
		    hop_parm >>= 3;
	        }
	        if(nhops > 1) {
		    set_err_code(103, "sc_cris_traj: more than 1 hop");
		    ret = -1;
		}
	    }
#endif /* ONE_HOP */
	    break;
	case 5: case 6: case 7: case 8:		/* multitraj */
	    ret = multitraj(e, t, cam, *ntraj, ISCRIS);
	    *ntraj = ret;
	    break;
	default:
	    set_err_code(101, "sc_cris_traj: garbled Traj_select_method = %d",
		Traj_select_method);
	    ret = -1;
	    break;
    }

    Evtprm.ret_code = ret;
    if (ret < 0) {
	return -1;
    } else {
	return 0;
    }
}

/* sc_msu_traj - returns 0 if all went well. If the cam 1 failed, the 0x1
 * bit is set and *ntraj1 = -1, if cam 2, 0x2 is set and *ntraj2 = -1. */
int 
sc_msu_traj(e, t1, t2, ntraj1, ntraj2)
    struct sc_msuEvt *e;
    struct sc_Coords_Traj *t1;
    struct sc_Coords_Traj *t2;
    int *ntraj1;
    int *ntraj2;
{
    int fixret;
    int ret;
    int ret1;
    int ret2;

    if (e == NULL || t1 == NULL || t2 == NULL || 
		ntraj1 == NULL || ntraj2 == NULL) {
    	set_err_code(110, "sc_msu_traj: recv'd NULL parameter");
	return -1;
    }

    ret1 = ret2 = -1;

    fixret = sc_msu_fix_blobs(e);
    if (fixret == 0x3) {
	/* both cameras failed fix_blobs() */
	set_err_code(109, "sc_msu_traj: can't fix blobs both cameras");
    }
    
    clear_blobid();

    switch (Traj_select_method) {
	case 0:					/* brightest each layer */
	case 1: case 2: case 3: case 4:		/* fiber fit */
	    if ((fixret & 0x1) == 0) {
		ret = sc_msu_coords(e, &t1->C, &t2->C);
		if ((ret & 0x10000) == 0) {
		    ret1 = sc_trajectory(&t1->C, &t1->T);
		}
		t1->Q.hop_param = get_hop_param();
	    }
	    *ntraj1 = (ret1 == 0) ? 1 : 0;

	    if ((fixret & 0x2) == 0) {
		if ((ret & 0x20000) == 0) {
		    ret2 = sc_trajectory(&t2->C, &t2->T);
		}
		t2->Q.hop_param = get_hop_param();
	    }
	    *ntraj2 = (ret2 == 0) ? 1 : 0;
	    break;

	case 5: case 6: case 7: case 8:		/* multitraj */
	    if ((fixret & 0x1) == 0) {
		ret1 = multitraj(e, t1, SC_CAM1, *ntraj1, ISMSU);
	    }
	    *ntraj1 = ret1;

	    if ((fixret & 0x2) == 0) {
		ret2 = multitraj(e, t2, SC_CAM2, *ntraj2, ISMSU);
	    }
	    *ntraj2 = ret2;
	    break;

	default:
	    set_err_code(111, "sc_msu_traj: garbled Traj_select_method = %d",
		Traj_select_method);
	    ret1 = ret2 = -1;
	    break;
    }

    ret = 0;
    if (ret1 < 0) {
	ret = 0x1;
    }
    if (ret2 < 0) {
	ret |= 0x2;
    }
    return ret;
}

/* sc_raw_traj() - calculate the coords and trajectory with input from
 * arrays of X, Y, and intensity instead of a cris or msu event. */
int
sc_raw_traj(x, y, I, nblobs, t, ntraj)
    int *x;
    int *y;
    long *I;
    int nblobs;
    struct sc_Coords_Traj *t;	/* ptr to array of traj structs */
    int *ntraj;			/* no. trajectories desired for multitraj */
{
#ifdef ONE_HOP
    int nhops;
    unsigned hop_parm;
#endif
    int cam;
    int i;
    struct raw_evt r;
    int ret;
    float xoffset,yoffset;

    if (x == NULL || y == NULL || I == NULL || t == NULL || ntraj == NULL) {
	set_err_code(120, 
	    "sc_raw_traj: recv'd NULL parameter x,y,I,c,t, or ntraj");
	return -1;
    }

    r.x = x;
    r.y = y;
    r.I = I;
    r.nblobs = nblobs;

    cam = sc_cris_get_cam();

    clear_blobid();

    switch (Traj_select_method) {
	case 0:					/* brightest each layer */
	case 1: case 2: case 3: case 4:		/* fiber fit */
	    ret = sc_raw_coords(x, y, I, nblobs, cam, &t->C);
	    if (ret == 0) {
		ret = sc_trajectory(&t->C, &t->T);
	    }
	    *ntraj = (ret == 0) ? 1 : 0;
	    for(i=0; i<SC_NLAYERS;i++) {
	        t->blobid[i] = Blobid[i];
	    }
	    xoffset = sc_x_dev(t->C.x1,t->C.x2,t->C.x3);
	    yoffset = sc_y_dev(t->C.y1,t->C.y2,t->C.y3);
	    t->Q.fit_param = (float)sqrt((double)SQR(xoffset) + 
	        (double)SQR(yoffset));
	    if(Traj_select_method < 1) {
		break;
	    }
	    t->Q.hop_param = get_hop_param();
#ifdef ONE_HOP
	    hop_parm = t->Q.hop_param;
	    if(hop_parm > 0) {
	        nhops = 0;
	        for(i=0; i<SC_NLAYERS; i++) {
	            if(hop_parm & 0x7) {
			nhops++;
		    }
		    hop_parm >>= 3;
	        }
	        if(nhops > 1) {
		    set_err_code(123, "sc_raw_traj: more than 1 hop");
		    ret = -1;
		}
	    }
#endif /* ONE_HOP */
	    break;
	case 5: case 6: case 7: case 8:		/* multitraj */
	    ret = multitraj(&r, t, cam, *ntraj, ISRAW);
	    *ntraj = ret;
	    break;
	default:
	    set_err_code(121, "sc_raw_traj: garbled Traj_select_method = %d",
		Traj_select_method);
	    ret = -1;
	    break;
    }

    if (ret < 0) {
	return -1;
    } else {
	return 0;
    }
}

/* hop parameter stuff - we localize the calculation of the hop parameter
 * here. These functions are used by fiberfit() and multitraj(). */

static int Hop_param = 0;	/* current value of the hop parameter */

static void
init_hop_param(void)
{
    Hop_param = 0;
}

/* accum_hop_param - should be called once for each hop factor in the blob
 * used in each layer for the trajectory */
static void
accum_hop_param(hop, n)
    int hop;	/* hop value */
    int n;	/* layer */
{
    Hop_param |= (hop << (3 * n));
}

static int
get_hop_param(void)
{
    return Hop_param;
}

/* sc_get_x_dev_limits - return the values of the lower and upper delta X2
 * limits and the X fit factor */
void
sc_get_x_dev_limits(limlo, limhi, factor)
    float *limlo;
    float *limhi;
    float *factor;
{
    *limlo = Real_x_lim_lo;
    *limhi = Real_x_lim_hi;
    *factor = Fit_x_factor;
}

void
sc_set_x_dev_limits(limlo, limhi, factor)
    float limlo;
    float limhi;
    float factor;
{
    Real_x_lim_lo = limlo;
    Real_x_lim_hi = limhi;
    Fit_x_factor = factor;
}

void
sc_set_y_dev_limits(limlo, limhi, factor)
    float limlo;
    float limhi;
    float factor;
{
    Real_y_lim_lo = limlo;
    Real_y_lim_hi = limhi;
    Fit_y_factor = factor;
}

/* sc_get_y_dev_limits - return the values of the lower and upper delta Y2
 * limits and the Y fit factor */
void
sc_get_y_dev_limits(limlo, limhi, factor)
    float *limlo;
    float *limhi;
    float *factor;
{
    *limlo = Real_y_lim_lo;
    *limhi = Real_y_lim_hi;
    *factor = Fit_y_factor;
}

/* sc_x_dev_ok - return 1 if delta X2 is within limits, else 0 */
int
sc_x_dev_ok(x1, x2, x3, xdev) 
    float x1;
    float x2;
    float x3;
    float *xdev;
{
    float d;
    d = sc_x_dev(x1, x2, x3);
    if (xdev != NULL) {
	*xdev = d;
    }
    if (d < Real_x_lim_lo || d > Real_x_lim_hi) {
	return 0;
    } else {
	return 1;
    }
}

/* sc_y_dev_ok - return 1 if delta Y2 is within limits, else 0 */
int
sc_y_dev_ok(y1, y2, y3, ydev) 
    float y1;
    float y2;
    float y3;
    float *ydev;
{
    float d;
    d = sc_y_dev(y1, y2, y3);
    if (ydev != NULL) {
	*ydev = d;
    }
    if (d < Real_y_lim_lo || d > Real_y_lim_hi) {
	return 0;
    } else {
	return 1;
    }
}

/* sc_x_dev - return the value of "delta X2" */
float
sc_x_dev(x1, x2, x3) 
    float x1;
    float x2;
    float x3;
{
    float dev;
    dev = x2 - x1 - Fit_x_factor * (x3-x1);
    return dev;
}

/* sc_y_dev - return the value of "delta Y2" */
float
sc_y_dev(y1, y2, y3)
    float y1;
    float y2;
    float y3;
{
    float dev;
    dev = y2 - y1 - Fit_y_factor * (y3-y1);
    return dev;
}

int
sc_check_e1(int doit)
{
    int tmp = Check_E1;

    if (doit) {
	Check_E1 = 1;
    } else {
	Check_E1 = 0;
    }
    return tmp;
}

int
sc_check_bundle(int doit)
{
    int tmp = Check_fiber_bundle_hit;

    if (doit) {
	Check_fiber_bundle_hit = 1;
    } else {
	Check_fiber_bundle_hit = 0;
    }
    return tmp;
}

int
sc_check_dev(int doit)
{
    int tmp = Check_deviation;

    if (doit) {
	Check_deviation = 1;
    } else {
	Check_deviation = 0;
    }
    return tmp;
}


int
sc_check_telescope(int doit)
{
    int tmp = Check_telescope;

    if (doit) {
	Check_telescope = 1;
    } else {
	Check_telescope = 0;
    }
    return tmp;
}

void
sc_check_chisq(int x, int y)
{
    Check_chi_x = x;
    Check_chi_y = y;
}

void
sc_set_max_traj_chisq(float x, float y)
{
    Max_chi_x = x;
    Max_chi_y = y;
}

void
sc_get_max_traj_chisq(float *x, float *y)
{
    *x = Max_chi_x;
    *y = Max_chi_y;
}

/* fib_bundle_chk  checks whether the trajectory defined by the bottom two
 * layers hits the fiber bundle in the top layer.  If it does, returns -1
 * else returns 0. (Joe Klarmann) */
static int
fib_bundle_chk(xb2, xb3, yb2, yb3)
    float xb2, xb3, yb2, yb3;	/* real coords in layers 2 and 3 */
{
    float mx, bx;	/* slope and intercept for x */
    float my, by;	/* slope and intercept for y */
    float xb1, yb1;	/* computed coords in layer 1 */

    mx = (xb3 - xb2) / (SC_Z[SC_X3] - SC_Z[SC_X2]);
    bx = xb3 - mx*SC_Z[SC_X3];
    my = (yb3 - yb2) / (SC_Z[SC_Y3] - SC_Z[SC_Y2]);
    by = yb3 - my*SC_Z[SC_Y3];
    xb1 = mx*SC_Z[SC_X1] + bx;
    yb1 = my*SC_Z[SC_Y1] + by;

    if(xb1 >= 10. && xb1 <= 300. && yb1 >= 264. && yb1 <= 300.) {
	return -1;
    }
    if(xb1 >= 264. && xb1 <= 297. && yb1 >= 20. && yb1 <= 320.) {
	return -1;
    }
    if(xb1 >= -40. && xb1 <= 260. && yb1 >= -24. && yb1 <= 10.) {
	return -1;
    }
    if(xb1 >= -20. && xb1 <= 12. && yb1 > 10. && yb1 <= 260.) {
	return -1;
    }
    return 0;
}

void
sc_set_z_coord(float x1, float y1, float x2, float y2, float x3, float y3)
{
    SC_Z[SC_X1] = x1;
    SC_Z[SC_Y1] = y1;
    SC_Z[SC_X2] = x2;
    SC_Z[SC_Y2] = y2;
    SC_Z[SC_X3] = x3;
    SC_Z[SC_Y3] = y3;
}

void
sc_get_z_coord(float *x1, float *y1, float *x2, float *y2, 
		float *x3, float *y3)
{
    *x1 = SC_Z[SC_X1];
    *y1 = SC_Z[SC_Y1];
    *x2 = SC_Z[SC_X2];
    *y2 = SC_Z[SC_Y2];
    *x3 = SC_Z[SC_X3];
    *y3 = SC_Z[SC_Y3];
}

/* sc_set_soft_stack_sep - set the separation between the bottom of SOFT to
 * the top of the stack detector. */
void
sc_set_soft_stack_sep(float sep)
{
    Soft_stack_sep = sep;
    set_stack_z();
}

float
sc_get_soft_stack_sep(void)
{
    return (Soft_stack_sep);
}

/* set_stack_z - set the values of SC_STACK_Z based on the current
 * Soft_stack_sep value. */
static void
set_stack_z(void)
{
    int i;
    for (i=0; i < NSTACKLAYER; i++) {
	SC_STACK_Z[i] = Stack_sep[i] - Soft_stack_sep;
    }
}

/* sc_get_stack_z - return a pointer to the SC_STACK_Z array containing the
 * Z coordinate of each of the stack layers. The parameter 'n' points to an
 * int containing the number of elements in the array. Each element can be
 * accessed using the stack layer ID macros in softcris.h (SC_E1, SC_E2,
 * etc.) */
float *
sc_get_stack_z(int *n)
{
    *n = NSTACKLAYER;
    return SC_STACK_Z;
}

/* sc_get_evt_params - get the per-event parameters accumulated by multitraj */
void
sc_get_evt_params(struct sc_evt_params *ep)
{
	ep->nextqual = Evtprm.nextqual; 
	ep->nxtrack = Evtprm.nxtrack; 
	ep->nytrack = Evtprm.nytrack; 
	ep->ntraj = Evtprm.ntraj; 
	ep->ret_code = Evtprm.ret_code; 
	ep->tel = Evtprm.tel; 
}

static void 
clear_blobid(void)
{
    int i;
    for (i=0; i<SC_NLAYERS; i++) {
	Blobid[i] = SC_NOLAYER;
    }
}

/* sc_set_maxtracks - set the value of Maxtracks. Returns the previous
 * value of Maxtracks, or -1 if the desired value is negative. */
int
sc_set_maxtracks(int n)
{
    int tmp = Maxtracks;

    if (n < 0) {
	return -1;
    } else {
	Maxtracks = n;
	return tmp;
    }
}

int
sc_get_maxtracks(void)
{
    return Maxtracks;
}

/* sc_set_maxtraj - set the value of Maxtraj. Returns the previous
 * value of Maxtraj, or -1 if the desired value is negative. */
int
sc_set_maxtraj(int n)
{
    int tmp = Maxtraj;

    if (n < 0) {
	return -1;
    } else {
	Maxtraj = n;
	Traj_init = 0;	/* force a traj_init() */
	return tmp;
    }
}

int
sc_get_maxtraj(void)
{
    return Maxtraj;
}

void
sc_set_fit_limits(xfac, xlo, xhi, yfac, ylo, yhi)
    float xfac;
    float xlo;
    float xhi;
    float yfac;
    float ylo;
    float yhi;
{
    Fit_x_factor = xfac;
    Fit_x_lim_lo = xlo;
    Fit_x_lim_hi = xhi;
    Fit_y_factor = yfac;
    Fit_y_lim_lo = ylo;
    Fit_y_lim_hi = yhi;
}

void
sc_get_fit_limits(xfac, xlo, xhi, yfac, ylo, yhi)
    float *xfac;
    float *xlo;
    float *xhi;
    float *yfac;
    float *ylo;
    float *yhi;
{
    *xfac = Fit_x_factor;
    *xlo = Fit_x_lim_lo;
    *xhi = Fit_x_lim_hi;
    *yfac = Fit_y_factor;
    *ylo = Fit_y_lim_lo;
    *yhi = Fit_y_lim_hi;
}

long *
sc_layer_maxI(void)
{
    return Layer_maxI;
}
